/*
 *  linux/arch/arm/mach-mb86hxx/timer.c
 *
 * Copyright (C) 2008-2010 FUJITSU MICROELECTRONICS LIMITED
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/clocksource.h>
#include <linux/clockchips.h>
#include <mach/hardware.h>
#include <mach/timer.h>
#include <asm/io.h>
#include <asm/mach/time.h>
#include <asm/hardware/arm_timer.h>

#define TIMER_RELOAD    (TIMER_INTERVAL)
#define TICKS2USECS(x)  ((x) / TICKS_PER_uSEC)
#define TIMER_INTERVAL  (TICKS_PER_uSEC * mSEC_10)

#ifdef CONFIG_MB86H6X_TIMER1_CLOCKSOURCE
#define TIMER1_OFFSET   0x40
#define USE_SCHED_CLOCK 1
#else
#define TIMER1_OFFSET   0x0
#define USE_SCHED_CLOCK 0
#endif

static void timer_set_mode(enum clock_event_mode mode,
			   struct clock_event_device *clk)
{
	switch(mode) {
	case CLOCK_EVT_MODE_PERIODIC:
		writel(0, TIMER_VA_BASE + TIMER_ENABLE);
		writel(0, TIMER_VA_BASE + TIMER_COUNT_PRE);
		writel(TIMER_RELOAD, TIMER_VA_BASE + TIMER_COUNT_LOW);
		writel(0, TIMER_VA_BASE + TIMER_COUNT_HI);

		/* Enable timer interrupt in endless */
		writel(7, TIMER_VA_BASE + TIMER_ENABLE);
		break;
	case CLOCK_EVT_MODE_ONESHOT:
		writel(0, TIMER_VA_BASE + TIMER_ENABLE);
		writel(0, TIMER_VA_BASE + TIMER_COUNT_PRE);
		writel(0, TIMER_VA_BASE + TIMER_COUNT_LOW);
		writel(0, TIMER_VA_BASE + TIMER_COUNT_HI);

		/* Enable timer interrupt, but timer is off */
		writel(2, TIMER_VA_BASE + TIMER_ENABLE);
		break;
	case CLOCK_EVT_MODE_UNUSED:
	case CLOCK_EVT_MODE_SHUTDOWN:
		writel(0, TIMER_VA_BASE + TIMER_ENABLE);
		break;
	default:
		break;
	}
}

static int timer_set_next_event(unsigned long evt,
				struct clock_event_device *unused)
{
	writel(0, TIMER_VA_BASE + TIMER_ENABLE);
	writel(0, TIMER_VA_BASE + TIMER_COUNT_PRE);
	writel(evt, TIMER_VA_BASE + TIMER_COUNT_LOW);
	writel(0, TIMER_VA_BASE + TIMER_COUNT_HI);

	/* Enable timer interrupt in endless */
	writel(3, TIMER_VA_BASE + TIMER_ENABLE);

	return 0;
}

static struct clock_event_device timer0_clockevent =	 {
	.name		= "timer0",
	.shift		= 32,
	.features       = CLOCK_EVT_FEAT_PERIODIC | CLOCK_EVT_FEAT_ONESHOT,
	.set_mode	= timer_set_mode,
	.set_next_event	= timer_set_next_event,
};

/*
 * IRQ handler for the timer
 */
static irqreturn_t mb86h6X_timer_interrupt(int irq, void *dev_id)
{
	struct clock_event_device *evt = &timer0_clockevent;

	/* Clear timer interrupt */
	writel(1, TIMER_VA_BASE + TIMER_IRQ_RESET);

	evt->event_handler(evt);

	return IRQ_HANDLED;
}

static struct irqaction mb86h6X_timer_irq = {
	.name		= "MB86H6X Timer Tick",
	.flags		= IRQF_DISABLED | IRQF_TIMER,
	.handler	= mb86h6X_timer_interrupt,
};

#ifdef CONFIG_MB86H6X_TIMER1_CLOCKSOURCE
#if USE_SCHED_CLOCK
/*
 * This idea was borrowed from plat-omap/common.c
 * Maybe, we should call mb86h6X_clocksource_init in arch_initcall()
 */
/*
 * Kernel assumes that sched_clock can be called early but may not have
 * things ready yet.
 */
static notrace cycle_t mb86h6X_get_cycles_dummy(struct clocksource *cs)
{
	return 0;
}
#endif /* USE_SCHED_CLOCK */

static notrace cycle_t mb86h6X_get_cycles(struct clocksource *cs)
{
	unsigned long time_lo, time_hi, time_hi2;
	unsigned long long cycles;

	time_hi = readl(TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_HI);
	time_lo = readl(TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_LOW);
	time_hi2 = readl(TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_HI);
	if (time_hi2 != time_hi) {
		time_hi = time_hi2;
		time_lo = readl(TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_LOW);
	}
	cycles = ~0;
	cycles -= ((((cycle_t)time_hi)<<32) + time_lo);
	return cycles;
}

static struct clocksource clocksource_mb86h6X = {
	.name 		= "timer1",
 	.rating		= 200,
#if 1 /* USE_SCHED_CLOCK */
	.read		= mb86h6X_get_cycles_dummy,
#else
	.read		= mb86h6X_get_cycles,
#endif /* USE_SCHED_CLOCK */
	.mask		= CLOCKSOURCE_MASK(32),
 	.shift 		= 20,
	.flags		= CLOCK_SOURCE_IS_CONTINUOUS,
};

#if USE_SCHED_CLOCK
/*
 * Returns current time from boot in nsecs. It's OK for this to wrap
 * around for now, as it's just a relative time stamp.
 */
unsigned long long notrace sched_clock(void)
{
	unsigned long long ret;

	ret = (unsigned long long)clocksource_mb86h6X.read(&clocksource_mb86h6X);
	ret = (ret * clocksource_mb86h6X.mult) >> clocksource_mb86h6X.shift;
	return ret;
}
#endif /* USE_SCHED_CLOCK */

static int __init mb86h6X_clocksource_init(void)
{
	writel(0, TIMER_VA_BASE + TIMER1_OFFSET + TIMER_ENABLE);
	writel(0, TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_PRE);
	writel(0xffffffff, TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_LOW);
	writel(0xffffffff, TIMER_VA_BASE + TIMER1_OFFSET + TIMER_COUNT_HI);

	/* Enable timer interrupt in endless */
	writel(7, TIMER_VA_BASE + TIMER1_OFFSET + TIMER_ENABLE);

 	clocksource_mb86h6X.mult =
 		clocksource_khz2mult(81000, clocksource_mb86h6X.shift);
 	clocksource_register(&clocksource_mb86h6X);

 	return 0;
}
#endif /* CONFIG_MB86H6X_TIMER1_CLOCKSOURCE */

/*
 * Set up timer interrupt, and return the current time in seconds.
 */
void __init mb86h6X_timer_init(void)
{
	/* 
	 * Make irqs happen for the system timer
	 */
	setup_irq(IRQ_TIMER0, &mb86h6X_timer_irq);

#ifdef CONFIG_MB86H6X_TIMER1_CLOCKSOURCE
#if USE_SCHED_CLOCK
	clocksource_mb86h6X.read = mb86h6X_get_cycles;
#endif /* USE_SCHED_CLOCK */
	mb86h6X_clocksource_init();
#endif /* CONFIG_MB86H6X_TIMER1_CLOCKSOURCE */

	timer0_clockevent.mult =
		div_sc(81000000, NSEC_PER_SEC, timer0_clockevent.shift);
	timer0_clockevent.max_delta_ns =
		clockevent_delta2ns(0xffffffff, &timer0_clockevent);
	timer0_clockevent.min_delta_ns =
		clockevent_delta2ns(0xf, &timer0_clockevent);

	timer0_clockevent.cpumask = get_cpu_mask(0);
	clockevents_register_device(&timer0_clockevent);
}
