/**
 * watchdog.c:  mb86h61 watchdog
 *
 * Copyright (C) 2011 by FUJITSU SEMICONDUCTOR LIMITED
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 */

#include <linux/init.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/watchdog.h>
#include <linux/module.h>
#include <mach/mb86hxx.h>

//#define MB86H61_WATCHDOG_DEBUG
#define MB86H61_WATCHDOG_REG (*((volatile uint32_t*) (MB86HXX_CLKPWR_VA_BASE + 0x3c)))
#define MB86H61_WATCHDOG_BASE_COUNT_MIN 0x00001000
#define MB86H61_WATCHDOG_BASE_COUNT_MAX 0x0ffff000
#define MB86H61_WATCHDOG_BASE_COUNT MB86H61_WATCHDOG_BASE_COUNT_MIN // use min as start watchdog value
#define MB86H61_WATCHDOG_COUNT_MAX 0x0fffffff
#define MB86H61_WATCHDOG_FREQ (36 * 1000 * 1000)    // 36MHz
#define MB86H61_WATCHDOG_OVERFLOW_JIFFIES(baseCount) ((MB86H61_WATCHDOG_COUNT_MAX - \
                (baseCount)) / MB86H61_WATCHDOG_FREQ * HZ)

static struct timer_list *pFeedDogTimer=NULL;
static spinlock_t  feed_lock;

static inline void setWatchDogCount(unsigned int count) {
    uint32_t value = (MB86H61_WATCHDOG_REG & 0xf0000000) | count;
    MB86H61_WATCHDOG_REG = value;
}

static void feedDog(unsigned long timer) {
    struct timer_list *p_timer = NULL;
    // see system alive times
#ifdef MB86H61_WATCHDOG_DEBUG
    printk(KERN_INFO "jiffies = %lu\n", jiffies);
#endif
    
    spin_lock(&feed_lock);
    p_timer = (struct timer_list *) timer;

    setWatchDogCount(MB86H61_WATCHDOG_BASE_COUNT);
    p_timer->expires = jiffies + MB86H61_WATCHDOG_OVERFLOW_JIFFIES(MB86H61_WATCHDOG_BASE_COUNT) / 2;
    add_timer(p_timer);   // will re-feeddog
#ifdef MB86H61_WATCHDOG_DEBUG
    printk(KERN_INFO "feed watchdog\n");
#endif
    spin_unlock(&feed_lock);

}

static int __init startWatchDog(void) {
    
    setWatchDogCount(MB86H61_WATCHDOG_BASE_COUNT);

#ifdef MB86H61_WATCHDOG_DEBUG
    printk(KERN_INFO "start watchdog, jiffies = %lu\n", jiffies);
#endif

    pFeedDogTimer = (struct timer_list *) kmalloc(sizeof(struct timer_list), GFP_KERNEL);
    if (pFeedDogTimer == NULL) {
        return -1;
    }
    spin_lock_init(&feed_lock);
    init_timer(pFeedDogTimer);
    pFeedDogTimer->expires = jiffies + MB86H61_WATCHDOG_OVERFLOW_JIFFIES(MB86H61_WATCHDOG_BASE_COUNT) / 2;
    pFeedDogTimer->function = feedDog;
    pFeedDogTimer->data = (unsigned long) pFeedDogTimer;
    add_timer(pFeedDogTimer);
    
    return 0;
}

int stop_watchdog_ext(void)
{
    /*
    ****************************************************************************
    ** "pFeedDogTimer" can only exist once and will be destroyed in the following
    ** codes. So this check should be safe by now. After any further modification
    ** on this file, this check should be reviewed to avoid any side effect.
    ****************************************************************************
    */
    if (!pFeedDogTimer){
        return 0;
    }

    /* stop feed timer */
    spin_lock(&feed_lock);

    setWatchDogCount(0);

    del_timer(pFeedDogTimer);
    kfree(pFeedDogTimer);
    pFeedDogTimer = NULL;

#ifdef MB86H61_WATCHDOG_DEBUG
    printk(KERN_INFO "stop watchdog\n");
#endif

    spin_unlock(&feed_lock);
    return 0;
}
EXPORT_SYMBOL(stop_watchdog_ext);

arch_initcall(startWatchDog);
