/*
 *
 * Device driver for MB86h61 remote control interfaces
 * on MB86H61_PDK_LUX_V2 board.
 *
 * Copyright (c) 2010 Fujitsu Semiconductor.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/delay.h>
#include <linux/input.h>
#include <linux/pci.h>
#include <linux/module.h>

#include <media/ir-common.h>

#include "fapi/drv_ir.h"
#include "fapi/drv_ir_nec.h"
#include "fapex_config_cmn.h"
#include "fapex_ir_func.h"
extern FAPEX_IR_FuncStrT ir_funcStr;
/* ---------------------------------------------------------------------- */

struct mb86h61_IR{
	struct input_dev *input;
	struct ir_input_state ir;
	char name[32];
	char phys[32];
        u32 lastKey;
        void* pFAPICtx;
};
static struct mb86h61_IR   *mb86h61_ir;

static void mb86h61_ir_input_key_event(struct input_dev *dev, struct ir_input_state *ir)
{
	if (KEY_RESERVED == ir->keycode) {
		printk(KERN_INFO "%s: unknown key: key=0x%02x raw=0x%02x down=%d\n",
		       dev->name,ir->ir_key,ir->ir_raw,ir->keypressed);
		return;
	}
	input_report_key(dev,ir->keycode,ir->keypressed);
	//input_sync(dev);
}

static void mb86h61_ir_input_keydown(struct input_dev *dev, struct ir_input_state *ir,
		      u32 ir_key, u32 ir_raw)
{
        u32 keycode = IR_KEYCODE(ir->ir_codes, ir_key);
	# if 0
	if (ir->keypressed && ir->keycode != keycode)
	{
		//if (ir->keypressed ) {
		ir->ir_key  = 0;
		ir->ir_raw  = 0;
		ir->keycode = 0;
		ir->keypressed = 0;
	 	//mb86h61_ir_input_key_event(dev,ir);	 	 	
	}
	#	endif
	//if (!ir->keypressed) 
	{
		ir->ir_key  = ir_key;
		ir->ir_raw  = ir_raw;
		ir->keycode = keycode;
		ir->keypressed = 1;	
		mb86h61_ir_input_key_event(dev,ir);
	}

}


/* ---------------------------------------------------------------------- */
void mb86h61_ir_event_Notifier(FAPI_SYS_HandleT handle, uint32_t* optDataPtr)
{
     int32_t  deviceId   = 0;
     uint32_t key        = 0;
     uint32_t rcmCode    = 0;
     uint32_t repeatFlag = 0;
   
     /* Get IR key */
     ir_funcStr.GetKey(handle, &deviceId, &key, &rcmCode, &repeatFlag);
     /* input event */
     if(repeatFlag == 1)
         key = mb86h61_ir->lastKey;
     else
        mb86h61_ir->lastKey = key;
     mb86h61_ir_input_keydown(mb86h61_ir->input, &(mb86h61_ir->ir),key,key);
 }


static int mb86h61_ir_open(struct input_dev *dev)
{
    FAPI_IR_OpenParamsT ir_open_params;
    FAPI_SYS_HandleT appIrHandle;
    int error_code = 0;

    ir_open_params.version           = FAPI_IR_VERSION;
    ir_open_params.swProtocol        = FAPI_IR_NEC_Mode;
    ir_open_params.rcmNotifyFunction = mb86h61_ir_event_Notifier;
    ir_open_params.optDataPtr        = (void*)mb86h61_ir;
    appIrHandle = ir_funcStr.Open( &ir_open_params, &error_code );
    if ( !appIrHandle )
    {
        printk("Open the IR failed\n");
        return error_code;
    }
    if(mb86h61_ir != NULL)
    {
        mb86h61_ir->pFAPICtx = appIrHandle;
    }
    return error_code;
}

static void mb86h61_ir_close(struct input_dev *dev)
{
    if((void*)mb86h61_ir != (void*)dev)
        return;
    
    if(mb86h61_ir != NULL && mb86h61_ir->pFAPICtx != NULL)
       ir_funcStr.Close(mb86h61_ir->pFAPICtx);
}

/* ---------------------------------------------------------------------- */
static int __init  mb86h61_ir_init(void)
{
        int ret = 0;

	struct input_dev *input_dev;
	IR_KEYTAB_TYPE *ir_codes = NULL;
	int ir_type = IR_TYPE_OTHER;
	int err = -ENOMEM;
        char* name="IR-receiver";
        char*  phys="fujitsu mb86h61";

        ret = FAPEXCFG_IR_Init();
        if(ret != 0){
            printk("FAPEXCFG_IR_Init:error %d\n",ret);
            FAPEXCFG_IR_Cleanup();
            return -1;
        }

	mb86h61_ir = kzalloc(sizeof(*mb86h61_ir), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!mb86h61_ir || !input_dev)
		goto err_out_free;

	mb86h61_ir->input = input_dev;

	/* detect & configure */
	ir_codes = ir_codes_mb86h61_nec;

	/* init input device */
	memcpy(mb86h61_ir->name,name,sizeof(name));
        memcpy(mb86h61_ir->phys,phys,sizeof(phys));
	
        ir_input_init(input_dev, &mb86h61_ir->ir, ir_type, ir_codes);
	input_dev->name = mb86h61_ir->name;
	input_dev->phys = mb86h61_ir->phys;
        input_dev->open = mb86h61_ir_open;
        input_dev->close = mb86h61_ir_close;
        input_dev->rep[REP_DELAY] = -1;  //don't use software repeat
        input_dev->rep[REP_PERIOD] = -1;  //don't use software repeat
	/* all done */
	err = input_register_device(mb86h61_ir->input);
	if (err)
		goto err_out_free;

	return 0;

 err_out_free:
	input_free_device(input_dev);
	kfree(mb86h61_ir);
	return err;
}

static void __exit  mb86h61_ir_exit(void)
{
	struct mb86h61_IR *ir = mb86h61_ir;

        FAPEXCFG_IR_Cleanup();

	/* skip detach on non attached boards */
	if (NULL == ir)
		return ;

	input_unregister_device(ir->input);
	kfree(ir);

	/* done */
	return;
}

module_init(mb86h61_ir_init);
module_exit(mb86h61_ir_exit);

/* ---------------------------------------------------------------------- */

MODULE_AUTHOR("Fujitsu Semiconductor ShangHai");
MODULE_DESCRIPTION("input ir driver for mb86h61");
MODULE_LICENSE("GPL");
