/*
 *  drivers/mtd/devices/mb86hxx.c
 *
 * Copyright (C) 2008 FUJITSU MICROELECTRONICS LIMITED
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/mutex.h>
#include <linux/math64.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>

#include <linux/spi/spi.h>
#include <asm/mach/flash.h>

/*get register address*/
#include <mach/mb86hxx.h>
/*include write and read function*/
#include <asm/io.h>

#define FLASH_PAGESIZE    256

/* Flash opcodes. */
#define OPCODE_WREN             0x06  /* Write enable */
#define OPCODE_RDSR             0x05  /* Read status register */
#define OPCODE_WRSR             0x01  /* Write status register 1 byte */
#define OPCODE_NORM_READ        0x03  /* Read data bytes (low frequency) */
#define OPCODE_FAST_READ        0x0b  /* Read data bytes (high frequency) */
#define OPCODE_PP               0x02  /* Page program (up to 256 bytes) */
#define OPCODE_BE_4K            0x20  /* Erase 4KiB block */
#define OPCODE_BE_32K           0x52  /* Erase 32KiB block */
#define OPCODE_CHIP_ERASE       0xc7  /* Erase whole flash chip */
#define OPCODE_SE               0xd8  /* Sector erase (usually 64KiB) */
#define OPCODE_RDID             0x9f  /* Read JEDEC ID */
#define OPCODE_WRDI             0x04

/* Status Register bits. */
#define SR_WIP                  1  /* Write in progress */
#define SR_WEL                  2  /* Write enable latch */
/* meaning of other SR_* bits may differ between vendors */
#define SR_BP0                  4  /* Block protect 0 */
#define SR_BP1                  8  /* Block protect 1 */
#define SR_BP2                  0x10  /* Block protect 2 */
#define SR_SRWD                 0x80  /* SR write protect */

#define SFLASH_PORT             0x0
#define SFLASH_COMMAND          0x4
#define SFLASH_SPEED            0x8

/* Define max times to check status register before we give up. */
#define MAX_READY_WAIT_JIFFIES  (40 * HZ)  /* M25P16 specs 40s max chip erase */
#define CMD_SIZE                4

#ifdef CONFIG_M25PXX_USE_FAST_READ
#define OPCODE_READ          OPCODE_FAST_READ
#define FAST_READ_DUMMY_BYTE 1
#else
#define OPCODE_READ          OPCODE_NORM_READ
#define FAST_READ_DUMMY_BYTE 0
#endif

/****************************************************************************/

struct mb86hxx_sflash{
        struct spi_device       *spi;
        struct mutex            lock;
        struct mtd_info         mtd;
        unsigned                partitioned:1;
        u8                      erase_opcode;
        u8                      command[CMD_SIZE + FAST_READ_DUMMY_BYTE];
};

static inline struct mb86hxx_sflash *mtd_to_mb86hxx_sflash(struct mtd_info *mtd)
{
    return container_of(mtd, struct mb86hxx_sflash, mtd);
}

/****************************************************************************/

/*
 * Internal helper functions
 */

/*
 * Read the status register, returning its value in the location
 * Return the status register value.
 * Returns negative if error occurred.
 */
static int read_sr(struct mb86hxx_sflash *flash)
{
    u8 code = OPCODE_RDSR;
    u32 val1;

    writel(code, MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    val1=readl(  MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    return val1;
}

/*
 * Write status register 1 byte
 * Returns negative if error occurred.
 */
static int write_sr(struct mb86hxx_sflash *flash, u8 val)
{
    return 0;
}

/*
 * Set write enable latch with Write Enable command.
 * Returns negative if error occurred.
 */
static inline int write_enable(struct mb86hxx_sflash *flash)
{
    u8 code = OPCODE_WREN;

    writel(code,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);

    return 0;
}


/*
 * Service routine to read status register until ready, or timeout occurs.
 * Returns non-zero if error.
 */
static int wait_till_ready(struct mb86hxx_sflash *flash)
{
    unsigned long deadline;
    int sr;

    deadline = jiffies + MAX_READY_WAIT_JIFFIES;

    do{
        sr = read_sr(flash);
        if(!(sr&SR_WIP))
        {
             return 0;
        }
        cond_resched();

    } while (!time_after_eq(jiffies, deadline));

    return 1;
}

/*
 * Erase the whole flash memory
 *
 * Returns 0 if successful, non-zero otherwise.
 */
static int erase_chip(struct mb86hxx_sflash *flash)
{
    DEBUG(MTD_DEBUG_LEVEL3, "%s: %s %lldKiB\n",
          dev_name(&flash->spi->dev), __func__,
          (long long)(flash->mtd.size >> 10));
    
    /* Send write enable, then erase commands. */
    write_enable(flash);
    writel(OPCODE_CHIP_ERASE,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    writel(0,MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    if (wait_till_ready(flash))
        return 1;
    writel(OPCODE_WRDI,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    
    return 0;
}

/*
 * Erase one sector of flash memory at offset ``offset'' which is any
 * address within the sector which should be erased.
 *
 * Returns 0 if successful, non-zero otherwise.
 */
static int erase_sector(struct mb86hxx_sflash *flash, u32 offset)
{
    u32 ret=0;
    DEBUG(MTD_DEBUG_LEVEL3, "%s: %s %dKiB at 0x%08x\n",
          dev_name(&flash->spi->dev), __func__,
          flash->mtd.erasesize / 1024, offset);
    
    write_enable(flash);

    writel(OPCODE_SE,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    writel(offset,MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);/*address should be written*/
    if (wait_till_ready(flash))
        ret = 1;
    writel(OPCODE_WRDI,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    return ret;
}

/****************************************************************************/

/*
 * MTD implementation
 */

/*
 * Erase an address range on the flash chip.  The address range may extend
 * one or more erase sectors.  Return an error is there is a problem erasing.
 */
static int sflash_erase(struct mtd_info *mtd, struct erase_info *instr)
{
    struct mb86hxx_sflash *flash = mtd_to_mb86hxx_sflash(mtd);
    u32 addr,len;
    uint32_t rem;

    DEBUG(MTD_DEBUG_LEVEL2, "%s: %s %s 0x%llx, len %lld\n",
          dev_name(&flash->spi->dev), __func__, "at",
          (long long)instr->addr, (long long)instr->len);

    /* sanity checks */
    if (instr->addr + instr->len > flash->mtd.size)
        return -EINVAL;
    div_u64_rem(instr->len, mtd->erasesize, &rem);
    
    if (rem)
        return -EINVAL;

    addr = instr->addr;
    len = instr->len;

    mutex_lock(&flash->lock);

    /* whole-chip erase? */
    if (len == flash->mtd.size) {
        if (erase_chip(flash)) {
            instr->state = MTD_ERASE_FAILED;
            mutex_unlock(&flash->lock);
            return -EIO;
        }

    /* REVISIT in some cases we could speed up erasing large regions
     * by using OPCODE_SE instead of OPCODE_BE_4K.  We may have set up
     * to use "small sector erase", but that's not always optimal.
     */

    /* "sector"-at-a-time erase */
    } else {
        while (len) {
            if (erase_sector(flash, addr)) {
                instr->state = MTD_ERASE_FAILED;
                mutex_unlock(&flash->lock);
                return -EIO;
            }

            addr += mtd->erasesize;
            len -= mtd->erasesize;
        }
    }

    mutex_unlock(&flash->lock);

    instr->state = MTD_ERASE_DONE;
    mtd_erase_callback(instr);

    return 0;
}

/*
 * Read an address range from the flash chip.  The address range
 * may be any size provided it is within the physical boundaries.
 */
static int sflash_read(struct mtd_info *mtd, loff_t from, size_t len,
                       size_t *retlen, u_char *buf)
{
    struct mb86hxx_sflash *flash = mtd_to_mb86hxx_sflash(mtd);
    u32 index,left,data;
    u_char *read_buf = buf;

    /* sanity checks */
    if (!len)
        return 0;

    if (from + len > flash->mtd.size)
        return -EINVAL;
    /* Byte count starts at zero. */
    if (retlen)
        *retlen = 0;

    left = len%sizeof(u32);

    mutex_lock(&flash->lock);
    /* FIXME switch to OPCODE_FAST_READ.  It's required for higher
     * clocks; and at this writing, every chip this driver handles
     * supports that opcode.
     */

     /* Wait till previous write/erase is done. */
    if (wait_till_ready(flash)) {
    /* REVISIT status return?? */
        mutex_unlock(&flash->lock);
        return 1;
    }

    writel(OPCODE_NORM_READ,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    writel((u32)(from),  MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    for( index=0; index < len/4; index++ )
    {
        data = readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
        memcpy(read_buf,&data,sizeof(u32));
        read_buf += sizeof(u32);
        *retlen+=sizeof(u32);
    }

    if(left>0)
    {
        data = 0xffffffff;
        data = readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
        memcpy(read_buf,&data,left);
        *retlen += left;
    }

    mutex_unlock(&flash->lock);

    return 0;
}

/*
 * Write an address range to the flash chip.  Data must be written in
 * FLASH_PAGESIZE chunks.  The address range may be any size provided
 * it is within the physical boundaries.
 */
static int sflash_write(struct mtd_info *mtd, loff_t to, size_t len,
                        size_t *retlen, const u_char *buf)
{
    struct mb86hxx_sflash *flash = mtd_to_mb86hxx_sflash(mtd);
    u32 i,addr,left,data;
    u_char *wptr = NULL;

    wptr =(u_char*) buf;
    addr = to;

    if (retlen)
        *retlen = 0;

    /* sanity checks */
    if (!len)
        return(0);

    if (to + len > flash->mtd.size)
        return -EINVAL;

    mutex_lock(&flash->lock);

    /* Wait until finished previous write command. */
    if (wait_till_ready(flash)) {
        mutex_unlock(&flash->lock);
        return 1;
    }
    /* what page do we start with? */
    left=len%sizeof(u32);
    data=0xffffffff;

    if((addr&0xff)!=0)
    {
        write_enable(flash);
        writel(OPCODE_PP,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
        writel((u32)addr,MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    }
    
    for(i=0;i<len/4;i++)
    {
        if((addr&0xff) ==0)
        {
            write_enable(flash);
            writel(OPCODE_PP,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
            writel((u32)(addr),MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
        }
        memcpy(&data,wptr,sizeof(u32));
        writel(data,MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
        
        (*retlen)+=sizeof(u32);
        addr+=sizeof(u32);
        wptr+=sizeof(u32);
        
        if((addr&0xff)==0)
        {
            wait_till_ready(flash);
        }
        
    }
    if(left>0)
    {
        if((addr&0xff) ==0)
        {
            write_enable(flash);
            writel(OPCODE_PP,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
            writel((u32)(addr),MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
        }
        data=0xffffffff;
        memcpy(&data,wptr,left);
        writel(data,MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
        (*retlen)+=left;

    }

    wait_till_ready(flash);
    writel(OPCODE_WRDI,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);

    writel(OPCODE_RDSR,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);

    mutex_unlock(&flash->lock);

    return 0;
}


/****************************************************************************/

/*
 * SPI device driver setup and teardown
 */

struct flash_info {
    char    *name;

    /* JEDEC id zero means "no ID" (most older chips); otherwise it has
     * a high byte of zero plus three data bytes: the manufacturer id,
     * then a two byte device id.
     */
    u32    jedec_id;
    u16    ext_id;

    /* The size listed here is what works with OPCODE_SE, which isn't
     * necessarily called a "sector" by the vendor.
    */
    unsigned    sector_size;
    u16    n_sectors;

    u16    flags;
#define SECT_4K     0x01    /* OPCODE_BE_4K works uniformly */
};


/* NOTE: double check command sets and memory organization when you add
 * more flash chips.  This current list focusses on newer chips, which
 * have been converging on command sets which including JEDEC ID.
 */
static struct flash_info __devinitdata sflash_data [] = {
    /* Spansion -- single (large) sector size only, at least
     * for the chips listed here (without boot sectors).
     */
    /* name          jedec_id  ext_id  sector_size  n_sector*/
    { "s25sl004a",   0x010212, 0,      64 * 1024,   8,  },
    { "s25sl008a",   0x010213, 0,      64 * 1024,   16, },
    { "s25sl016a",   0x010214, 0,      64 * 1024,   32, },
    { "s25sl032a",   0x010215, 0,      64 * 1024,   64, },
    { "s25sl064a",   0x010216, 0,      64 * 1024,  128, },
    { "s25sl12800",  0x012018, 0x0300, 256 * 1024,  64, },
    { "s25sl12801",  0x012018, 0x0301, 64 * 1024,  256, },
    { "s25fl128p",   0x012018, 0,      256 * 1024,  64, },
    { "mx25l12845e", 0xc22018, 0,      64 * 1024,  256, },
    { "w25q128bvfg", 0xef4018, 0,      64 * 1024,  256, },
    /* this is from drv_sflash_device.c */
};

static struct flash_info *__devinit jedec_probe(struct spi_device *spi)
{
    int    tmp;
    u8    code = OPCODE_RDID;
    u32    jedec;
    u16    ext_jedec;
    struct flash_info   *info;

    /* JEDEC also defines an optional "extended device information"
     * string for after vendor-specific data, after the three bytes
     * we use here.  Supporting some chips might require using it.
    */

    /*24.75mhz*/
    writel(0, MB86HXX_SFLASH_VA_BASE+SFLASH_SPEED);
    /* Read the ID codes */
    writel(code,MB86HXX_SFLASH_VA_BASE+SFLASH_COMMAND);
    jedec = readl(MB86HXX_SFLASH_VA_BASE+SFLASH_PORT);
    ext_jedec=0;
    printk( "%s reading JEDEC ID:  %xh \n",dev_name(&spi->dev), jedec);

    for (tmp = 0, info = sflash_data;tmp < ARRAY_SIZE(sflash_data);tmp++, info++) {
        if (info->jedec_id == (jedec)) {
            if (info->ext_id != 0 && info->ext_id != ext_jedec)
                continue;
            return info;
        }
    }

    dev_err(&spi->dev, "unrecognized JEDEC id %06x\n", jedec);
    return NULL;
}


/*
 * board specific setup should have ensured the SPI clock used here
 * matches what the READ command supports, at least until this driver
 * understands FAST_READ (for clocks over 25 MHz).
 */
static int __devinit mb86hxx_sflash_probe(struct spi_device *spi)
{
    struct flash_platform_data      *data;
    struct mb86hxx_sflash           *flash;
    struct flash_info               *info;
    unsigned int                    i;

    /* Platform data helps sort out which chip type we have, as
     * well as how this board partitions it.  If we don't have
     * a chip ID, try the JEDEC id commands; they'll work for most
     * newer chips, even if we don't recognize the particular chip.
     */
    data = (struct flash_platform_data*)(spi->dev.platform_data);
    if (data && data->name) {
        for (i = 0, info = sflash_data; i < ARRAY_SIZE(sflash_data);i++, info++) {
            if (strcmp(data->name, info->name) == 0)
                break;
        }
        /* unrecognized chip? */
       if (i == ARRAY_SIZE(sflash_data)) {
           DEBUG(MTD_DEBUG_LEVEL0, "%s: unrecognized id %s\n",
                      dev_name(&spi->dev), data->name);
           info = NULL;

           /* recognized; is that chip really what's there? */
        } else if (info->jedec_id) {
            struct flash_info   *chip = jedec_probe(spi);
            if (!chip || chip != info) {
                dev_warn(&spi->dev, "found %s, expected %s\n",
                         chip ? chip->name : "UNKNOWN",info->name);
                info = NULL;
            }
        }
    } else
        info = jedec_probe(spi);

    if (!info)
        return -ENODEV;

    flash = kzalloc(sizeof *flash, GFP_KERNEL);
    if (!flash)
        return -ENOMEM;

    flash->spi = spi;
    mutex_init(&flash->lock);
    dev_set_drvdata(&spi->dev, flash);

    /*
     * Atmel serial flash tend to power up
     * with the software protection bits set
     */

    if (info->jedec_id >> 16 == 0x1f) {
        write_enable(flash);
        write_sr(flash, 0);
    }

    if (data && data->name)
        flash->mtd.name = data->name;
    else
        flash->mtd.name = dev_name(&spi->dev);

    flash->mtd.type = MTD_NORFLASH;
    flash->mtd.writesize = 1;
    flash->mtd.flags = MTD_CAP_NORFLASH;
    flash->mtd.size = info->sector_size * info->n_sectors;
    flash->mtd.erase = sflash_erase;
    flash->mtd.read = sflash_read;
    flash->mtd.write = sflash_write;

    /* prefer "small sector" erase if possible */
    if (info->flags & SECT_4K) {
        flash->erase_opcode = OPCODE_BE_4K;
        flash->mtd.erasesize = 4096;
    } else {
        flash->erase_opcode = OPCODE_SE;
        flash->mtd.erasesize = info->sector_size;
    }

    flash->mtd.dev.parent = &spi->dev;

    dev_info(&spi->dev, "%s (%lld Kbytes)\n", info->name,
            (long long)flash->mtd.size >> 10);

    DEBUG(MTD_DEBUG_LEVEL2,
          "mtd .name = %s, .size = 0x%llx (%lldMiB) "
          ".erasesize = 0x%.8x (%uKiB) .numeraseregions = %d\n",
          flash->mtd.name,
          (long long)flash->mtd.size, (long long)(flash->mtd.size >> 20),
          flash->mtd.erasesize, flash->mtd.erasesize / 1024,
          flash->mtd.numeraseregions);

    if (flash->mtd.numeraseregions)
        for (i = 0; i < flash->mtd.numeraseregions; i++)
            DEBUG(MTD_DEBUG_LEVEL2,
                "mtd.eraseregions[%d] = { .offset = 0x%llx, "
                ".erasesize = 0x%.8x (%uKiB), "
                ".numblocks = %d }\n",
                i, (long long)flash->mtd.eraseregions[i].offset,
                flash->mtd.eraseregions[i].erasesize,
                flash->mtd.eraseregions[i].erasesize / 1024,
                flash->mtd.eraseregions[i].numblocks);


    /* partitions should match sector boundaries; and it may be good to
     * use readonly partitions for writeprotected sectors (BP2..BP0).
     */
    if (mtd_has_partitions()) {
        struct mtd_partition    *parts = NULL;
        int    nr_parts = 0;

        if (mtd_has_cmdlinepart()) {
            static const char *part_probes[]= { "cmdlinepart", NULL, };
            nr_parts = parse_mtd_partitions(&flash->mtd,part_probes, &parts, 0);
        }

        if (nr_parts <= 0 && data && data->parts) {
            parts = data->parts;
            nr_parts = data->nr_parts;
        }
        if (nr_parts > 0) {
            for (i = 0; i < nr_parts; i++) {
                DEBUG(MTD_DEBUG_LEVEL2, "partitions[%d] = "
                      "{.name = %s, .offset = 0x%llx, "
                      ".size = 0x%llx (%lldKiB) }\n",
                       i, parts[i].name,
                       (long long)parts[i].offset,
                       (long long)parts[i].size,
                       (long long)(parts[i].size >> 10));
            }
            flash->partitioned = 1;
            return add_mtd_partitions(&flash->mtd, parts, nr_parts);
        }
    } else if (data && data->nr_parts)
        dev_warn(&spi->dev, "ignoring %d default partitions on %s\n",
                 data->nr_parts, data->name);
    return add_mtd_device(&flash->mtd) == 1 ? -ENODEV : 0;
}


static int __devexit mb86hxx_sflash_remove(struct spi_device *spi)
{
    struct mb86hxx_sflash *flash = dev_get_drvdata(&spi->dev);
    int    status;

    /* Clean up MTD stuff. */
    if (mtd_has_partitions() && flash->partitioned)
        status = del_mtd_partitions(&flash->mtd);
    else
        status = del_mtd_device(&flash->mtd);
    if (status == 0)
        kfree(flash);
    return 0;
}


static struct spi_driver mb86hxx_sflash_driver = {
    .driver = {
        .name = "mb86hxx_sflash",
        .bus = &spi_bus_type,
        .owner = THIS_MODULE,
    },
    .probe = mb86hxx_sflash_probe,
    .remove = __devexit_p(mb86hxx_sflash_remove),
    /* REVISIT: many of these chips have deep power-down modes, which
     * should clearly be entered on suspend() to minimize power use.
     * And also when they're otherwise idle...
     */
};

/*
    GPIO         name     func
    GPIO_43      SDIN_SF   2
    GPIO_44      HOLD_SF   49
    GPIO_45      CS_SF     50
    GPIO_46      CLK_SF    51
    GPIO_47      SDO_SF    52
*/

static void mb86hxx_sflash_gpio_conf(void)
{
    writel(2,  MB86HXX_GPIO_VA_BASE+0x200+4*(43+2)); /*configure GPIO 43 as SDIN_SF*/
    writel(49, MB86HXX_GPIO_VA_BASE+4*44);           /*configure GPIO 44 as HOLD_SF*/
    writel(50, MB86HXX_GPIO_VA_BASE+4*45);           /*configure GPIO 45 as CS_SF*/
    writel(51, MB86HXX_GPIO_VA_BASE+4*46);           /*configure GPIO 46 as CLK_SF*/
    writel(52, MB86HXX_GPIO_VA_BASE+4*47);           /*configure GPIO 47 as SDO_SF*/

}

static int mb86hxx_sflash_init(void)
{
    mb86hxx_sflash_gpio_conf();
    return spi_register_driver(&mb86hxx_sflash_driver);
}

static void mb86hxx_sflash_exit(void)
{
    spi_unregister_driver(&mb86hxx_sflash_driver);
}

module_init(mb86hxx_sflash_init);
module_exit(mb86hxx_sflash_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Fujitsu FMC");
MODULE_DESCRIPTION("MTD serial flash driver for MTD");
