/*
 *  drivers/mtd/maps/mb86hxx.c
 *
 * Copyright (C) 2008 FUJITSU MICROELECTRONICS LIMITED
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/platform_device.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/slab.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>
#include <mach/hardware.h>
#include <asm/mach/flash.h>
#include <mach/mb86hxx.h>

/* UPI Device Config Register Base */
#define MB86H60_UPI0_CONFIG_BASE	(MB86HXX_UPI_VA_BASE + 0x18000000)
#define MB86H60_UPI1_CONFIG_BASE	(MB86HXX_UPI_VA_BASE + 0x18000040)
#define MB86H60_UPI2_CONFIG_BASE	(MB86HXX_UPI_VA_BASE + 0x18000080)

/* UPI Device Config Register Offset */
#define MB86H60_UPI_MODE_OFFSET		0x00
#define MB86H60_UPI_TALS_OFFSET		0x04
#define MB86H60_UPI_TALH_OFFSET		0x08
#define MB86H60_UPI_TRS_OFFSET		0x0C
#define MB86H60_UPI_TRDS_OFFSET		0x10
#define MB86H60_UPI_TRDH_OFFSET		0x14
#define MB86H60_UPI_TRR_OFFSET		0x18
#define MB86H60_UPI_TWS_OFFSET		0x1C
#define MB86H60_UPI_TWP_OFFSET		0x20
#define MB86H60_UPI_TWR_OFFSET		0x24
#define MB86H60_UPI_WMMASK_OFFSET	0x28
#define MB86H60_UPI_BUFEN_OFFSET	0x2C

#define MB86H60_UPI_ADDREXT	(MB86HXX_UPI_VA_BASE + 0x180000C0)

/* UPI_ADDREXT Bit Flags */
#define ADDREXT_EXT_MODE_0_BIT	(1 << 0)
#define ADDREXT_EXT_MODE_1_BIT	(1 << 1)
#define ADDREXT_EXT_MODE_2_BIT	(1 << 2)
#define ADDREXT_LIN_MODE_0_BIT	(1 << 3)
#define ADDREXT_LIN_MODE_1_BIT	(1 << 4)
#define ADDREXT_LIN_MODE_2_BIT	(1 << 5)

#define MB86H60_UPI_CSSEL	(MB86HXX_UPI_VA_BASE + 0x180000C4)
#define MB86H60_UPI_CSMODE	(MB86HXX_UPI_VA_BASE + 0x180000C8)

/* UPI_CSMODE Bit Flags*/
#define CSMODE_CS1_INVERT_BIT	(1 << 0)
#define CSMODE_CS2_INVERT_BIT	(1 << 1)
#define CSMODE_CS1_WEMODE_BIT	(1 << 2)
#define CSMODE_CS2_WEMODE_BIT	(1 << 3)

#define MB86H60_UPI_WAIT_TIME		(MB86HXX_UPI_VA_BASE + 0x18000030)
#define MB86H60_UPI_WAIT_TIME_EN	(MB86HXX_UPI_VA_BASE + 0x18000034)
#define MB86H60_UPI_LAST_WAIT_TIME	(MB86HXX_UPI_VA_BASE + 0x18000038)
#define MB86H60_UPI_WAIT_TIME_STATUS	(MB86HXX_UPI_VA_BASE + 0x1800003C)

/* UPI_WAIT_TIME_STATUS Bit Flags */
#define WTSTS_WAIT_TIME_ACT_BIT		(1 << 0)
#define WTSTS_WAIT_LINE_STATUS_BIT	(1 << 1)

#define UPI_MODE_PARAM		0x01
#define UPI_TALS_PARAM		0x02
#define UPI_TALH_PARAM		0x02
#define UPI_TRS_PARAM		0x02
#define UPI_TRDS_PARAM		0x13
#define UPI_TRDH_PARAM		0x02
#define UPI_TRR_PARAM		0x05
#define UPI_TWS_PARAM		0x05
#define UPI_TWP_PARAM		0x09
#define UPI_TWR_PARAM		0x02
#define UPI_WMMASK_PARAM	0x01
#define UPI_BUFEN_PARAM		0x00

#define UPI_ADDREXT_PARAM		0x00
#define UPI_CSSEL_PARAM			0x00
#define UPI_CSMODE_PARAM		0x00
#define UPI_WAIT_TIME_PARAM		0x00
#define UPI_WAIT_TIME_EN_PARAM		0x00
#define UPI_LAST_WAIT_TIME_PARAM	0x00
#define UPI_WAIT_TIME_STATUS_PARAM	0x03

#ifdef CONFIG_MB86HXX_UPI_INIT
static inline void mb86hxx_upi_init(void)
{
#ifdef CONFIG_MB86HXX_UPI0
	unsigned long base = MB86H60_UPI0_CONFIG_BASE;
#endif
#ifdef CONFIG_MB86HXX_UPI1
	unsigned long base = MB86H60_UPI1_CONFIG_BASE;
#endif
#ifdef CONFIG_MB86HXX_UPI2
	unsigned long base = MB86H60_UPI2_CONFIG_BASE;
#endif

	/* initialize UPI */
	writel(UPI_MODE_PARAM, base + MB86H60_UPI_MODE_OFFSET);
	writel(UPI_TALS_PARAM, base + MB86H60_UPI_TALS_OFFSET);
	writel(UPI_TALH_PARAM, base + MB86H60_UPI_TALH_OFFSET);
	writel(UPI_TRS_PARAM, base + MB86H60_UPI_TRS_OFFSET);
	writel(UPI_TRDS_PARAM, base + MB86H60_UPI_TRDS_OFFSET);
	writel(UPI_TRDH_PARAM, base + MB86H60_UPI_TRDH_OFFSET);
	writel(UPI_TRR_PARAM, base + MB86H60_UPI_TRR_OFFSET);
	writel(UPI_TWS_PARAM, base + MB86H60_UPI_TWS_OFFSET);
	writel(UPI_TWP_PARAM, base + MB86H60_UPI_TWP_OFFSET);
	writel(UPI_TWR_PARAM, base + MB86H60_UPI_TWR_OFFSET);
	writel(UPI_WMMASK_PARAM, base + MB86H60_UPI_WMMASK_OFFSET);
	writel(UPI_BUFEN_PARAM, base + MB86H60_UPI_BUFEN_OFFSET);

	writel(UPI_ADDREXT_PARAM, MB86H60_UPI_ADDREXT);
	writel(UPI_CSSEL_PARAM, MB86H60_UPI_CSSEL);
	writel(UPI_CSMODE_PARAM, MB86H60_UPI_CSMODE);

	writel(UPI_WAIT_TIME_PARAM, MB86H60_UPI_WAIT_TIME);
	writel(UPI_WAIT_TIME_EN_PARAM, MB86H60_UPI_WAIT_TIME_EN);
	writel(UPI_LAST_WAIT_TIME_PARAM, MB86H60_UPI_LAST_WAIT_TIME);
	writel(UPI_WAIT_TIME_STATUS_PARAM, MB86H60_UPI_WAIT_TIME_STATUS);
}
#else
static inline void mb86hxx_upi_init(void)
{
}
#endif

static struct mtd_partition mb86hxx_partitions[] = {
	{
		.name = "rootfs",
		.size = CONFIG_MTD_MB86HXX_PARTITION_SIZE,
		.offset = CONFIG_MTD_MB86HXX_PARTITION_OFFSET,
		.mask_flags = 0,
	}
};

struct mb86hxx_info {
	struct mtd_info *mtd;
	struct map_info map;
};

static int __devinit mb86hxx_pflash_probe(struct platform_device *pdev)
{
	int err;
	struct mb86hxx_info *info;
	struct flash_platform_data *pdata = pdev->dev.platform_data;
	struct resource *res = pdev->resource;
	unsigned long size = res->end - res->start + 1;

	info = kzalloc(sizeof(struct mb86hxx_info), GFP_KERNEL);
	if (!info) {
		return -ENOMEM;
	}

	if (!request_mem_region(res->start, size, "pflash")) {
		err = -EBUSY;
		goto out_free_info;
	}

        info->map.virt = (void *)MB86HXX_UPI_VA_BASE; 

	if (!info->map.virt) {
		err = -ENOMEM;
		goto out_release_mem_region;
	}
	info->map.name = "mb86hxx_pflash";
	info->map.phys = res->start;
	info->map.size = size;
	info->map.bankwidth = pdata->width;

	simple_map_init(&info->map);
	info->mtd = do_map_probe(pdata->map_name, &info->map);
	if (!info->mtd) {
		err = -EIO;
		goto out_release_mem_region;
	}
	info->mtd->owner = THIS_MODULE;

	err = add_mtd_partitions(info->mtd, mb86hxx_partitions,
				 ARRAY_SIZE(mb86hxx_partitions));
	if (err != 0) {
		goto out_map_destroy;
	}

	platform_set_drvdata(pdev, info);

	return 0;

 out_map_destroy:
	map_destroy(info->mtd);

 out_release_mem_region:
	release_mem_region(res->start, size);
 out_free_info:
	kfree(info);

	return err;
}

static int __devexit mb86hxx_pflash_remove(struct platform_device *pdev)
{
	struct mb86hxx_info *info = platform_get_drvdata(pdev);

	platform_set_drvdata(pdev, NULL);

	if (info) {
		del_mtd_partitions(info->mtd);
		map_destroy(info->mtd);
		iounmap((void __iomem *)info->map.virt);
		release_mem_region(info->map.phys, info->map.size);
		kfree(info);
	}

	return 0;
}

static struct platform_driver mb86hxx_pflash_driver = {
	.probe = mb86hxx_pflash_probe,
	.remove = __devexit_p(mb86hxx_pflash_remove),
	.driver = {
		.name = "mb86hxx-pflash",
	},
};

static int __init mb86hxx_init(void)
{
	mb86hxx_upi_init();
	return platform_driver_register(&mb86hxx_pflash_driver);
}

static void __exit mb86hxx_exit(void)
{
	platform_driver_unregister(&mb86hxx_pflash_driver);
}

module_init(mb86hxx_init);
module_exit(mb86hxx_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MTD map driver for MB86HXX");
