/*
 *  drivers/mtd/mb86h61_nand.c
 *
 *  Copyright (c) 2011 Fujitsu Semiconductor ShangHai 
 *
 *  Copyright (c) 2002 Thomas Gleixner <tgxl@linutronix.de>
 *
 *  Derived from drivers/mtd/spia.c
 *	 Copyright (C) 2000 Steven J. Hill (sjhill@realitydiluted.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   FUJITSU MB86H61_PDK_LUX_V2 board
 *
 *    2011-2-17 Initial version
 */

#include <linux/slab.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <mach/hardware.h>
#include <asm/sizes.h>
#include <mach/mb86hxx.h>

/* UPI Device Config Register Base */
#define MB86H61_UPI0_CONFIG_BASE        (MB86HXX_UPI_VA_BASE + 0x18000000)
#define MB86H61_UPI1_CONFIG_BASE        (MB86HXX_UPI_VA_BASE + 0x18000040)
#define MB86H61_UPI2_CONFIG_BASE        (MB86HXX_UPI_VA_BASE + 0x18000080)

#ifdef CONFIG_MB86H61_NAND_UPI0
   #define MB86H61_NAND_BASE MB86HXX_UPI_VA_BASE
#endif
#ifdef CONFIG_MB86H61_NAND_UPI1
    #define MB86H61_NAND_BASE (MB86HXX_UPI_VA_BASE + 0x08000000)
#endif
#ifdef CONFIG_MB86H61_NAND_UPI2
    #define MB86H61_NAND_BASE (MB86HXX_UPI_VA_BASE + 0x10000000)
#endif

#define MB86H61_NAND_CMD_OFFSET     0x0
#define MB86H61_NAND_ADDR_OFFSET    0x4
#define MB86H61_NAND_DATA_OFFSET    0x8
#define MB86H61_NAND_END_OFFSET     0xc


/* UPI Device Config Register Offset */
#define MB86H61_UPI_MODE_OFFSET         0x00
#define MB86H61_UPI_TALS_OFFSET         0x04
#define MB86H61_UPI_TALH_OFFSET         0x08
#define MB86H61_UPI_TRS_OFFSET          0x0C
#define MB86H61_UPI_TRDS_OFFSET         0x10
#define MB86H61_UPI_TRDH_OFFSET         0x14
#define MB86H61_UPI_TRR_OFFSET          0x18
#define MB86H61_UPI_TWS_OFFSET          0x1C
#define MB86H61_UPI_TWP_OFFSET          0x20
#define MB86H61_UPI_TWR_OFFSET          0x24
#define MB86H61_UPI_WMMASK_OFFSET       0x28
#define MB86H61_UPI_BUFEN_OFFSET        0x2C

#define MB86H61_UPI_ADDREXT     (MB86HXX_UPI_VA_BASE + 0x180000C0)
#define MB86H61_UPI_CSSEL       (MB86HXX_UPI_VA_BASE + 0x180000C4)
#define MB86H61_UPI_CSMODE      (MB86HXX_UPI_VA_BASE + 0x180000C8)

#define MB86H61_UPI_WAIT_TIME           (MB86HXX_UPI_VA_BASE + 0x18000030)
#define MB86H61_UPI_WAIT_TIME_EN        (MB86HXX_UPI_VA_BASE + 0x18000034)
#define MB86H61_UPI_LAST_WAIT_TIME      (MB86HXX_UPI_VA_BASE + 0x18000038)
#define MB86H61_UPI_WAIT_TIME_STATUS    (MB86HXX_UPI_VA_BASE + 0x1800003C)


#define UPI_MODE_PARAM          0x03
#define UPI_TALS_PARAM          0x0
#define UPI_TALH_PARAM          0x0
#define UPI_TRS_PARAM           0x0
#define UPI_TRDS_PARAM          0x06
#define UPI_TRDH_PARAM          0x0
#define UPI_TRR_PARAM           0x0
#define UPI_TWS_PARAM           0x0
#define UPI_TWP_PARAM           0x06
#define UPI_TWR_PARAM           0x0
#define UPI_WMMASK_PARAM        0x01
#define UPI_BUFEN_PARAM         0x0

#define UPI_ADDREXT_PARAM               0x00
#define UPI_CSSEL_PARAM                 0x00
#define UPI_CSMODE_PARAM                0x00
#define UPI_WAIT_TIME_PARAM             0x00
#define UPI_WAIT_TIME_EN_PARAM          0x00

#ifdef CONFIG_MB86H61_UPI_NAND_INIT
static inline void mb86h61_upi_nand_init(void)
{
#ifdef CONFIG_MB86H61_NAND_UPI0
        unsigned long base = MB86H61_UPI0_CONFIG_BASE;
#endif
#ifdef CONFIG_MB86H61_NAND_UPI1
        unsigned long base = MB86H61_UPI1_CONFIG_BASE;
#endif
#ifdef CONFIG_MB86H61_NAND_UPI2
        unsigned long base = MB86H61_UPI2_CONFIG_BASE;
#endif
        writel(134, MB86HXX_GPIO_VA_BASE+4*6	); /*configure GPIO 6 as CLE*/
        writel(1, MB86HXX_GPIO_VA_BASE+4*66	);     /*configure GPIO 66 as CS */
        writel(1, MB86HXX_GPIO_VA_BASE+4*68     ); /*configure GPIO 68 as CS */

        /* initialize UPI */
        writel(UPI_MODE_PARAM, 		base + 	MB86H61_UPI_MODE_OFFSET);
        writel(UPI_TALS_PARAM, 		base + 	MB86H61_UPI_TALS_OFFSET);
        writel(UPI_TALH_PARAM, 		base + 	MB86H61_UPI_TALH_OFFSET);
        writel(UPI_TRS_PARAM, 		base + 	MB86H61_UPI_TRS_OFFSET);
        writel(UPI_TRDS_PARAM,		base + 	MB86H61_UPI_TRDS_OFFSET);
        writel(UPI_TRDH_PARAM, 		base + 	MB86H61_UPI_TRDH_OFFSET);
        writel(UPI_TRR_PARAM, 		base + 	MB86H61_UPI_TRR_OFFSET);
        writel(UPI_TWS_PARAM, 		base + 	MB86H61_UPI_TWS_OFFSET);
        writel(UPI_TWP_PARAM, 		base + 	MB86H61_UPI_TWP_OFFSET);
        writel(UPI_TWR_PARAM, 		base + 	MB86H61_UPI_TWR_OFFSET);
        writel(UPI_WMMASK_PARAM, 	base + 	MB86H61_UPI_WMMASK_OFFSET);
        writel(UPI_BUFEN_PARAM, 	base + 	MB86H61_UPI_BUFEN_OFFSET);

        writel(UPI_ADDREXT_PARAM, 	MB86H61_UPI_ADDREXT);
        writel(UPI_CSSEL_PARAM, 	MB86H61_UPI_CSSEL);
        writel(UPI_CSMODE_PARAM, 	MB86H61_UPI_CSMODE);

        writel(UPI_WAIT_TIME_PARAM, 	MB86H61_UPI_WAIT_TIME);
        writel(UPI_WAIT_TIME_EN_PARAM, 	MB86H61_UPI_WAIT_TIME_EN);
        
}
#else
{
}
#endif

static struct mtd_partition mb86h61_partitions[] = {
        {
                .name = "rootfs",
                .size = CONFIG_MTD_NAND_MB86H61_PARTITION_SIZE,
                .offset = CONFIG_MTD_NAND_MB86H61_PARTITION_OFFSET,
                .mask_flags = 0,
        }
};


/*
 * MTD structure for MB98H60 board
 */
static struct mtd_info *mb86h61_mtd = NULL;
static void __iomem *mb86h61_nand_base;


#define NUM_PARTITIONS 1

/*	hardware specific access to control-lines
 *
 */
static void mb86h61_hwcontrol(struct mtd_info *mtd, int cmd,
			       unsigned int ctrl)
{
    struct nand_chip *chip = mtd->priv;
 		
    if(ctrl & NAND_CTRL_CHANGE)
    {
        if((ctrl & NAND_CTRL_CLE) == NAND_CTRL_CLE)
            chip->IO_ADDR_W = mb86h61_nand_base + MB86H61_NAND_CMD_OFFSET;
        else if((ctrl & NAND_CTRL_ALE) == NAND_CTRL_ALE)
            chip->IO_ADDR_W = mb86h61_nand_base + MB86H61_NAND_ADDR_OFFSET;
        else
            chip->IO_ADDR_W = mb86h61_nand_base + MB86H61_NAND_DATA_OFFSET;
    }

    if (cmd != NAND_CMD_NONE)
        writeb(cmd, chip->IO_ADDR_W);	
}

static int mb86h61_nand_device_ready(struct mtd_info *mtd)
{
    unsigned int gpio;
    gpio = *((volatile unsigned int*)(MB86HXX_GPIO_VA_BASE+0x504));
    return (gpio>>8)&0x1;
}

static void mb86h61_select_chip(struct mtd_info *mtd,int chip_nr)
{
    int i;
    if(chip_nr == -1)
    {
        writel(1, MB86HXX_GPIO_VA_BASE+4*66	);   /*CS disable*/ 
    }
    else
    {
       	writel(0, MB86HXX_GPIO_VA_BASE+4*66	);   /*CS enable*/
    }
    for(i=0;i<2000;i++);
}


/*
 * Main initialization routine
 */
static int __init mb86h61_nand_init(void)
{
	struct nand_chip *this;
	int err = 0;

	mb86h61_upi_nand_init();
	/* Allocate memory for MTD device structure and private data */
	mb86h61_mtd = kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip),
			       GFP_KERNEL);
	if (!mb86h61_mtd) {
		printk("Unable to allocate MB86H61 NAND MTD device structure.\n");
		err = -ENOMEM;
		goto out;
	}

	/* map physical address */
	mb86h61_nand_base = (void __iomem *)MB86H61_NAND_BASE;
	if (!mb86h61_nand_base) {
		printk("Ioremap MB86H61 Nand Device failed\n");
		err = -EIO;
		goto out_mtd;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *)(&mb86h61_mtd[1]);

	/* Initialize structures */
	memset(mb86h61_mtd, 0, sizeof(struct mtd_info));
	memset(this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	mb86h61_mtd->priv = this;
	mb86h61_mtd->owner = THIS_MODULE;

	/* Set address of NAND IO lines */
	this->IO_ADDR_R = mb86h61_nand_base + MB86H61_NAND_DATA_OFFSET;
	this->IO_ADDR_W = mb86h61_nand_base + MB86H61_NAND_DATA_OFFSET;
	this->cmd_ctrl = mb86h61_hwcontrol;
	this->select_chip = mb86h61_select_chip;
	this->dev_ready = mb86h61_nand_device_ready;
	/* 50 us command delay time */
	this->chip_delay = 50;
	this->ecc.mode = NAND_ECC_SOFT;

	/* Enable the following for a flash based bad block table */
	/*
	   this->options = NAND_USE_FLASH_BBT;
	 */
	
        /* Scan to find existance of the device */
	if (nand_scan(mb86h61_mtd, 1)) {
		err = -ENXIO;
		goto out_mtd;
	}
  	err = add_mtd_partitions(mb86h61_mtd,mb86h61_partitions,NUM_PARTITIONS);
  
  	if(err == 0)
		return 0;

out_mtd:
	kfree(mb86h61_mtd);
out:
	return err;
}

module_init(mb86h61_nand_init);

/*
 * Clean up routine
 */
static void __exit mb86h61_nand_cleanup(void)
{
	/* Release resources, unregister device */
	nand_release(mb86h61_mtd);

	/* Free the MTD device structure */
	kfree(mb86h61_mtd);
}

module_exit(mb86h61_nand_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Fujitsu Semiconductor ShangHai");
MODULE_DESCRIPTION("Glue layer for MB86H61 Nand MTD device");
