/*
 *  linux/drivers/net/fmac3h.c
 *
 * Copyright (C) 2007-2009 FUJITSU MICROELECTRONICS LIMITED
 * Copyright (C) 2010-2011 FUJITSU SEMICONDUCTOR LIMITED
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
#define DRV_NAME       "F_MAC3H"
#define DRV_VERSION    "1.1.1"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/ethtool.h>
#include <linux/compiler.h>
#include <linux/rtnetlink.h>
#include <linux/crc32.h>
#include <linux/mii.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/skbuff.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <mach/hardware.h>

#include "fmac3h.h"

/* The driver uses a "RX_COPYBREAK" scheme. Rather than a fixed intermediate 
 * receive buffer, this scheme allocates full-sized skbuffs as receive buffers.
 * The value RX_COPYBREAK is used as the copying breakpoint: it is chosen to 
 * trade-off the memory wasted by passing the full-sized skbuff to the queue 
 * layer for all frames vs. the copying cost of copying a frame to a 
 * correctly-sized skbuff.
 *
 * ARM systems perform better by disregarding the bus-master transfer capability
 *  of these cards.
 *
 * the explanation of RX_COPYBREAK stolen from 3c59x.c */

/*
 * The definition used in this driver
 */

/* MAX size of TX&RX size */
/* frame(1500)+EthernetHeader(14)+FCS(4)+NET_IP_ALIGN(2) */
#define BUFFER_SIZE      1520
#define JUMBO_FRAME_MTU  9000  /* Jumbo Frame */

/* MAX address of 32 PHYs */
#define MAX_PHY_ADR  31

/* After reset device, have to wait 2000ns before access the register */
#define DELAY_TIME 2      /* 2us = 2000ns */

/* hope the option will end within this time */
#define WAIT_TIME      1000
#define ANEG_WAIT_TIME (WAIT_TIME*1000)

/* check carrier every 1s */
#define CHECK_CARRIER_TIME  (jiffies + HZ)

/* the number of free descriptors */
#define TX_RING_AVAIL(RP) \
        (((RP)->tx_tail <= (RP)->tx_head) ? \
         (RP)->tx_tail + (FMAC3H_RDESC_NUM - 1) - (RP)->tx_head : \
         (RP)->tx_tail - (RP)->tx_head - 1)

#define NEXT_TX(N)    (((N) + 1) % FMAC3H_TDESC_NUM)
#define NEXT_RX(N)    (((N) + 1) % FMAC3H_RDESC_NUM)

/* the size of descriptor spaces */
#define FMAC3H_DESC_BYTES  ((sizeof(struct fmac3h_desc) * FMAC3H_TDESC_NUM)) + \
                           ((sizeof(struct fmac3h_desc) * FMAC3H_RDESC_NUM))

/* Read MMC register */
#define READ_MMC(REG)  (fmac3h_reg_read(lp->base_addr, (REG)))

/* Dump control */
#define REG_DUMP  0x01
#define DESC_DUMP 0x02
#define MEM_DUMP  0x04
#define WORD0(val) ((val) & 0x000000FF)
#define WORD1(val) (((val) & 0x0000FF00) >> 8)
#define WORD2(val) (((val) & 0x00FF0000) >> 16)
#define WORD3(val) (((val) & 0xFF000000) >> 24)
#define READ_REG(REG) __raw_readl(base_addr + (REG))

/* Send engin error number */
#define FMAC3H_TX_OK    NETDEV_TX_OK    /* driver took care of packet */
#define FMAC3H_TX_BUSY  NETDEV_TX_BUSY  /* driver tx path was busy*/

/* SYS_CLK is based on AHB */
#define SYS_CLK ETH_MHZ

extern int eth_mac_addr(struct net_device *dev, void *p);

/* suggested PHY type's ID */
#define BCM5241_ID    0x0143BC31
#define LAN8700_ID    0x0007C0C4
#define LAN8710_ID    0x0007c0f1
#if 1 /* Add for #28840 Realtek RTL8201 に対応 by hiroshi.hasegawa@pixela.jp 2011.09.10 */
#define RTL8201_ID    0x001cc815 /* 上位16bitはRegister2, 下位16bitはRegister3 の値 */
#endif

/* 
 * if you have tested this driver on a new type PHY and make sure the driver
 * supports the new type PHY, you should add the ID(get from register ID1&ID2) 
 * to this array.
 */
static u32 known_phy_id[] = {
    BCM5241_ID,
    LAN8700_ID,
    LAN8710_ID,
#if 1 /* Add for #28840 Realtek RTL8201 に対応 by hiroshi.hasegawa@pixela.jp 2011.09.10 */
    RTL8201_ID,
#endif
    /* Add the ID above here
     * 0 should always be the last one of array */
    0   
};

/* Set MAC address */
#ifndef CONFIG_ARCH_MB86HXX
#define FMAC3H_MAC_ADDR
/* MAC address register base */
#define MACADDR_BASE MB86HXX_SRAM_BASE
/* MAC address register offset */
#define MACL        0x10
#define MACH        0x14
#endif /* Set MAC address */

#if 1 /* Add for #28357 起動引数からMACアドレスを取得 by hiroshi.hasegawa@pixela.jp 2011.09.08
       * (Linuxの起動引数(console=ttyAMA0,115200 root=/dev/mtdblock6 rw ip=none です)に
       *  mac_adrs=00:11:22:33:44:55 を追加すれば 有線LANのMACアドレスをそのアドレスにする)
       */
static char *cmdline;
static void macadrs_setup_from_cmdline(u8 *addr)
{
	int cnt;
	u8 num;
	u8 *p_addr = addr;
	char *str = cmdline;
	
	/* 起動引数にセットされたMACアドレスの文字列 xx:xx:xx:xx:xx:xx をデコード */
	if (!str) {
		cnt = 0;
	} else {
		for (cnt = 0; cnt < 18; cnt++) {
			if ( cnt == 17 ) {
				if (*str != '\0') {
					break; /* 18文字目が終端文字でなければ break */
				}
			} else if ( (cnt % 3) == 2 ) {
				if (*str != ':') {
					break; /* 3,6,9,12,15文字目が':'文字でなければ break */
				}
			} else {
				if ('0' <= *str && *str <= '9') {
					num = *str - '0';
				} else if ('a' <= *str && *str <= 'f') {
					num = *str - 'a' + 0xa;
				} else if ('A' <= *str && *str <= 'F') {
					num = *str - 'A' + 0xA;
				} else {
					break; /* 16進数ではない文字があれば break */
				}
				if ( (cnt % 3) == 0 ) {
					*p_addr = num << 4;
				} else {
					*p_addr |= num;
					p_addr++;
				}
			}
			str++; /* 次の文字へ */
		}
	}
	
	/* 正しいMACアドレスが設定されていなければ デフォルト値へ */
	if (cnt != 18) {
		addr[0] = 0xca; addr[1] = 0xe7; addr[2] = 0x96; addr[3]=0x99;
		addr[4] = 0x6b; addr[5] = 0xef; // refs #29889 デフォルトのMACアドレスを AirTuner-VEは
		                                // ca:e7:96:99:6b:ef (DT-270は最下位桁が ee) にする
		PMSG(KERN_INFO, "Use default MAC address: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n", 
		     addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
	} else {
		PMSG(KERN_INFO, "MAC address: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n", 
		     addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
	}
}
#endif

#ifdef FMAC3H_DEBUG
/*
 * debug_dump -- dump data in debug mode
 * @base_addr: the base address of FMAC3H's register space
 * @ctl: dump message control
 *       bit0:dump FMAC3H register
 *       bit1:dump TX/RX descriptors
 *       bit2:dump memory from *startp to (*startp + size)
 * @txring: TX descriptor ring start address
 * @rxring: RX descriptor ring start address
 * @startp: the start address of dump memory.
 * @size: the size of dump meory.
 *
 * Description: This function will be effective when FMAC3H_DEBUG define
 *  is enable.
 */
static void debug_dump(void __iomem *base_addr, int ctl, int *txring,
                       int *rxring, char *startp, unsigned int size)
{
    int *p;
    int i;
    char *mp;

    /* 
     * Dump Register
     */  
    if (ctl & REG_DUMP) { 
        printk(TEMP_DEBUG DUMPH 
           "+-------------------------------------------------------------+\n");
        printk(TEMP_DEBUG DUMPH 
           "|                     FMAC3H Register Dump                    |\n");
        printk(TEMP_DEBUG DUMPH 
           "+-------------------------------------------------------------+\n");
        printk(TEMP_DEBUG DUMPH 
               "MCR(0x%04x)    : %08x     |     MFFR(0x%04x)   : %08x\n", 
               FMAC3H_MCR, READ_REG(FMAC3H_MCR), 
               FMAC3H_MFFR, READ_REG(FMAC3H_MFFR));
        printk(TEMP_DEBUG DUMPH 
               "MHTRH(0x%04x)  : %08x     |     MHTRL(0x%04x)  : %08x\n", 
               FMAC3H_MHTRH, READ_REG(FMAC3H_MHTRH), 
               FMAC3H_MHTRL, READ_REG(FMAC3H_MHTRL));
        printk(TEMP_DEBUG DUMPH 
               "FCR(0x%04x)    : %08x     |     VTR(0x%04x)    : %08x\n", 
               FMAC3H_FCR, READ_REG(FMAC3H_FCR), 
               FMAC3H_VTR, READ_REG(FMAC3H_VTR));
        printk(TEMP_DEBUG DUMPH 
               "RWFFR(0x%04x)  : %08x     |     PMTR(0x%04x)   : %08x\n", 
               FMAC3H_RWFFR, READ_REG(FMAC3H_RWFFR), 
               FMAC3H_PMTR, READ_REG(FMAC3H_PMTR));
        printk(TEMP_DEBUG DUMPH 
               "MAR0H(0x%04x)  : %08x     |     MAR0L(0x%04x)  : %08x\n", 
               FMAC3H_MAR0H, READ_REG(FMAC3H_MAR0H), 
               FMAC3H_MAR0L, READ_REG(FMAC3H_MAR0L));
        printk(TEMP_DEBUG DUMPH 
               "BMR(0x%04x)    : %08x     |     TPDR(0x%04x)   : %08x\n", 
               FMAC3H_BMR, READ_REG(FMAC3H_BMR), 
               FMAC3H_TPDR, READ_REG(FMAC3H_TPDR));
        printk(TEMP_DEBUG DUMPH 
               "RPDR(0x%04x)   : %08x     |     RDLAR(0x%04x)  : %08x\n", 
               FMAC3H_RPDR, READ_REG(FMAC3H_RPDR), 
               FMAC3H_RDLAR, READ_REG(FMAC3H_RDLAR));
        printk(TEMP_DEBUG DUMPH 
               "TDLAR(0x%04x)  : %08x     |     SR(0x%04x)     : %08x\n", 
               FMAC3H_TDLAR, READ_REG(FMAC3H_TDLAR), 
               FMAC3H_SR, READ_REG(FMAC3H_SR));
        printk(TEMP_DEBUG DUMPH 
               "OMR(0x%04x)    : %08x     |     IER(0x%04x)    : %08x\n", 
               FMAC3H_OMR, READ_REG(FMAC3H_OMR), 
               FMAC3H_IER, READ_REG(FMAC3H_IER));
        printk(TEMP_DEBUG DUMPH
               "MFC(0x%04x)    : %08x     |     CHTDR(0x%04x)  : %08x\n",
               FMAC3H_MFC, READ_REG(FMAC3H_MFC),
               FMAC3H_CHTDR, READ_REG(FMAC3H_CHTDR));
        printk(TEMP_DEBUG DUMPH
               "CHRDR(0x%04x)  : %08x     |     CHTBAR(0x%04x) : %08x\n",
               FMAC3H_CHRDR, READ_REG(FMAC3H_CHRDR),
               FMAC3H_CHTBAR, READ_REG(FMAC3H_CHTBAR));
        printk(TEMP_DEBUG DUMPH "CHRBAR(0x%04x) : %08x\n\n",
               FMAC3H_CHRBAR, READ_REG(FMAC3H_CHRBAR));

    } /* if (ctl & REG_DUMP) { */


    /* 
     * Dump TX/RX Descriptor
     */ 
    if (ctl & DESC_DUMP) {
        
        if (txring) {
            printk(TEMP_DEBUG DUMPH 
             "+--------------------------------------------------------------+\n");
            printk(TEMP_DEBUG DUMPH 
             "|                     TX Descriptor Dump                       |\n");
            printk(TEMP_DEBUG DUMPH 
             "+--------------------------------------------------------------+\n");
            p = txring;
            for (i = 0; i < FMAC3H_TDESC_NUM; i++) {
                printk(TEMP_DEBUG DUMPH "DESC%03d(0x%08x) : %02x%02x %02x%02x", 
                       i, (unsigned int)p, WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
                printk(TEMP_DEBUG " %02x%02x %02x%02x",WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
                printk(TEMP_DEBUG "  %02x%02x %02x%02x", WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
                printk(TEMP_DEBUG " %02x%02x %02x%02x\n", WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
            }
            printk(TEMP_DEBUG "\n");
        }

        if (rxring) {
            printk(TEMP_DEBUG DUMPH 
             "+--------------------------------------------------------------+\n");
            printk(TEMP_DEBUG DUMPH 
             "|                     RX Descriptor Dump                       |\n");
            printk(TEMP_DEBUG DUMPH 
             "+--------------------------------------------------------------+\n");
            p = rxring;
            for (i = 0; i < FMAC3H_RDESC_NUM; i++) {
                printk(TEMP_DEBUG DUMPH "DESC%03d(0x%08x) : %02x%02x %02x%02x", 
                       i, (unsigned int)p, WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
                printk(TEMP_DEBUG " %02x%02x %02x%02x", WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
                printk(TEMP_DEBUG "  %02x%02x %02x%02x",WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
                printk(TEMP_DEBUG " %02x%02x %02x%02x\n",WORD3(*p), WORD2(*p), 
                       WORD1(*p), WORD0(*p));
                p++;
            }
 
            printk(TEMP_DEBUG "\n");
        }
    } /* if (ctl & DESC_DUMP) { */


    /* 
     * Dump Memory
     */  
    if (ctl & MEM_DUMP) {
        if (size == 0) {
            printk(TEMP_DEBUG DBGH "Passed SIZE is 0. No Dump!\n");
            return;
        }

        printk(TEMP_DEBUG DUMPH 
             "+-----------------------------------------------------------+\n");
        printk(TEMP_DEBUG DUMPH 
             "|                     Memory Dump                           |\n");
        printk(TEMP_DEBUG DUMPH 
             "+-----------------------------------------------------------+\n");
     
        mp = (char *)startp;
        for (i = 0; i < size; i++) {
            if (i % 16 == 0) {
                printk(DUMPH "0x%08x :", (unsigned int)mp);
            }

            printk(" %02x", *mp);

            if ((i + 1) == size) { /* the last one */
                printk("\n");
            } else {
                if ((i + 1) % 16 == 0) {
                    printk("\n");
                }                
                mp++;
            }
        }
    } /* if (ctl & MEM_DUMP) { */

    return;
}
#endif /* #ifdef FMAC3H_DEBUG */


/*
 * fmac3h_change_mtu -- change the MTU of system 
 * @dev: target net device structure
 * @new_mtu: new mtu value
 *
 * Description: change the MTU of system.  
 */
static int fmac3h_change_mtu(struct net_device *dev, int new_mtu)
{
    struct fmac3h_private *lp;
    int ret = 0;    /* the return value */

    lp = netdev_priv(dev);
    if (new_mtu < ETH_ZLEN || new_mtu > JUMBO_FRAME_MTU)
        return -EINVAL;

    dev->mtu = new_mtu;
 
    return ret;
}


/*
 * fmac3h_reg_read -- read FMAC3H's registers 
 * @base_addr: the base address of FMAC3H's register space
 * @reg: register address
 *
 * Description: return the value of register.  
 */
static u32 fmac3h_reg_read(void __iomem *base_addr, u32 reg)
{
    return __raw_readl(base_addr + reg);
}


/*
 * fmac3h_reg_write -- write FMAC3H's registers 
 * @base_addr: the base address of FMAC3H's register space
 * @reg: register address
 * @val: written value
 *
 * Description: write FMAC3H's registers without return value
 */
static void fmac3h_reg_write(void __iomem *base_addr, u32 reg, u32 val)
{

    __raw_writel(val, base_addr + reg);
    
    return;
}


/*
 * fmac3h_phy_wait -- confirm that PHY is not busy
 * @base_addr: the base address of FMAC3H's registers space.
 *
 * Description: In this function, the driver will confirm whether PHY is 
 *  not busy.  If PHY does not become free in time, the driver will send 
 *  warning message to user and continue.
 */
static int fmac3h_phy_wait(void __iomem *base_addr)
{
    u32 wtime;    /* waiting time */
    u32 val;

    /* confirm that PHY is not busy befor write/read GAR and GDR.
     * We set a waitingtime, if PHY does not become free in time, 
     * we will send warning message to user and continue. */
    wtime = WAIT_TIME;
    while (wtime--) {
        val = __raw_readl(base_addr + FMAC3H_GAR);
        if(!(val & GAR_GB)) {
            /*DBG_PRINT("PHY is not busy.\n");*/
            return 0;
        }
        udelay(DELAY_TIME); 
    }

    PMSG(KERN_ALERT, "PHY is busy!\n");
    DBG_DUMP(base_addr, REG_DUMP, NULL, NULL, 0, 0);

    return 1;
}


/*
 * fmac3h_phy_read -- read PHY's register 
 * @dev: target net device structure
 * @phy_id: the id of phy. range:0-31
 * @reg: phy register address
 * @read_val: the readen PHY register's value 
 * 
 * Description: read PHY's register by setting FMAC3H's GAR register and 
 *  GDR register.
 */
static int fmac3h_phy_read(struct net_device *dev, int phy_id, 
                           int reg, u32 *read_val)
{
    struct fmac3h_private *lp;
    int wval;
    int ret;    /* the return value */

    lp = netdev_priv(dev);

    /* Set the GAR register */
    wval = (phy_id << 11) | ((reg & 0x1f) << 6) | GAR_GW_R 
          | ((lp->mdc_clk & 0x07) << 2) | GAR_GB;

#ifndef LPDCP_FMAC3H_WORKAROUND
    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait(lp->base_addr);
    if (ret) {
        PMSG(KERN_ERR, "Read PHY Error!\n");
        DBG_FUN_ERREND();
        return ret;
    }
   
    fmac3h_reg_write(lp->base_addr, FMAC3H_GAR, wval);

    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait(lp->base_addr);
    if (ret) {
        PMSG(KERN_ERR, "Read PHY Error!\n");
        DBG_FUN_ERREND();
        return ret;
    }

    *read_val = fmac3h_reg_read(lp->base_addr, FMAC3H_GDR);

#else

    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait((void __iomem *)IO_ADDRESS(LPDCP_CPLD_BASE));
    if (ret) {
        PMSG(KERN_ERR, "Read PHY Error!\n");
        DBG_FUN_ERREND();
        return ret;
    }
   
    __raw_writel(wval, IO_ADDRESS(LPDCP_CPLD_BASE) + FMAC3H_GAR);

    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait((void __iomem *)IO_ADDRESS(LPDCP_CPLD_BASE));
    if (ret) {
        PMSG(KERN_ERR, "Read PHY Error!\n");
        DBG_FUN_ERREND();
        return ret;
    }

    *read_val = __raw_readl(IO_ADDRESS(LPDCP_CPLD_BASE) + FMAC3H_GDR);
#endif

    return 0;
}


/*
 * fmac3h_phy_write -- write PHY's register 
 * @dev: target net device structure
 * @phy_id: the id of phy. range:0-31
 * @reg: phy register address
 * @val: written value
 * 
 * Description: write PHY's register by setting FMAC3H's GAR register and 
 *  GDR register.
 */
static int fmac3h_phy_write(struct net_device *dev, int phy_id, 
                             int reg, int val)
{
    u32 wval;
    struct fmac3h_private *lp;
    int ret;
    
    lp = netdev_priv(dev);

    wval = (phy_id << 11) | (reg & 0x1f) << 6 | GAR_GW_W 
          | ((lp->mdc_clk & 0x03) << 2) | GAR_GB;

#ifndef LPDCP_FMAC3H_WORKAROUND
    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait(lp->base_addr);
    if (ret) {
        PMSG(KERN_ERR, "Write PHY Error!!!\n");
        DBG_FUN_ERREND();
        return ret;
    }

    /* Set the value that is want to be written in PHY register */
    fmac3h_reg_write(lp->base_addr, FMAC3H_GDR, val);
    /* Set PHY Register access information */
    fmac3h_reg_write(lp->base_addr, FMAC3H_GAR, wval);

    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait(lp->base_addr);
    if (ret) {
        PMSG(KERN_ERR, "Write PHY Error!!!\n");
        DBG_FUN_ERREND();
        return ret;
    }

#else

    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait((void __iomem *)IO_ADDRESS(LPDCP_CPLD_BASE));
    if (ret) {
        PMSG(KERN_ERR, "Write PHY Error!!!\n");
        DBG_FUN_ERREND();
        return ret;
    }

    /* Set the value that is want to be written in PHY register */
    __raw_writel(val, IO_ADDRESS(LPDCP_CPLD_BASE) + FMAC3H_GDR);
    /* Set PHY Register access information */
    __raw_writel(wval, IO_ADDRESS(LPDCP_CPLD_BASE) + FMAC3H_GAR);

    /* Wait until GMII/MII is not busy */
    ret = fmac3h_phy_wait((void __iomem *)IO_ADDRESS(LPDCP_CPLD_BASE));
    if (ret) {
        PMSG(KERN_ERR, "Write PHY Error!!!\n");
        DBG_FUN_ERREND();
        return ret;
    }
#endif

    return 0;
}


/*
 * fmac3h_mii_read -- ethool use this function to read PHY's registers
 * @dev: target net device structure
 * @phy_id: the id of phy. range:0-31
 * @reg: phy register address
 * 
 * Description: ethool use this function to read PHY's registers.
 */
static int fmac3h_mii_read(struct net_device *dev, int phy_id, int reg)
{
    int val;
    int ret;    /* the return value */

    ret = fmac3h_phy_read(dev, phy_id, reg, &val);

    return val;
}


/*
 * fmac3h_mii_write -- ethool use this function to write PHY's registers
 * @dev: target net device structure
 * @phy_id: the id of phy. range:0-31
 * @reg: phy register address
 * @val: written value
 * 
 * Description: ethool use this function to write PHY's registers.
 */
static void fmac3h_mii_write(struct net_device *dev, int phy_id, 
                             int reg, int val)
{
    int ret;
    
    ret = fmac3h_phy_write(dev, phy_id, reg, val);

    return;
}


/*
 * fmac3h_stop_rxtx -- stop rx engine and tx engine 
 * @lp: taget F_MAC3H device structure
 *
 * Description: In this function, the driver will stop RX/TX DMA by setting
 *  OMR register. If RX/TX engines do not stop in time, the driver will send
 *  warning message to user and continue.
 */
static void fmac3h_stop_rxtx(struct fmac3h_private *lp)
{
    u32 val;
    u32 wtime;    /* waiting time */
    u32 i;

    DBG_FUN_START();

    val = fmac3h_reg_read(lp->base_addr, FMAC3H_OMR);

    /* Stop RX:set 0 to OMR[bit1]; Stop TX:set 0 to OMR[bit13] */
    val &= ~(START_TX | START_RX);

    /* Send stop command to TX and RX process */
    fmac3h_reg_write(lp->base_addr, FMAC3H_OMR, val);

    /* TX/RX process will not stop immediately with stop command, until
     * they finish the work which has been started.
     * We set a waitingtime, if TX/RX process do not stop in time, we 
     * will send warning message to user and continue. */
    wtime = WAIT_TIME;
    while (wtime--) {
        val = fmac3h_reg_read(lp->base_addr, FMAC3H_OMR);
        if(!(val & (START_TX | START_RX))) {
            DBG_PRINT("TX/RX Stoped.\n");
            goto drop_packet;
        }
        udelay(DELAY_TIME); 
    }

    PMSG(KERN_CRIT, "Can not to stop DMA!!!\n");

drop_packet:
    DBG_DUMP(lp->base_addr, REG_DUMP | DESC_DUMP, (int *)lp->tx_ring, 
               (int *)lp->rx_ring, 0, 0);

    /* after stop TX/RX, the packets that has not been tx/rx will be dropped */
    /* Check TX descriptors */
    for (i = 0; i < FMAC3H_TDESC_NUM; i++) {
        if (lp->tx_ring[i].opts1 & OWN_BIT) {
            PMSG(KERN_WARNING,  
                 "The packet of TX descirptor(%d) will be dropped!!!\n", i);
            lp->net_stats.tx_dropped++;
        }
    }

    /* Check RX descriptors */
    for (i = 0; i < FMAC3H_RDESC_NUM; i++) {
        if (!(lp->rx_ring[i].opts1 & OWN_BIT)) {
            PMSG(KERN_WARNING,  
                 "The packet of RX descirptor(%d) will be dropped!!!\n", i);
            lp->net_stats.rx_dropped++;
        }
    }

    DBG_FUN_END();
    return;
}


/*
 * fmac3h_get_stats_from_mfc -- get RX status by reading MFC register 
 * @lp: taget F_MAC3H device structure
 *
 * Description: In this function, the driver will read MFC register to fill
 *  net_stats structure.
 */
static void fmac3h_get_stats_from_mfc(struct fmac3h_private *lp)
{
    u32 val;

    DBG_FUN_START();

    val = fmac3h_reg_read(lp->base_addr, FMAC3H_MFC);
    /* Miss error = Number of missed frame by FMAC3H (MFC[bit27:17])
     *             + Number of missed frame by HOST (MFC[bit15:0]) */ 
    lp->net_stats.rx_missed_errors += ((val & 0x0000FFFF)            /* NMFF */
                                      + ((val >> 17) & 0x000003FF)); /* NMFH */
    
    DBG_FUN_END();
    return;
}


/*
 * fmac3h_setup_aneg -- setup autonegotiation 
 * @lp: taget F_MAC3H device structure
 * @aneg_level: autonegotiation wait level
 *
 * Description: In this function, the driver will check whether 
 *  auto-negotiation is permitted before setup auto-negotiation.
 */
static int fmac3h_setup_aneg(struct fmac3h_private *lp, int aneg_level)
{
    struct mii_if_info *mii;
    u32 adv, ctl; 
    u32 wtime;
    int ret;

    DBG_FUN_START();

    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (unlikely(lp->dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    mii = &lp->mii;

    if (mii->force_media) {
        PMSG(KERN_ERR, "Not allow to autonegotiation!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    /* reset phy */
    ret = fmac3h_phy_read(lp->dev, mii->phy_id, MII_BMCR, &ctl);
    if (ret) {
        PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    ctl |= BMCR_RESET;
    ret = fmac3h_phy_write(lp->dev, mii->phy_id, MII_BMCR, ctl);
    if (ret) {
        PMSG(KERN_ERR, "Failed to write PHY register!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    /* to sure reset complete in time */
    wtime = WAIT_TIME;
    DBG_PRINT("Wait start. Time(%d)\n", wtime);
    while (wtime--) {
        ret = fmac3h_phy_read(lp->dev, mii->phy_id, MII_BMCR, &ctl);
        if (ret) {
            PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
            DBG_FUN_ERREND();
            return -EACCES;
        }
        if (!(ctl & BMCR_RESET)) {
            wtime = 0;
            break;
        }
        udelay(DELAY_TIME);
    }

    if (ctl & BMCR_RESET) {
        PMSG(KERN_ERR, "Failed to setup autonegotiation!!!\n");
        DBG_FUN_ERREND();
        return -EBUSY;
    }

    /* setup standard advertise */
    ret = fmac3h_phy_read(lp->dev, mii->phy_id, MII_ADVERTISE, &adv);
    if (ret) {
        PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    adv &= ~(ADVERTISE_ALL | ADVERTISE_100BASE4 
                | ADVERTISE_PAUSE_CAP | ADVERTISE_PAUSE_ASYM);
    if (mii->advertising & ADVERTISED_10baseT_Half)
        adv |= ADVERTISE_10HALF;
    if (mii->advertising & ADVERTISED_10baseT_Full)
        adv |= ADVERTISE_10FULL;
    if (mii->advertising & ADVERTISED_100baseT_Half)
        adv |= ADVERTISE_100HALF;
    if (mii->advertising & ADVERTISED_100baseT_Full)
        adv |= ADVERTISE_100FULL;
    if (mii->advertising & ADVERTISED_Pause)
        adv |= ADVERTISE_PAUSE_CAP;
    if (mii->advertising & ADVERTISED_Asym_Pause)
        adv |= ADVERTISE_PAUSE_ASYM;
    ret = fmac3h_phy_write(lp->dev, mii->phy_id, MII_ADVERTISE, adv);
    if (ret) {
        PMSG(KERN_ERR, "Failed to write PHY register!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    if (mii->supports_gmii) {
        DBG_PRINT("GMII Port\n");
        ret = fmac3h_phy_read(lp->dev, mii->phy_id, MII_CTRL1000, &adv);
        if (ret) {
            PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
            DBG_FUN_ERREND();
            return -EACCES;
        }
   
        adv &= ~(ADVERTISE_1000FULL | ADVERTISE_1000HALF);
        if (mii->advertising & ADVERTISED_1000baseT_Full)
            adv |= ADVERTISE_1000FULL;
        if (mii->advertising & ADVERTISED_1000baseT_Half)
            adv |= ADVERTISE_1000HALF;
        ret = fmac3h_phy_write(lp->dev, mii->phy_id, MII_CTRL1000, adv);
        if (ret) {
            PMSG(KERN_ERR, "Failed to write PHY register!!!\n");
            DBG_FUN_ERREND();
            return -EACCES;
        }
    }

    /* start auto negotiation */
    DBG_PRINT("Start autonegotiation\n");
    ret = fmac3h_phy_read(lp->dev, mii->phy_id, MII_BMCR, &ctl);
    if (ret) {
        PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    ctl |= BMCR_ANENABLE | BMCR_ANRESTART;
    ret = fmac3h_phy_write(lp->dev, mii->phy_id, MII_BMCR, ctl);
    if (ret) {
        PMSG(KERN_ERR, "Failed to write PHY register!!!\n");
        DBG_FUN_ERREND();
        return -EACCES;
    }

    /* sure autonegotiation success */
    wtime = (aneg_level == 0) ? WAIT_TIME : ANEG_WAIT_TIME;
    while (wtime--) {
        ret = fmac3h_phy_read(lp->dev, mii->phy_id, MII_BMSR, &ctl);
        if (ret) {
            PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
            DBG_FUN_ERREND();
            return -EACCES;
        }

        if ((ctl & BMSR_LSTATUS) && (ctl & BMSR_ANEGCOMPLETE)) {
            wtime = 0;
            DBG_FUN_END();
            return 0;
        }
        udelay(DELAY_TIME);
    }

    if (aneg_level != 0) {
        if (!((ctl & BMSR_LSTATUS) && (ctl & BMSR_ANEGCOMPLETE))) {
            PMSG(KERN_ERR, "Setup Autonegotiation is failed!!!\n");
            DBG_PRINT("BMSR(0x%08x) Wish(0x%08x)\n", ctl, 
                      BMSR_LSTATUS | BMSR_ANEGCOMPLETE);
            DBG_FUN_ERREND();
            return -EBUSY;
        }
    }
    
    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_init_hw -- initializate the registers of F_MAC3H  
 * @lp: taget F_MAC3H device structure
 * @aneg_level: autonegotiation wait level
 *
 * Description: In this function, the driver will setup autonegotiation and
 *  set registers of FMAC3H.
 */
static int fmac3h_init_hw(struct fmac3h_private *lp, int aneg_level)
{
    u32 val, lpa, advert;
    int err;
    dma_addr_t rx_adr, tx_adr;
    
    DBG_FUN_START();

    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    /* 1.Set BMR(MDC Bus Mode Register) */
    val = BMR_DBO_LIT     /* Descriptor Little Endian    */
         | BMR_BLE_LIT    /* F_MAC3H->HOST Little Endian */
         | BMR_DSL        /* Descripter Skip Length is 0 */
#ifndef MB86HXX_FMAC3H_WORKAROUND
         | BMR_PBL_32;    /* Burst Length is 32 Bytes    */
#else
         | BMR_PBL_1;     /* Burst Length is 1 Bytes     */
#endif
    fmac3h_reg_write(lp->base_addr, FMAC3H_BMR, val);

    /* 2.Set IER(MDC Interrupt Enable Register) */
    val = (INT_EARY_RX & INT_MASK)        /* Eary Receive Int Unable          */
         | (INT_BUS_ERR & INT_ENABLE)     /* Fatal Bus Error Int Enable       */
         | (INT_EARY_TX & INT_MASK)       /* Eary Transmit Int Unable         */
         | (INT_RX_WT & INT_MASK)         /* RX Watchdog Timeout Int Unable   */
         | (INT_RX_STOP & INT_ENABLE)     /* Receive Process Stop Int Enable  */
         | (INT_RXB_UNAV & INT_ENABLE)    /* RX Buffer Unaballable Int Enable */
         | (INT_RX_INT & INT_ENABLE)      /* Receive Int Enable               */
         | (INT_TX_UFLOW & INT_ENABLE)    /* Transmit Underflow Int Enable    */
         | (INT_RX_OVERF & INT_ENABLE)    /* Receive Overlow Int Enable       */
         | (INT_TX_JT & INT_MASK)         /* TX Jabber Timeout Int Unable     */
         | (INT_TXB_UNAV & INT_MASK)      /* TX Buffer Unaballable Int Enable */
         | (INT_TX_STOP & INT_ENABLE)     /* Transmit Process Stop Int Enable */
         | (INT_TX_INT & INT_ENABLE)      /* Transmit Int Enable              */
         | (INT_NORMAL & INT_ENABLE)      /* Normal Interrupt Summary Enable  */
         | (INT_ABNORMAL & INT_ENABLE);   /* Abnormal Interrupt Summary       */
    fmac3h_reg_write(lp->base_addr, FMAC3H_IER, val);

    /* 3.Set RDLAR(MDC Receive Descriptor List Address Register)
     *       TDLAR(MDC Transmit Descriptor List Address Register) */
    rx_adr = lp->ring_dma;
    tx_adr = lp->ring_dma + (sizeof(struct fmac3h_desc) * FMAC3H_RDESC_NUM);
    fmac3h_reg_write(lp->base_addr, FMAC3H_RDLAR, rx_adr);
    fmac3h_reg_write(lp->base_addr, FMAC3H_TDLAR, tx_adr);

    /* 4.Set MFFR(MAC Frame Filter Register) 
     *   With this set the LAN can do:
     *   (1).Receive broadcast frames
     *   (2).Receive packets that has the same multicast address as MARn 
     *   (3).Receive packets that has the same unicast address as MARn   */
    val = REC_ALL_ON          /* Receive all frame        */
         | PROMIS_MODE;       /* Promiscuous Mode         */
    fmac3h_reg_write(lp->base_addr, FMAC3H_MFFR, val);

    /* 5.Set MHTRH(MAC Hash Table Register High)
     *       MHTRL(MAC Hash Table Register Low) */
    fmac3h_reg_write(lp->base_addr, FMAC3H_MHTRH, 0);
    fmac3h_reg_write(lp->base_addr, FMAC3H_MHTRL, 0);

    /* 6.Set MMC Register */
    fmac3h_reg_write(lp->base_addr, FMAC3H_MMC_CNTL, CLEAR_ALL);
    val = RESET_ON_READ       /* Auto clear after is readed */
         | STOP_ROLL_OVER;    /* Stop count when arrive MAX */
    fmac3h_reg_write(lp->base_addr, FMAC3H_MMC_CNTL, val);
    /* Mask all interrupt of MMC */
    fmac3h_reg_write(lp->base_addr, FMAC3H_MMC_INTR_MASK_RX, 0x00FFFFFF);
    fmac3h_reg_write(lp->base_addr, FMAC3H_MMC_INTR_MASK_TX, 0x01FFFFFF);

    /* 7.Setup PHY Autonegotiation */
    err = fmac3h_setup_aneg(lp, aneg_level);
    if (err) {
        PMSG(KERN_ERR, "Failed to setup AutoNegotiation!!!\n");
        DBG_FUN_ERREND();
        return err;
    }

    /* 8.Set MCR(MAC Configuration Register) */
    val = (MCR_WD & SET_1)            /* Disable RX Watchdog timeout */
         | (MCR_JD & SET_1)           /* Disable TX Jabber timer     */
         | (MCR_BE & SET_0)           /* Frame Burst disable         */
         | (MCR_JE & SET_1)           /* Jumbo Frame Enable          */
         | (MCR_EM & SET_0)           /* Little Endian Mode          */
         | (MCR_DO & SET_0)           /* Enable Receive Own          */
         | (MCR_LM & SET_0)           /* Not Loop-back Mode          */
         | (MCR_DR & SET_0)           /* Enable Retry                */
         | (MCR_APS & SET_0)          /* Not Automatic Pad Stripping */
         | (MCR_ACS & SET_0)          /* Not Automatic CRC Stripping */
         | (MCR_DC & SET_0)           /* Disable Deferral Check      */
         | (MCR_TX_ENABLE & SET_1)    /* Enable Transmitter          */
         | (MCR_RX_ENABLE & SET_1)    /* Enable Receiver             */
         | MCR_BL_00;                 /* Back-off Limit is setted 0  */
    /* MCR[bit15,11] is seted based on PHY`s setting        */
    if (!lp->mii.supports_gmii) {   /* MII Port */
        val |= MII_PORT;
    }
    wmb();
    err = fmac3h_phy_read(lp->dev, lp->mii.phy_id, MII_LPA, &lpa);
    if (err) {
        PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
        DBG_FUN_ERREND();
        return err;
    }
    err = fmac3h_phy_read(lp->dev, lp->mii.phy_id, MII_ADVERTISE, &advert);
    if (err) {
        PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
        DBG_FUN_ERREND();
        return err;
    }
    
    if ((lpa & advert) & (LPA_10FULL | LPA_100FULL)) {    /* Full Duplex */
       val |= FULL_DUPLEX;
    }
    fmac3h_reg_write(lp->base_addr, FMAC3H_MCR, val);

    /* add wait time after setting MCR over 1.6us */
    udelay(10); /* wait 10us */

    /*
     * CAUTION!
     * For F_MAC3H V3.0, RDLAR/TDLAR must be set after setting the MCR.
     */

    /* 8.Set RDLAR(MDC Receive Descriptor List Address Register)
     *       TDLAR(MDC Transmit Descriptor List Address Register) */
    rx_adr = lp->ring_dma;
    tx_adr = lp->ring_dma + (sizeof(struct fmac3h_desc) * FMAC3H_RDESC_NUM);
    fmac3h_reg_write(lp->base_addr, FMAC3H_RDLAR, rx_adr);
    fmac3h_reg_write(lp->base_addr, FMAC3H_TDLAR, tx_adr);

    /* 9.Set OMR to start RX and TX */
#ifndef MB86HXX_FMAC3H_WORKAROUND
    val = /* OMR_SF */ 
          OMR_TTC_64B       /* TX after 64byte written in FIFO */
         /* | START_TX */   /* Start Transmissin */
         | START_RX;        /* Start Receive     */
#else
    val = OMR_SF 
         /*OMR_TTC_64B */   /* TX after 32byte written in FIFO */
         /* | START_TX */   /* Start Transmissin */
         | START_RX;        /* Start Receive     */
#endif
    fmac3h_reg_write(lp->base_addr, FMAC3H_OMR, val);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_clean_ring -- free the skbuff of every descriptor 
 * @lp: taget F_MAC3H device structure
 *
 * Description: Free the skbuff which has not been freed and clean the 
 *  setting of all descriptors. 
 */
static int fmac3h_clean_ring(struct fmac3h_private *lp)
{
    int i;

    DBG_FUN_START();

    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    memset(lp->tx_ring, 0, sizeof(struct fmac3h_desc) * FMAC3H_TDESC_NUM);
    lp->tx_ring[FMAC3H_TDESC_NUM - 1].opts2 = END_RING;
    wmb();
    memset(lp->rx_ring, 0, sizeof(struct fmac3h_desc) * FMAC3H_RDESC_NUM);
    lp->rx_ring[FMAC3H_RDESC_NUM - 1].opts2 = END_RING;
    wmb();

    /* free all of receive skb */ 
    for (i = 0; i < FMAC3H_RDESC_NUM; i++) {
        if (lp->rx_skb[i].skb) {
            DBG_PRINT("SKB(RX DESC:%d) free\n", i);
            /* delete the mapping of skbuff */
            dma_unmap_single(NULL, lp->rx_skb[i].mapping, 
                             lp->rx_buf_sz, DMA_FROM_DEVICE);
            /* free buffer space */
            dev_kfree_skb(lp->rx_skb[i].skb);
            lp->rx_skb[i].skb = NULL;
            lp->rx_skb[i].mapping = 0;
        }
    }

    /* free the transmit skb that has not been freed */
    for (i = 0; i < FMAC3H_TDESC_NUM; i++) {
        struct sk_buff *temp_skb = lp->tx_skb[i].skb;
        if (temp_skb) {
            DBG_PRINT("SKB(TX DESC:%d) free\n", i);
            /* delete the mapping of skbuff */
            dma_unmap_single(NULL, lp->tx_skb[i].mapping, 
                             temp_skb->len, DMA_TO_DEVICE);
            /* free buffer space */
            dev_kfree_skb(temp_skb);
            lp->tx_skb[i].skb = NULL;
            lp->tx_skb[i].mapping = 0;
        }
    }

    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_free_ring -- unmap and free descirptor and buffer spaces 
 * @lp: target F_MAC3H device structure
 * 
 * Description: Check every descriptor's skbuff pointer. If the skbuff has
 *  not been freed then free it. At last free all of descriptor spaces 
 */
static int fmac3h_free_ring(struct fmac3h_private *lp)
{
    int err;

    DBG_FUN_START();

    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (lp->rx_ring) {
        DBG_DUMP(lp->base_addr, DESC_DUMP, (int *)lp->tx_ring, 
                   (int *)lp->rx_ring, 0, 0);

        /* free alloced buffer and clean descriptor */
        err = fmac3h_clean_ring(lp);
        if (unlikely(err)) {
            PMSG(KERN_ERR, "Failed to free buffer space!!!\n");
            DBG_FUN_ERREND();
            return err;
        }
    
        /* delete the head and tail flag of descriptor ring */
        lp->tx_head = lp->tx_tail = 0;
        lp->rx_tail = 0;
    }

    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_refill_rx -- refill rx descriptors
 * @lp: target F_MAC3H device structure
 *
 * Description: In this function, the driver will malloc a skbuffer for 
 *  every rx descriptor and set OWN bit.
 */
static int fmac3h_refill_rx(struct fmac3h_private *lp)
{
    int i, err;

    DBG_FUN_START();

    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* alloc a receive buffer for every receive descriptor */
    for (i = 0; i < FMAC3H_RDESC_NUM; i++) {
        struct sk_buff *skb;

        skb = dev_alloc_skb(lp->rx_buf_sz);    /* alloc receive buffer */
        if (!skb) {
            PMSG(KERN_ERR, "Failed to alloc receive buffer!!!\n");
            err = -ENOMEM;
            goto err_out;
        }

        skb->dev = lp->dev;

        /* set the alloced skb in receive skb array */
        lp->rx_skb[i].mapping = dma_map_single(NULL, skb->data, 
                                               lp->rx_buf_sz, DMA_FROM_DEVICE);
        lp->rx_skb[i].skb = skb;

        /* set receive descriptor */
        if (i == (FMAC3H_RDESC_NUM - 1)) {    /* the last one in ring */
            lp->rx_ring[i].opts2 = (END_RING | lp->rx_buf_sz);
        } else {
            lp->rx_ring[i].opts2 = lp->rx_buf_sz;
        }
        lp->rx_ring[i].addr1 = lp->rx_skb[i].mapping;
        lp->rx_ring[i].addr2 = 0;
        lp->rx_ring[i].opts1 = OWN_BIT;    /* set OWN bit */

    }

    DBG_FUN_END();
    return 0;

err_out:
    fmac3h_clean_ring(lp);
    DBG_FUN_ERREND();
    return err;
}


/*
 * fmac3h_init_rings -- initializate TX and RX rings 
 * @lp: target F_MAC3H device structure
 *
 * Description: In this function, the driver will initializate the head and 
 *  tail flag of TX/RX rings and refill rx descriptors.
 */
static int fmac3h_init_rings(struct fmac3h_private *lp)
{
    int ret = 0;

    DBG_FUN_START();

    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp->rx_tail = 0;
    lp->tx_head = lp->tx_tail = 0;

    ret = fmac3h_refill_rx(lp);

    DBG_FUN_END();
    return ret;
}


/*
 * fmac3h_ioctl -- performs interface-specific ioctl commands 
 * @dev: target net device structure
 * @rq: the rq pointer points to a kernel-space address that holds a copy
 *      of the structure passed by the user.
 * @cmd: ioctl command
 *
 * Description: In FMAC3H driver, three commands are supported. SIOCGMIIPHY, 
 *  SIOCGMIIREG and SIOCSMIIREG.   
 */
static int fmac3h_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
    struct fmac3h_private *lp;
    int ret = 0;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (unlikely(rq == NULL)) {
        PMSG(KERN_CRIT, "Not proper req!!!\n");
        DBG_FUN_ERREND();
        return -EINVAL;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (!netif_running(dev)) {
        PMSG(KERN_CRIT, "Device is unvaluable!!!\n");
        DBG_FUN_ERREND();
        return -EINVAL;
    }

    spin_lock_irq(&lp->lock);

    switch (cmd) {
        case SIOCGMIIPHY:
        case SIOCGMIIREG:
        case SIOCSMIIREG:
            DBG_PRINT("Valuable ioctl command\n");    
            ret = generic_mii_ioctl(&lp->mii, if_mii(rq), cmd, NULL);
            break;
        default:
            PMSG(KERN_ERR, "Do not support the command(0x%04x)!!!\n", cmd);
            ret = EOPNOTSUPP;
            break;
    }
    spin_unlock_irq(&lp->lock);

    DBG_FUN_END();
    return ret;
}


/*
 * fmac3h_get_drvinfo -- report driver information 
 * @dev: target net device structure
 * @info: ethtool driver information structure
 *
 * Description: set driver information of structure ethtool_drvinfo.
 */
static void fmac3h_get_drvinfo(struct net_device *dev, 
                               struct ethtool_drvinfo *info)
{
    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    if (unlikely(info == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper command!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    strlcpy(info->driver, DRV_NAME, sizeof(info->driver));
    strlcpy(info->version, DRV_VERSION, sizeof(info->version));
    strlcpy(info->bus_info, "mb86hxx_fmac3h", sizeof(info->bus_info));

    DBG_FUN_END();
    return;
}


/*
 * fmac3h_get_settings -- get device-specific settings 
 * @dev: target net device structure
 * @cmd: ethtool command 
 *
 * Description: fmac3h_get_settings is passed an &ethtool_cmd to fill in. 
 *  It returns an negative errno or zero.
 */
static int fmac3h_get_settings(struct net_device *dev, 
                               struct ethtool_cmd *cmd)
{
    struct fmac3h_private *lp;
    int ret;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (unlikely(cmd == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper command!!!\n");
        DBG_FUN_ERREND();
        return -EINVAL;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    spin_lock_irq(&lp->lock);
    ret = mii_ethtool_gset(&lp->mii, cmd);
    spin_unlock_irq(&lp->lock);

    DBG_FUN_END();
    return ret;
}


/*
 * fmac3h_set_settings -- Set device-specific settings 
 * @dev: target net device structure
 * @cmd: ethtool command 
 *
 * Description: fmac3h_set_settings is passed an &ethtool_cmd and should 
 *  attempt to set all the settings this device supports. It returns an 
 *  error value if something goes wrong (otherwise 0).
 */
static int fmac3h_set_settings(struct net_device *dev, struct ethtool_cmd *cmd)
{
    struct fmac3h_private *lp;
    u32 val, lpa, advert;
    u32 ctl = 0;
    u32 wtime;
    int ret;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (unlikely(cmd == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper command!!!\n");
        DBG_FUN_ERREND();
        return -EINVAL;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    spin_lock_irq(&lp->lock);

    ret = mii_ethtool_sset(&lp->mii, cmd);
    /* MCR[bit15,11] is seted based on PHY`s setting */
    if (!ret) {
        val = fmac3h_reg_read(lp->base_addr, FMAC3H_MCR);
        if (lp->mii.supports_gmii) {        /* GMII Port */
            /* MCR[bit15]: 0 = GMII Port; 1 = MII Port */
            val &= ~ MII_PORT;
        } else {                            /* MII Port */
            val |= MII_PORT;
        }

        if (!lp->mii.force_media) { 
            /* sure autonegotiation success */
            wtime = ANEG_WAIT_TIME;
            while (wtime--) {
                ctl = lp->mii.mdio_read(dev, lp->mii.phy_id, MII_BMSR);

                if ((ctl & BMSR_LSTATUS) && (ctl & BMSR_ANEGCOMPLETE)) {
                    wtime = 0;
                    DBG_FUN_END();
                    break;
                }
                udelay(DELAY_TIME);
            }

            if (!((ctl & BMSR_LSTATUS) && (ctl & BMSR_ANEGCOMPLETE))) {
                PMSG(KERN_ERR, "Setup Autonegotiation is failed!!!\n");
                DBG_PRINT("BMSR(0x%08x) Wish(0x%08x)\n", ctl, 
                          BMSR_LSTATUS | BMSR_ANEGCOMPLETE);
                DBG_FUN_ERREND();
                return -EBUSY;
            }
        }

        advert = lp->mii.mdio_read(dev, lp->mii.phy_id, MII_ADVERTISE);
        lpa = lp->mii.mdio_read(dev, lp->mii.phy_id, MII_LPA);
        if ((lpa & advert) & (LPA_10FULL | LPA_100FULL)) {    /* Full Duplex */
           val |= FULL_DUPLEX;
        } else {    /* Half Duplex */
           val &= ~FULL_DUPLEX;
        }
        
        fmac3h_reg_write(lp->base_addr, FMAC3H_MCR, val);
    } else {
        PMSG(KERN_ERR, "Failed to set device!!!");    
    }
    spin_unlock_irq(&lp->lock);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    DBG_FUN_END();
    return ret;
}


/*
 * fmac3h_nway_reset -- restart autonegotiation 
 * @dev: target net device structure
 *
 * Description: In this function, the driver will restart autonegotiation 
 *  and set the result of autonegotiation to MCR registers.
 */
static int fmac3h_nway_reset(struct net_device *dev)
{
    struct fmac3h_private *lp;
    u32 val, lpa, advert;
    u32 ctl = 0;
    u32 wtime;
    int ret;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    spin_lock_irq(&lp->lock);

    ret = mii_nway_restart(&lp->mii);
    /* MCR[bit15,11] is seted based on PHY`s setting */
    if (!ret) {
        val = fmac3h_reg_read(lp->base_addr, FMAC3H_MCR);
        if (lp->mii.supports_gmii) {        /* GMII Port */
            /* MCR[bit15]: 0 = GMII Port; 1 = MII Port */
            val &= ~ MII_PORT;
        } else {                            /* MII Port */
            val |= MII_PORT;
        }

        if (!lp->mii.force_media) { 
            /* sure autonegotiation success */
            wtime = ANEG_WAIT_TIME;
            while (wtime--) {
                ctl = lp->mii.mdio_read(dev, lp->mii.phy_id, MII_BMSR);

                if ((ctl & BMSR_LSTATUS) && (ctl & BMSR_ANEGCOMPLETE)) {
                    wtime = 0;
                    DBG_FUN_END();
                    break;
                }
                udelay(DELAY_TIME);
            }

            if (!((ctl & BMSR_LSTATUS) && (ctl & BMSR_ANEGCOMPLETE))) {
                PMSG(KERN_ERR, "Setup Autonegotiation is failed!!!\n");
                DBG_PRINT("BMSR(0x%08x) Wish(0x%08x)\n", ctl, 
                          BMSR_LSTATUS | BMSR_ANEGCOMPLETE);
                DBG_FUN_ERREND();
                return -EBUSY;
            }
        }

        advert = lp->mii.mdio_read(dev, lp->mii.phy_id, MII_ADVERTISE);
        lpa = lp->mii.mdio_read(dev, lp->mii.phy_id, MII_LPA);
        if ((lpa & advert) & (LPA_10FULL | LPA_100FULL)) {    /* Full Duplex */
           val |= FULL_DUPLEX;
        } else {    /* Half Duplex */
           val &= ~FULL_DUPLEX;
        }
        
        fmac3h_reg_write(lp->base_addr, FMAC3H_MCR, val);
    } else {
        PMSG(KERN_ERR, "Failed to restart device!!!");
    }

    spin_unlock_irq(&lp->lock);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    DBG_FUN_END();
    return ret;
}


/* Ethtool supported by F_MAC3H */
static const struct ethtool_ops fmac3h_ethtool_ops = {
    .get_link     = ethtool_op_get_link,
    .get_drvinfo  = fmac3h_get_drvinfo,
    .get_settings = fmac3h_get_settings,
    .set_settings = fmac3h_set_settings,
    .nway_reset   = fmac3h_nway_reset,
};


/*
 * fmac3h_stop_hw -- stop hardware 
 * @lp: target F_MAC3H device structure
 *
 * Description: In this function, the driver will stop the TX/RX engine and
 *  clear the status of hardware.
 */
static void fmac3h_stop_hw(struct fmac3h_private *lp)
{
    u32 val;

    DBG_FUN_START();

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    /* disable all interrupt */
    fmac3h_reg_write(lp->base_addr, FMAC3H_IER, 0);

    /* stop TX and RX process */
    fmac3h_stop_rxtx(lp);

    /* clear status register */
    val = fmac3h_reg_read(lp->base_addr, FMAC3H_SR);
    fmac3h_reg_write(lp->base_addr, FMAC3H_SR, val);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    DBG_FUN_END();
    return;
}


/*
 * fmac3h_check_carrier -- check the status of carrier 
 * @data: target net device structure
 *
 * Description: this function will be called every one second to check 
 *  the status of carrier.
 */
static void fmac3h_check_carrier(unsigned long data)
{
    struct net_device *dev;
    struct fmac3h_private *lp;
    u32 oldlink, newlink;
    int err;

    dev = (struct net_device*)data;
    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return;
    }
    
    spin_lock_irq(&lp->lock);

    /* check the old link state of carrier 
     * the return value of netif_carrier_ok()
     *  1: carrier on
     *  0: carrier off */
    oldlink = netif_carrier_ok(dev);

    /* check the new link state of carrier 
     * the return value of mii_link_ok()
     *  1: carrier on 
     *  0: carrier off */
    newlink = (unsigned int)mii_link_ok(&lp->mii);

    /* update link state in dev->state */
    if (!oldlink && newlink) {
        /* Link OFF->ON */
        /* free descriptor and buffer space */
        err = fmac3h_free_ring(lp);
        if (unlikely(err)) {
            PMSG(KERN_ERR, "Failed to free dscriptor and buffer!!!\n");
            DBG_FUN_ERREND();
            goto out;
        }
        err = fmac3h_init_rings(lp);
        if (err) {
            PMSG(KERN_ERR, "Failed to initializate DMA descriptor!!!\n");
            DBG_FUN_ERREND();
            goto out;
        }
        err = fmac3h_init_hw(lp, 1);
        if (err) {
            PMSG(KERN_WARNING, "Failed to initializate device!!!\n");
            DBG_FUN_ERREND();
            goto out;
        }
        /* start queue to allow transmit */
        netif_carrier_on(dev);
        netif_start_queue(dev);

        PMSG(KERN_NOTICE, "Link UP\n");
    }
    else if (oldlink && !newlink) {
        /* Link ON->OFF */
        fmac3h_stop_hw(lp);
        netif_stop_queue(dev);
        netif_carrier_off(dev);

        PMSG(KERN_NOTICE, "Link DOWN\n");
    }

out:
    spin_unlock_irq(&lp->lock);

    /* set the next check */
    mod_timer(&lp->timer, CHECK_CARRIER_TIME);

    return;
}


/*
 * fmac3h_rx -- this function will be called when there is receive 
 *  interrupt(SR[bit6]) or receive buffer unavallable interrupt(SR[bit7]) 
 *  or receive FIFO overflow interrupt(SR[bit4])
 * @dev: target net device structure
 *
 * Description: In this function, the driver will check rx descriptor and 
 *  post skb to next network layer.
 */
static int fmac3h_rx(struct net_device *dev)
{
    struct fmac3h_private *lp;
    u32 rxtail, desc_num;
    u32 fdesc, ldesc;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    desc_num = FMAC3H_RDESC_NUM;    /* check the whole RX ring */
    rxtail = lp->rx_tail;           /* start check from this descriptor */ 

    DBG_DUMP(lp->base_addr, DESC_DUMP, NULL, (int *)lp->rx_ring, 0, 0);

    /* Check every RxDesc. if its OWN bit is 0, post its skbuf to nex layer */
    while (desc_num) {
        struct sk_buff *skb;
        struct sk_buff *reskb;  /* the skbuf passed to the upper layer */
        u32 rx_sta, datalen, buflen;
        u32 bnum, i;

        rx_sta = lp->rx_ring[rxtail].opts1;
        if (rx_sta & OWN_BIT) {    /* OWN bit is 1 */
            DBG_PRINT("RX Desc(%d) is not mine.\n", rxtail);
            break;
        }

        fdesc = rxtail;
        DBG_PRINT("First Desc(%d):0x%08x\n", fdesc, rx_sta);
        ldesc = rxtail;
        for (bnum = 1; bnum <= desc_num; bnum++) {
            rx_sta = lp->rx_ring[ldesc].opts1;

            if (rx_sta & OWN_BIT) {    /* OWN bit is 1 */
                DBG_PRINT("NO LAST ONE!!!\n");
                goto rx_end;
            }

            if (rx_sta & RX_DESC_LS) {
                DBG_PRINT("Last Desc(%d):0x%08x\n", ldesc, rx_sta);
                break;
            } else {
                DBG_PRINT("Middle Desc(%d):0x%08x\n", ldesc, rx_sta);
            }
            
            if (fdesc == NEXT_RX(ldesc)) {
                DBG_PRINT("NO LAST ONE!!!\n");
                goto rx_end;
            } else {
                ldesc = NEXT_RX(ldesc);
            }
        }            

        if (bnum > desc_num) {  
            DBG_PRINT("NOT FIND LAST ONE!!!\n");
            goto rx_end;
        }

        DBG_PRINT("Frame's descriptor:%d\n", bnum);

        if (rx_sta & RX_DESC_VLAN) {
            DBG_PRINT("Received VLAN Frame.\n");
        }

        if (rx_sta & RX_DESC_ERR) {
            PMSG(KERN_NOTICE, "RX Desc(%d) has error status(0x%08x).\n", 
                 rxtail, rx_sta);
  
            /* renew the value of error */
            lp->net_stats.rx_errors++;

            /* when descripter error happen, it means rx ring buffer overflow. */
            if (rx_sta & RX_DESC_DE_ERR) {
                lp->net_stats.rx_over_errors++;
            }

            if ((rx_sta & RX_DESC_LE_ERR) && !(rx_sta & RX_DESC_GF_ERR)){
                /* Length Error without Giant Frame setting */
                lp->net_stats.rx_length_errors++;
            }

            if (rx_sta & RX_DESC_CE_ERR) {
                /* CRC Error */
                lp->net_stats.rx_crc_errors++;
            }
            
            if (rx_sta & RX_DESC_OE_ERR) {
                /* Overflow Error */
                lp->net_stats.rx_fifo_errors++;
            }

            goto rx_next;
        }

        /* Get data length from desc without 4bytes FCS */
        datalen = ((rx_sta >> 16) & 0x3FFF) - 4;
        /* DestinationAddr(6)+SourceAddr(6)+Type(2)+NET_IP_ALIGN(2)=16Bytes */
        buflen = datalen + NET_IP_ALIGN;
        DBG_PRINT("Receive Data Length:%d\n", buflen);

        /* alloc a properly sized skbuff */        
        reskb = dev_alloc_skb(buflen);
        if (unlikely(reskb == NULL)) {
            PMSG(KERN_WARNING, 
                 "Failed to alloc skb, Drop! RxDesc(%d)  DataLen(%d)!!!",
                 rxtail, datalen);
            /* renew status struct */
            lp->net_stats.rx_dropped++;
            goto rx_next;
        }
        reskb->dev = dev;

        /* insert 16 bytes befor ethernet header 
         * |pad 2B|desAddr 6B|souAddr 6B|type 2B| */
        skb_reserve(reskb, NET_IP_ALIGN);

        i = bnum;
        while (i) {
            skb = lp->rx_skb[fdesc].skb;

            /* let cpu get ownership of skbuf from DMA */
            dma_sync_single_for_cpu(NULL, lp->rx_skb[fdesc].mapping,
                                    lp->rx_buf_sz, DMA_FROM_DEVICE);

            if (fdesc == ldesc) {
                /* copy data from RX skbuffer to a properly sized skbuffer */
                memcpy(skb_put(reskb, datalen), skb->data, datalen);
            } else {
                /* copy data from RX skbuffer to a properly sized skbuffer */
                memcpy(skb_put(reskb, lp->rx_buf_sz), skb->data, lp->rx_buf_sz);
                datalen -= lp->rx_buf_sz;
            }        
            wmb();

            /* return the ownership of skbuf to DMA */
            dma_sync_single_for_device(NULL, lp->rx_skb[rxtail].mapping,
                                       lp->rx_buf_sz, DMA_FROM_DEVICE);

            fdesc = NEXT_RX(fdesc);
            i--;
        }

        /* determine the packet type which will be used to identify net work */
        reskb->protocol = eth_type_trans(reskb, dev);

        lp->net_stats.rx_packets++;
        lp->net_stats.rx_bytes += reskb->len;

        dev->last_rx = jiffies;
        /* post skb to next network layer */
        netif_rx(reskb);

rx_next:
        for (i =0; i < bnum; i++) {
            /* initializate descriptor */
            lp->rx_ring[rxtail].addr1 = lp->rx_skb[rxtail].mapping;
            lp->rx_ring[rxtail].opts1 = 0;
            if (rxtail == (FMAC3H_RDESC_NUM - 1)) {
                lp->rx_ring[rxtail].opts2 = END_RING | lp->rx_buf_sz;
            } else {
                lp->rx_ring[rxtail].opts2 = lp->rx_buf_sz;
            }
            wmb();
            lp->rx_ring[rxtail].opts1 = OWN_BIT;    /* set own bit to 1 */

            /* get next descriptor */
            rxtail = NEXT_RX(rxtail);
        }

        desc_num -= bnum;        
    }

rx_end:
    /* send RX POLL Demand to wake suspend receive process */ 
    fmac3h_reg_write(lp->base_addr, FMAC3H_RPDR, 1);

    lp->rx_tail = rxtail;

    DBG_DUMP(lp->base_addr, DESC_DUMP, NULL, (int *)lp->rx_ring, 0, 0);

    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_tx -- this function will be called when there is transmit finish 
 *  interrupt(SR[bit0]) or transmit buffer unavallable interrupt(SR[bit2]) 
 *  or transmit underflow interrupt(SR[bit5]) 
 * @dev: target net device structure
 *
 * Description: In this function, the driver will check tx descriptor and 
 *  free the used skbuffer of tx descriptor.
 */
static int fmac3h_tx(struct net_device *dev)
{ 
    struct fmac3h_private *lp;
    int txtail;    /* the head of TX rings whose OWN bit is 1 */
    int txhead;    /* the head of TX ring whose OWN bit is 0 */

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* begin to check the TX descriptor that has done the job */
    txtail = lp->tx_tail;
    txhead = lp->tx_head;

    DBG_DUMP(lp->base_addr, DESC_DUMP, (int *)lp->tx_ring, NULL, 0, 0);

    /* we start check from descriptor(tx_tail), because it first get job
     * and must be first finish the job. */
    while (txtail != txhead) {
        u32 dstatus;    /* the status of descriptor */
            
        dstatus = lp->tx_ring[txtail].opts1;
        if (dstatus & OWN_BIT) {
            DBG_PRINT("the deal of TX interrupt over\n");
            break;
        }

        if (dstatus & TX_DESC_ERR) {
            DBG_PRINT("There is error(0x%08x) in this TX Desc(%d).\n", 
                      dstatus, txtail);

            /* renew the value of errors */
            lp->net_stats.tx_errors++;

            /* When carrier error, late collision error, Excessive collision 
             * error, underflow error happened, the packet is aborted. */ 
            if ((dstatus & TX_DESC_LC_ERR) | (dstatus & TX_DESC_NC_ERR) 
                | ((dstatus & TX_DESC_LCO_ERR) & !(dstatus & TX_DESC_UF_ERR))
                | (dstatus & TX_DESC_EC_ERR)
                | (dstatus & TX_DESC_UF_ERR)) { 

                lp->net_stats.tx_aborted_errors++;

                if ((dstatus & TX_DESC_LC_ERR) | (dstatus & TX_DESC_NC_ERR)) {
                    /* Loss of Carrier | No Carrier */
                    lp->net_stats.tx_carrier_errors++;
                }

                if (dstatus & TX_DESC_UF_ERR) {
                    /* Undeflow Error */
                    lp->net_stats.tx_fifo_errors++;
                } else if (dstatus & TX_DESC_LCO_ERR) {
                    /* Late Collision without Undeflow Error */
                    lp->net_stats.tx_window_errors++;
                }
            }
        }

        lp->net_stats.tx_packets++;
        lp->net_stats.tx_bytes += lp->tx_skb[txtail].skb->len;

        /* renew the status of collisions count */
        if ((dstatus & TX_DESC_EC_ERR)) {
            lp->net_stats.collisions += 16;
        } else {
            lp->net_stats.collisions += (dstatus >> 3) | 0x0000000F;
        }

        if (unlikely(lp->tx_skb[txtail].skb == NULL)) {
            PMSG(KERN_WARNING, 
                 "TX Desc(%d) does not have skbuffer!!!\n", txtail);
            goto next_desc;
        }
 
        /* delete the mapping with skbuffer */
        dma_unmap_single(NULL, lp->tx_skb[txtail].mapping, 
                         lp->tx_skb[txtail].skb->len, DMA_TO_DEVICE);

        /* free the skbuffer */
        dev_kfree_skb_irq(lp->tx_skb[txtail].skb);

next_desc:
        /* clean desciptor status */
        lp->tx_skb[txtail].skb = NULL;
        lp->tx_skb[txtail].mapping = 0;
        lp->tx_ring[txtail].opts1 = 0;
        lp->tx_ring[txtail].opts2 = 0;
        lp->tx_ring[txtail].addr1 = 0;
        lp->tx_ring[txtail].addr2 = 0;

        /* get next descriptor */
        txtail = NEXT_TX(txtail);
    }

    /* renew the flag of ring */
    lp->tx_tail = txtail;

    DBG_DUMP(lp->base_addr, DESC_DUMP, (int *)lp->tx_ring, NULL, 0, 0);

    if (netif_carrier_ok(dev) && 
        netif_queue_stopped(dev) && (TX_RING_AVAIL(lp) > 0)) {
        /* wake queue */
        netif_wake_queue(dev);
        DBG_PRINT("Wake up queue\n");
    }

    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_interrupt -- this function will be called when there is interrupts  
 * @irq: the interrupt number being requested
 * @dev_id: target net device id
 *
 * Description: In this function, the driver will handle the device's 
 *  interrupts.
 */
static irqreturn_t fmac3h_interrupt(int irq, void *dev_id)
{
    struct net_device *dev;
    struct fmac3h_private *lp;
    u32 ints;      /* recived interrupts */
    int ret;

    DBG_FUN_START();

    dev = (struct net_device*)dev_id;
    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "irq(%d) for unknown device!!!\n", irq);
        DBG_FUN_ERREND();
        return IRQ_NONE;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "irq(%d) not for F_MAC3H!!!\n", irq);
        DBG_FUN_ERREND();
        return IRQ_NONE;
    }

    /* get spin lock to refuse others */
    spin_lock_irq(&lp->lock);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    /* get interrupts from SR register */
    ints = fmac3h_reg_read(lp->base_addr, FMAC3H_SR);

    /* clear interrupt */
    fmac3h_reg_write(lp->base_addr, FMAC3H_SR, ints);

    DBG_PRINT("interrupts 0x%08x \n", ints);

    if (!(ints & CHECK_INT)) {    /* did not find any interrupt */
        /* PMSG(KERN_NOTICE, "Do not find interrupt.\n"); */
        spin_unlock_irq(&lp->lock);
        DBG_FUN_END();
        return IRQ_NONE;
    }

    /*    RX stop interrupt. RPS(SR[bit8]) 
     * OR Receive Buffer Unavallable interrupt. RU(SR[bit7]) 
     * OR Receive FIFO overflow interrupt. OVF(SR[bit4]) */
    if (ints & (INT_RX_STOP | INT_RXB_UNAV | INT_RX_OVERF)) {
        if (ints & INT_RX_STOP) {
            /* find out where RX stop by RDLAR register */
            PMSG(KERN_ALERT, 
            "RX stop interrupt. Stop at (0x%08x). Need to restart LAN device!!!\n", 
                 fmac3h_reg_read(lp->base_addr, FMAC3H_RDLAR));
        }

        if (ints & INT_RXB_UNAV) {
            /* tell user what kind of interrupt */
            DBG_PRINT("Receive Buffer Unavallable interrupt\n");
        }

        if (ints & INT_RX_OVERF) {
            /* tell user what kind of interrupt */
            DBG_PRINT("Receive FIFO overflow interrupt\n");
        }

        /* collect status */
        fmac3h_get_stats_from_mfc(lp);

        DBG_PRINT("The deal with RX error interrupt over\n"); 
    }    /* if (ints & (INT_RX_STOP | INT_RXB_UNAV | INT_RX_OVERF)) { */

    /* Receive interrupt. RI(SR[bit6]) */
    if (ints & INT_RX_INT) {
        /* tell user what kind of interrupt */
        DBG_PRINT("Receive interrupt\n");

        /* begin pass received packet and 
         * send RX poll demand to wake suspending RX */
        ret = fmac3h_rx(dev);
        if (ret) {
            DBG_PRINT("The deal with RX interrupt ERROR\n");
        } else {
            DBG_PRINT("The deal with RX interrupt OK\n");
        }

    }    /* if (ints & INT_RX_INT) { */

    /*    TX stop interrupt. TPS(SR[bit1]) 
     * OR Transmit buffer unavallable interrupt. TU(SR[bit2]) 
     * OR Transmit underflow interrupt. UNF(SR[bit5]) */
    if (ints & (INT_TXB_UNAV | INT_TX_UFLOW | INT_TX_STOP)) {
        if (ints & INT_TX_STOP) {
            /* find out where TX stop by TDLAR register */
            PMSG(KERN_ALERT, 
            "TX stop interrupt. stop at (0x%08x). Need to restart LAN device!!!\n", 
                 fmac3h_reg_read(lp->base_addr, FMAC3H_TDLAR));
        }

        if (ints & INT_TXB_UNAV) {
            /* tell user what kind of interrupt */
            DBG_PRINT("Transmit Buffer Unavallable interrupt\n");
        }

        if (ints & INT_TX_UFLOW) {
            /* tell user what kind of interrupt */
            DBG_PRINT("Transmit underflow interrupt\n");
        }

        DBG_DUMP(lp->base_addr, REG_DUMP | DESC_DUMP, (int *)lp->tx_ring, 
                 (int *)lp->rx_ring, 0, 0);

        DBG_PRINT("The deal with TX error interrupt over\n");
    }    /* if (ints & (INT_TXB_UNAV | INT_TX_UFLOW | INT_TX_STOP)) { */

    /* Transmit finish interrupt. TI(SR[bit0]) */
    if (ints & INT_TX_INT) {
        /* tell user what kind of interrupt */
        DBG_PRINT("Transmit finish interrupt\n");

        /* Check every TX descriptor's status and free skbuff */
        ret = fmac3h_tx(dev);
        if (ret) {
            DBG_PRINT("The deal with TX interrupt ERROR\n");
        } else {
            DBG_PRINT("The deal with TX interrupt OK\n");
        } 

    }    /* if (ints & INT_TX_INT) { */

    /* Fatal bus error interrupt. FBE(SR[bit13]) */
    if (ints & INT_BUS_ERR) {
        /* tell user what kind of interrupt */
        DBG_PRINT("Fatal bus error interrupt\n");

        /* collect status */
        fmac3h_get_stats_from_mfc(lp);
    
        /* XXX XXX XXX XXX ? need something hereXXX XXX XXX XXX XXX XXX XXX ? */

        DBG_PRINT("The deal with Fatal bus error interrupt over\n");
    }    /* if (ints & INT_BUS_ERR) { */

    /* free spin lock */
    spin_unlock_irq(&lp->lock);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    DBG_FUN_END();
    return IRQ_HANDLED;
}


/*
 * fmac3h_timeout -- this function will be called when send engine timeout 
 * @dev: target net device structure
 *
 * Description: When send engine timeout this function will be called. In 
 *  this funtion, the driver will reset the whole device and initializate
 *  send and receive engine. 
 */
static void fmac3h_timeout(struct net_device *dev)
{
    struct fmac3h_private *lp;
    u32 val;
    int err;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    /* get spin lock and mask interrupt */
    spin_lock_irq(&lp->lock);

    DBG_DUMP(lp->base_addr, REG_DUMP | DESC_DUMP, (int *)lp->tx_ring, 
             (int *)lp->rx_ring, 0, 0); 

    val = fmac3h_reg_read(lp->base_addr, FMAC3H_SR);
    PMSG(KERN_NOTICE, "Transmit timed out, status 0x%08x\n", val);

    DBG_PRINT("Now TX Desc(0x%08x)\n", 
              fmac3h_reg_read(lp->base_addr, FMAC3H_TDLAR));
    DBG_PRINT("Now RX Desc(0x%08x)\n", 
              fmac3h_reg_read(lp->base_addr, FMAC3H_RDLAR));

    /* Watchdog Timeout will not happen unless there is something wrong
     * in processer. Now we reset the processor to correct the error. 
     * Before begin reset, we stop TX/RX process with stop command. */
    fmac3h_stop_rxtx(lp);

    /* collect the status of transfer befor reset */
    fmac3h_get_stats_from_mfc(lp);

    /* Reset the processer */
    fmac3h_reg_write(lp->base_addr, FMAC3H_BMR, BMR_SOFTWARE_RESET);
    /* have to wait morn than 1600ns before access the register */
    udelay(DELAY_TIME);

    /* free spin lock and enable interrupt */
    spin_unlock_irq(&lp->lock);

    /* initializate the hardware */
    err = fmac3h_init_hw(lp, 1);
    if (err) {
        PMSG(KERN_WARNING, "Failed to initializate device!!!\n");
    }

    /* free the buffer space and initializate the descripor */
    err = fmac3h_clean_ring(lp);
    if (unlikely(err)) {
        PMSG(KERN_ERR, 
             "Failed to free buffer space, didn't wake queue!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    /* initializate the TX/RX ring */
    err = fmac3h_init_rings(lp);
    if (unlikely(err)) {
        PMSG(KERN_ERR, 
             "Failed to initializate TX/RX ring, didn't wake queue!!!\n");
        DBG_FUN_ERREND();
        return;
    }

    DBG_DUMP(lp->base_addr, DESC_DUMP, (int *)lp->tx_ring, 
             (int *)lp->rx_ring, 0, 0);

    /* wake queue to start transfer */
    netif_wake_queue(dev);

    DBG_FUN_END();
    return;
}


/*
 * fmac3h_stats -- get the status of send and receive from mmc registers 
 * @dev: target net device structure
 *
 * Description: In this function, the driver collect the status of send and
 *  receive engine in mmc registers.
 */
static struct net_device_stats *fmac3h_stats(struct net_device *dev)
{
    struct fmac3h_private *lp;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper device!!!\n");
        DBG_FUN_ERREND();
        return NULL;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return NULL;
    }

    spin_lock_irq(&lp->lock);

    if (netif_running(dev)) {
        /* Get status from MFC registers */
        fmac3h_get_stats_from_mfc(lp);
        DBG_DUMP(lp->base_addr, REG_DUMP | DESC_DUMP, (int *)lp->tx_ring, 
                 (int *)lp->rx_ring, 0, 0);
    }

    spin_unlock_irq(&lp->lock);

    DBG_FUN_END();
    return &lp->net_stats;
}


/*
 * fmac3h_get_phymap -- find out all the PHYs on the device
 * @dev: target net device structure
 *
 * Description: Reads the ID registers of the every PHY addr(0-31) on the
 *   device. If the ID<>0xFFFFFFFF then set the bit of phy_map.
 */
static int fmac3h_get_phymap(struct net_device *dev)
{
    struct fmac3h_private *lp;
    int phy_reg;
    int i;
    int ret;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* initializate PHY installation information */
    lp->phy_map = 0;

    for (i = 0; i <= MAX_PHY_ADR; i++) {
        u32 phy_id, temp_id;
        int j;

        /* Read phy ID from PHY's register2&register3 */
        ret = fmac3h_phy_read(dev, i, MII_PHYSID1, &phy_reg);
        if (ret) {
            PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
            DBG_FUN_ERREND();
            return -EACCES;
        }

        if (phy_reg <= 0) {
            DBG_PRINT("There is no PHY in slot(%d).\n", i);
            continue;
        }
        
        phy_id = (phy_reg & 0xFFFF) << 16;    /* get PHYID high */

        ret = fmac3h_phy_read(dev, i, MII_PHYSID2, &phy_reg);
        if (ret) {
            PMSG(KERN_ERR, "Failed to read PHY register!!!\n");
            DBG_FUN_ERREND();
            return -EACCES;
        }

        if (phy_reg <= 0) {
            DBG_PRINT("There is no PHY in slot(%d).\n", i);
            continue;
        }

        phy_id |= (phy_reg & 0xFFFF);    /* get PHYID low */
        DBG_PRINT("PHY%d: ID:0x%08x\n", i, phy_id);

        if (phy_id == 0xFFFFFFFF) {
            DBG_PRINT("There is no PHY in slot(%d).\n", i);
            continue;
        }

        /* determine whether this phy type is known before */
        for (j = 0; (temp_id = known_phy_id[j]) != 0; j++) {
            if (phy_id == temp_id) {
                break;
            }
        }

        if (!temp_id) {
            PMSG(KERN_WARNING, 
                 "PHY(TYPE:0x%08x ADDRESS:%d) is not known before!!!\n", 
                 phy_id, i);
        }

        lp->phy_map |= (1 << i);    /* set the bit of installed PHY */
    }

    DBG_PRINT("phy_map : 0x%08x\n", lp->phy_map);    
    DBG_FUN_END();
    return 0;
}


/*
 * fmac3h_start_xmit -- this function will send the skbuffer 
 * @skb: the socket buffer which need to be sended
 * @dev: target net device structure
 * 
 * Description: This function is the send data engine of the lan device. In
 *  this function, the driver will send the skbuffer by send DMA. 
 */
static int fmac3h_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
    struct fmac3h_private *lp;
    struct fmac3h_desc *desp;  /* the pointer to descriptor */
    u32 ring_num;        /* the number of free descriptors */
    u32 txhead;          /* The first descriptor of this transfer ring */ 
    u32 skblen, flag;
    u32 val;
    dma_addr_t mapping;
    u32 bnum, i;
    unsigned long flags;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    if (unlikely(skb == NULL)) {
        PMSG(KERN_CRIT, "Unaballlable SK Buffer!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    spin_lock_irqsave(&lp->lock, flags);

    DBG_PRINT("FIRST:skb->data(0x%08x)\n", (u32)skb->data);
    
    if (skb->len < ETH_ZLEN) {
        DBG_PRINT("skb_padto\n");
        if (skb_padto(skb, ETH_ZLEN)) {
            DBG_FUN_END();
            return FMAC3H_TX_OK;
        }
    }

    /* set up DMA mappings of sk buffer */
    skblen = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;

    DBG_DUMP(lp->base_addr, REG_DUMP | DESC_DUMP, (int *)lp->tx_ring, 
               NULL, 0, 0);

    /* Use bnum descriptors to send one frame */
    if ((skblen % lp->rx_buf_sz) == 0) {
        bnum = skblen / lp->rx_buf_sz;
    } else {
        bnum = skblen / lp->rx_buf_sz + 1;
    }
    DBG_PRINT("bnum:%d\n", bnum);
   
    /* get the number of free descriptor */
    ring_num = TX_RING_AVAIL(lp);
    if (ring_num < bnum) {
        netif_stop_queue(dev);

        /* drop this packet */
        lp->net_stats.tx_dropped++;

        spin_unlock_irqrestore(&lp->lock, flags);

        PMSG(KERN_ERR, "No space in TX ring, drop packet!!!\n");
        DBG_FUN_ERREND();
        return FMAC3H_TX_BUSY;
    }

    txhead = lp->tx_head;

    for (i = 1; i <= bnum; i++) {
        struct sk_buff *newskb;
   
        desp = &lp->tx_ring[txhead];
        DBG_PRINT("txhead(%d)\n", txhead);
    
        if (i == bnum) { /* The Last one of frame chain */
            newskb = dev_alloc_skb(skblen);
            newskb->dev = dev;
            memcpy(skb_put(newskb, skblen), skb->data, skblen);
        } else {
            newskb = dev_alloc_skb(lp->rx_buf_sz);
            newskb->dev = dev;
            memcpy(skb_put(newskb, lp->rx_buf_sz), skb->data, lp->rx_buf_sz);

            /* decrease the length of SKB */
            skblen -= lp->rx_buf_sz;            
            skb_pull(skb, lp->rx_buf_sz);            
        }

        mapping = dma_map_single(NULL, newskb->data, newskb->len, DMA_TO_DEVICE);
        DBG_PRINT("newskb->data(0x%08x)  mapping(0x%08x)\n", (u32)newskb->data, (u32)mapping);

        /* the settings of descriptor */
        flag = INT_COMPLET | newskb->len;
        
        if (i == 1) {               /* The first one of frame chain */
            flag |= FIRST_SEG;
        } 
        
        if (i == bnum ) {    /* The last one of frame chain  */
            flag |= LAST_SEG;
        }

        if (txhead == (FMAC3H_TDESC_NUM - 1)) {
            flag |= END_RING;    /* this descriptor is the end of ring */
        }

        desp->opts1 = 0;
        desp->opts2 = flag;
        desp->addr1 = mapping;
        desp->addr2 = 0;

        /* save the skb in TX skbuffer array, can be freed after TX */
        lp->tx_skb[txhead].skb = newskb;
        lp->tx_skb[txhead].mapping = mapping;
        wmb();

        /* set OWN bit */
        desp->opts1 = OWN_BIT;
        wmb();

        ring_num--;
        if (ring_num == 0) {
            PMSG(KERN_NOTICE, "No more space in TX ring.\n");
            netif_stop_queue(dev);
        }
    
        /* get the next descriptor */
        txhead = NEXT_TX(txhead);        
    }

    /* renew the flag of TX ring */
    lp->tx_head = txhead;

    /* 9.Set OMR to start TX */
    val = fmac3h_reg_read(lp->base_addr, FMAC3H_OMR);
    if (!(val & START_TX)) {
        val |= START_TX;        /* Start Transmissin */
        fmac3h_reg_write(lp->base_addr, FMAC3H_OMR, val);
    }

    spin_unlock_irqrestore(&lp->lock, flags);

    /* send TX POLL Demand to wake suspend transmit process */ 
    fmac3h_reg_write(lp->base_addr, FMAC3H_TPDR, 1);
    dev->trans_start = jiffies;

    kfree_skb(skb);

    DBG_FUN_END();
    return FMAC3H_TX_OK;
}

#ifdef FMAC3H_MAC_ADDR    /* Set MAC address */
/*
 * fmac3h_set_macaddr -- this function will be called when the device is loaded
 *
 * Description: Set MAC address.
 */
static int fmac3h_set_macaddr(struct net_device *dev)
{
    u32 *laddr_ptr, *haddr_ptr;

    laddr_ptr = (u32*)ioremap((MACADDR_BASE + MACL), sizeof(u32));
    haddr_ptr = (u32*)ioremap((MACADDR_BASE + MACH), sizeof(u32));

    if(laddr_ptr == NULL || haddr_ptr == NULL){
        if(laddr_ptr != NULL)
            iounmap(laddr_ptr);
        if(haddr_ptr != NULL)
            iounmap(haddr_ptr);
        return -ENOMEM;
    }

    dev->dev_addr[0] = ( (*laddr_ptr)        & 0xFF);
    dev->dev_addr[1] = (((*laddr_ptr) >> 8 ) & 0xFF);
    dev->dev_addr[2] = (((*laddr_ptr) >> 16) & 0xFF);
    dev->dev_addr[3] = (((*laddr_ptr) >> 24) & 0xFF);
    dev->dev_addr[4] = ( (*haddr_ptr)        & 0xFF);
    dev->dev_addr[5] = (((*haddr_ptr) >> 8 ) & 0xFF);
    iounmap(laddr_ptr);
    iounmap(haddr_ptr);

    return 0;
}

#endif
/*
 * fmac3h_open -- this function will be called when the device is started 
 * @dev: target net device structure
 *
 * Description: In this function, the driver will get the descriptor and 
 *  buffer space, set the device's registers, registe the interrupt handler
 *  function and start the queue.
 */
static int fmac3h_open(struct net_device *dev)
{
    struct fmac3h_private *lp;
    int err;
    u32 val;
    u32 mac_addrl, mac_addrh;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device, aborting!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* Reset the processer */
    fmac3h_reg_write(lp->base_addr, FMAC3H_BMR, BMR_SOFTWARE_RESET);
    fmac3h_reg_write(lp->base_addr, FMAC3H_MCR, MII_PORT);
    /* have to wait morn than 1600ns before access the register */
    udelay(DELAY_TIME);

    /* get spinlock */
    spin_lock_irq(&lp->lock);

    /* set the generated MAC address in MAC address register */
    mac_addrh = (dev->dev_addr[5] << 8) 
               | dev->dev_addr[4];
    mac_addrl =  (dev->dev_addr[3] << 24) 
               | (dev->dev_addr[2] << 16) 
               | (dev->dev_addr[1] << 8) 
               |  dev->dev_addr[0]; 
    fmac3h_reg_write(lp->base_addr, FMAC3H_MAR0H, mac_addrh);
    fmac3h_reg_write(lp->base_addr, FMAC3H_MAR0L, mac_addrl);

    err = fmac3h_init_rings(lp);
    if (err) {
        PMSG(KERN_ERR, "Failed to initializate DMA descriptor!!!\n");
        goto err_out_cleanring;
    }

    DBG_DUMP(lp->base_addr, DESC_DUMP, (int *)lp->tx_ring, 
               (int *)lp->rx_ring, 0, 0);

    /* Initiate hardware */
    val = (u32)mii_link_ok(&lp->mii);
    if (val) {
        err = fmac3h_init_hw(lp, 1);
        if (err) {
            PMSG(KERN_ERR, "Failed to initializate LAN's hardware!!!\n");
        }
        /* start queue to allow transmit */
        netif_carrier_on(dev);
        netif_start_queue(dev);
    }
    else {
        /* stop network queue */
        netif_stop_queue(dev);
        netif_carrier_off(dev);
    }

    /* Initiate timer function which will check the state of carrier timely */
    init_timer(&lp->timer);
    lp->timer.expires = CHECK_CARRIER_TIME;
    lp->timer.data = (unsigned long)dev;
    lp->timer.function = fmac3h_check_carrier;
    add_timer(&lp->timer);

    spin_unlock_irq(&lp->lock);

    DBG_FUN_END();
    return 0;

err_out_cleanring:
    fmac3h_free_ring(lp);
    spin_unlock_irq(&lp->lock);
    DBG_FUN_ERREND();
    return err;
}


/*
 * fmac3h_close -- this function will be called when the device is stopped 
 * @dev: target net device structure
 *
 * Description: In this function, the driver will stop the device, remove 
 *  the interrupt handler that is registed in fmac3h_open function, free
 *  all of spaces that are malloced in fmac3h_open function. 
 */
static int fmac3h_close(struct net_device *dev)
{
    struct fmac3h_private *lp;
    unsigned long flags;
    int err;

    DBG_FUN_START();

    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper net device, aborting!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* get spinlock and save the current irq mask in flags */
    spin_lock_irqsave(&lp->lock, flags);

    /* delete timer function which will check the state of carrier timely */ 
    del_timer_sync(&lp->timer);

    /* stop hardware first */
    fmac3h_stop_hw(lp);

    /* stop network queue */
    netif_stop_queue(dev);
    netif_carrier_off(dev);

    spin_unlock_irqrestore(&lp->lock, flags);

    /* free descriptor and buffer space */
    err = fmac3h_free_ring(lp);
    if (unlikely(err)) {
        PMSG(KERN_ERR, "Failed to free dscriptor and buffer!!!\n");
        DBG_FUN_ERREND();
        return err;
    }

    DBG_FUN_END();
    return 0;
}

#ifdef CONFIG_NET_POLL_CONTROLLER
static void fmac3h_enable_irq(struct fmac3h_private *lp)
{
    unsigned long flags;

    spin_lock_irqsave(&lp->lock, flags);
    fmac3h_reg_write(lp->base_addr, FMAC3H_IER, INT_ENABLE);
    spin_unlock_irqrestore(&lp->lock, flags);
}

static void fmac3h_disable_irq(struct fmac3h_private *lp)
{
    unsigned long flags;

    spin_lock_irqsave(&lp->lock, flags);
    fmac3h_reg_write(lp->base_addr, FMAC3H_IER, INT_MASK);
    spin_unlock_irqrestore(&lp->lock, flags);
}

static void fmac3h_netpoll(struct net_device *dev)
{
    struct fmac3h_private *lp = NULL;

    lp = netdev_priv(dev);

    /* disable irq */
    fmac3h_disable_irq(lp);

    /* handle irq */
    fmac3h_interrupt(dev->irq, dev);

    /* enable irq */
    fmac3h_enable_irq(lp);


}
#endif


static void fmac3h_set_rx_mode(struct net_device *dev)
{
}

/*
static int fmac3h_set_mac_addr(struct net_device *dev, void *addr)
{
    return 0;
}*/

static int fmac3h_validate_addr(struct net_device *dev)
{
    return 0;
}

static const struct net_device_ops fmac3h_netdev_ops = {
    .ndo_open               = fmac3h_open,
    .ndo_stop               = fmac3h_close,
    .ndo_start_xmit         = fmac3h_start_xmit,
    .ndo_get_stats          = fmac3h_stats,
    .ndo_do_ioctl           = fmac3h_ioctl,
    .ndo_tx_timeout         = fmac3h_timeout,
    .ndo_change_mtu         = fmac3h_change_mtu,
    .ndo_set_multicast_list = fmac3h_set_rx_mode,
    .ndo_set_mac_address    = eth_mac_addr,  // fmac3h_set_mac_addr,
    .ndo_validate_addr      = fmac3h_validate_addr,
#ifdef CONFIG_NET_POLL_CONTROLLER
    .ndo_poll_controller    = fmac3h_netpoll,
#endif
};

/*
 * fmac3h_probe -- this function will be called when the device is loaded 
 * @pdev: target platform device structure
 *
 * Description: In this function, the driver will probe for its device and
 *  its hardware location(I/O ports and IRQ line). 
 */
static int __init fmac3h_probe(struct platform_device *pdev)
{
    struct net_device *dev;
    struct fmac3h_private *lp;
    int i, err;

    DBG_FUN_START();

    if (unlikely(pdev == NULL)) {
        PMSG(KERN_CRIT, 
             "Can not find proper platform device, aborting!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* allocate a new ethernet device */
    DBG_PRINT("alloc etherdev\n");
    dev = alloc_etherdev(sizeof(struct fmac3h_private));
    if (!dev) {
        PMSG(KERN_ERR, "Etherdev alloc failed, aborting!!!\n");
        DBG_FUN_ERREND();
        return -ENOMEM;
    }
    SET_NETDEV_DEV(dev, &pdev->dev);

    /* get private struct's pointer */
    lp = netdev_priv(dev);
    lp->dev = dev;

    /* registe ethernet device */
    platform_set_drvdata(pdev, dev);

    /* initializate spin lock */
    spin_lock_init(&lp->lock);

    /* fill structure net_device in defaults        */
    ether_setup(dev);
    /* fill structure net_device by FMAC3H function */
    dev->netdev_ops = &fmac3h_netdev_ops;
    dev->ethtool_ops = &fmac3h_ethtool_ops;
    dev->watchdog_timeo = 1 * 500;         /* 500jiffy = 5s       */
    dev->irq = pdev->resource[1].start;    /* set IRQ             */
    /* Can not handle VLAN packets */
    dev->features = NETIF_F_VLAN_CHALLENGED;
    /* dev->flags is seted in ether_setup()                       */
    /* MTU should be changed by system                            */
    /* dev->mtu = JUMBO_FRAME_MTU; */      /* support Jumbo Frame */

#ifdef FMAC3H_MAC_ADDR    /* Set MAC address */
    if(MACADDR_BASE){
        err = fmac3h_set_macaddr(dev);
        if(err){
            PMSG(KERN_ERR, "Failed to set Ether MAC address.\n");
        }
        PMSG(KERN_INFO, "Static MAC address: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n", 
             dev->dev_addr[0], dev->dev_addr[1],
             dev->dev_addr[2], dev->dev_addr[3],
             dev->dev_addr[4], dev->dev_addr[5]);
    } else {
        /* When not get the MAC address from register, 
        * generate random MAC address */
        if (is_zero_ether_addr(dev->dev_addr)) {
            /* generate MAC address */
#if 1 /* Mod for #28357 起動引数からMACアドレスを取得 by hiroshi.hasegawa@pixela.jp 2011.09.08 */
			macadrs_setup_from_cmdline(dev->dev_addr);
#else
            random_ether_addr(dev->dev_addr);
            PMSG(KERN_INFO, "Random MAC address: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n", 
                 dev->dev_addr[0], dev->dev_addr[1],
                 dev->dev_addr[2], dev->dev_addr[3],
                 dev->dev_addr[4], dev->dev_addr[5]);
#endif
        }
    }
#endif
    if (is_zero_ether_addr(dev->dev_addr)) {
        /* generate MAC address */
#if 1 /* Mod for #28357 起動引数からMACアドレスを取得 by hiroshi.hasegawa@pixela.jp 2011.09.08 */
        macadrs_setup_from_cmdline(dev->dev_addr);
#else
        random_ether_addr(dev->dev_addr);
        PMSG(KERN_INFO, "Random MAC address: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n", 
             dev->dev_addr[0], dev->dev_addr[1],
             dev->dev_addr[2], dev->dev_addr[3],
             dev->dev_addr[4], dev->dev_addr[5]);
#endif
    }

    /* Get F_MAC3H register memory space */
    lp->regs = request_mem_region(pdev->resource[0].start,
                            pdev->resource[0].end - pdev->resource[0].start + 1,
                            DRV_NAME);
    if (lp->regs == NULL) {
        PMSG(KERN_ERR, "Can not alloc register memory space, aborting!!!\n");
        err = -ENOMEM;
        goto err_out_freedev;
    }
    /* map register address to virtual address space */
    lp->base_addr = ioremap(pdev->resource[0].start,
                           pdev->resource[0].end - pdev->resource[0].start + 1);
    if (lp->base_addr == NULL) {
        PMSG(KERN_ERR, "Failed to ioremap ethernet registers!!!\n");
        err = -ENOMEM;
        goto err_out_freesource;
    }

    /* Reset the processer */
    fmac3h_reg_write(lp->base_addr, FMAC3H_BMR, BMR_SOFTWARE_RESET);
    /* have to wait morn than 1600ns before access the register */
    udelay(DELAY_TIME);

    DBG_DUMP(lp->base_addr, REG_DUMP, NULL, NULL, 0, 0);

    /* get MDC clock from SYS_CLOCK */
    if (SYS_CLK  <  20) {
        PMSG(KERN_ERR, "Unavallable SYS_CLK(%d), aborting!!!\n", SYS_CLK);
        err = -EINVAL;
        goto err_out_unmap;
    } else if (SYS_CLK < 35) {
        lp->mdc_clk = 2;
    } else if (SYS_CLK < 60) {
        lp->mdc_clk = 3;
    } else if (SYS_CLK < 100) {
        lp->mdc_clk = 0;
    } else if (SYS_CLK < 150) {
        lp->mdc_clk =1;
    } else {
        PMSG(KERN_ERR, "Unavallable SYS_CLK(%d), aborting!!!\n", SYS_CLK);
        err = -EINVAL;
        goto err_out_unmap;
    }
    DBG_PRINT("SYS CLK:%d    MDC CLK:%d\n", SYS_CLK, lp->mdc_clk);

    err = fmac3h_get_phymap(dev);    /* get phy install information */
    if (err) {
        PMSG(KERN_ERR, "Failed to find PHY, aborting!!!\n");
        err = -EINVAL;
        goto err_out_unmap;
    }

    if (!lp->phy_map) {
        PMSG(KERN_ALERT, "There is no PHY, aborting!!!\n");
        err = -EINVAL;
        goto err_out_unmap;
    }

    /* use the PHY who's ID is the minimum one of all */
    for (i = 0; i <= MAX_PHY_ADR; i++) {
        if (lp->phy_map & (1 << i)) {
            lp->mii.phy_id = i;
            break;
        }
    }
    lp->mii.dev = dev;
    lp->mii.advertising = ADVERTISED_10baseT_Half | ADVERTISED_10baseT_Full 
                         | ADVERTISED_100baseT_Half | ADVERTISED_100baseT_Full
                         | ADVERTISED_Autoneg | ADVERTISED_MII;
    lp->mii.full_duplex = 1;    /* Full duplex */
    lp->mii.force_media = 0;    /* Auto negotiation */
    lp->mii.supports_gmii = 0;  /* not support GMII */
    lp->mii.mdio_read = fmac3h_mii_read;
    lp->mii.mdio_write = fmac3h_mii_write;
    lp->mii.phy_id_mask = 0x1f;
    lp->mii.reg_num_mask = 0x1f;

    /* Descriptor and Buffer initializate */
    lp->rx_ring = (struct fmac3h_desc *)dma_alloc_coherent(NULL, 
                               FMAC3H_DESC_BYTES, &lp->ring_dma, GFP_KERNEL);
    if (lp->rx_ring == NULL) {
        PMSG(KERN_ERR, "Failed to alloc DMA descriptor memory!!!\n");
        err = -ENOMEM;
        goto err_out_unmap;
    }
    lp->tx_ring = &lp->rx_ring[FMAC3H_RDESC_NUM];
    memset(lp->tx_ring, 0, sizeof(struct fmac3h_desc) * FMAC3H_TDESC_NUM);
    lp->tx_ring[FMAC3H_TDESC_NUM-1].opts2 = END_RING;
    lp->rx_buf_sz = BUFFER_SIZE;    /* 1520 bytes */

    /* Install the interrupt handler */
    err = request_irq(dev->irq, fmac3h_interrupt, 0, dev->name, dev);
    if (err) {
        PMSG(KERN_CRIT, "Failed to register IRQ%d, ERRNO:%d!!!\n", 
               dev->irq, err);
        goto err_out_freerxbuf;
    }

    err = register_netdev(dev);
    if (err) {
        PMSG(KERN_CRIT, "Failed to register netdev. ERRNO:%d!!!\n", err);
        goto err_out_freeirq;
    }

    /* initializate the hardware */
    fmac3h_init_hw(lp, 0);

    PMSG(KERN_INFO, "FMAC3H Driver Loaded.\n");

    DBG_FUN_END();
    return 0;

err_out_freeirq:
    free_irq(dev->irq, dev);
err_out_freerxbuf:
    dma_free_coherent(NULL, FMAC3H_DESC_BYTES, lp->rx_ring, lp->ring_dma);
err_out_unmap:
    iounmap(lp->base_addr);
err_out_freesource:
    release_resource(lp->regs);
err_out_freedev:
    free_netdev(dev);
    platform_set_drvdata(pdev, NULL);
    DBG_FUN_ERREND();
    return err;
}


/*
 * fmac3h_remove -- remove the device from the system 
 * @pdev: target platform device structure
 *
 * Description: Free the resource that was requested in probe and open 
 *  process, if it has not been freed.  
 */
static int fmac3h_remove(struct platform_device *pdev)
{
    struct net_device *dev;
    struct fmac3h_private *lp;

    DBG_FUN_START();
    if (unlikely(pdev == NULL)) {
        PMSG(KERN_CRIT, 
             "Can not find proper platform device, aborting!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    dev = platform_get_drvdata(pdev);
    if (unlikely(dev == NULL)) {
        PMSG(KERN_CRIT, "Can not find proper device, aborting!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    lp = netdev_priv(dev);
    if (unlikely(lp == NULL)) {
        PMSG(KERN_CRIT, "Can not get the device!!!\n");
        DBG_FUN_ERREND();
        return -ENODEV;
    }

    /* remove the interrupt handler */
    free_irq(dev->irq, dev);

    /* Force Down */
    unregister_netdev(dev);

    /* Unmap a ioremap()ed region */
    if (lp->base_addr != NULL) {
        DBG_PRINT("unmap memory\n");
        iounmap(lp->base_addr);
    }

    /* release a previously reserved resource in probe() */
    if (lp->regs != NULL) {
        DBG_PRINT("release resource\n");
        release_resource(lp->regs);
        kfree(lp->regs);
    }

    /* delete driver data */
    platform_set_drvdata(pdev, NULL);

    /* free descriptor space */
    dma_free_coherent(NULL, FMAC3H_DESC_BYTES, lp->rx_ring, lp->ring_dma);
    lp->rx_ring = NULL;
    lp->tx_ring = NULL;
    lp->ring_dma = 0;

    /* free the allocated net device */
    free_netdev(dev);

    DBG_FUN_END();
    return 0;

}

#if 1 /* Add for #28357 起動引数からMACアドレスを取得 by hiroshi.hasegawa@pixela.jp 2011.09.08 */
static int macadrs_setup(char *s)
{
	cmdline = s;
	return 1;
}

__setup("mac_adrs=", macadrs_setup);
#endif

/* initializate the platform_driver structure */
static struct platform_driver fmac3h_driver = {
    .probe     = fmac3h_probe,
    .remove    = fmac3h_remove,
    .driver    = {
        .name  = DRV_NAME,
        .owner = THIS_MODULE,
    },
};


static int __init fmac3h_module_init(void)
{
    return platform_driver_register(&fmac3h_driver);
}


static void __exit fmac3h_module_exit(void)
{
    platform_driver_unregister(&fmac3h_driver);
}


module_init(fmac3h_module_init)
module_exit(fmac3h_module_exit)


MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Fujitsu F_MAC3H Ethernet driver");
