/*
 *  drivers/video/mb86h61fb.c
 *  mb86h61fb: a linux frame buffer driver for Fujitsu MB86H61
 *
 * Copyright (C) 2009 FUJITSU MICROELECTRONICS LIMITED
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/fb.h>
#include <mach/hardware.h>
#include <asm/io.h>
#include <mach/mb86hxx.h>

#include "fapex/sys_config.h"
#include "fapex_fdlviout.h"
#include "fapex_func_ext.h"

extern FAPEX_VISCALE_FuncStrT viscale_funcStr;
extern FAPEX_VIENC_FuncStrT     vienc_funcStr;
extern unsigned long io_phys_to_virt(unsigned long x);

/*
 * Here we define the default structs fb_fix_screeninfo and fb_var_screeninfo
 * if we don't use modedb. If we do use modedb see xxxfb_init how to use it
 * to get a fb_var_screeninfo. Otherwise define a default var as well. 
 */

static struct fb_fix_screeninfo mb86h61fb_fix = {
	.id             = "MB86H61 OSD1",

	.smem_start     = FAPEX_SYS_MMAP_OSD1_PLANE_START,
	.smem_len       = FAPEX_SYS_MMAP_OSD1_PLANE_BYTES,	
	.type           = FB_TYPE_PACKED_PIXELS,
	.xpanstep       = 0,
	.ypanstep       = 1,
	.ywrapstep      = 0,
	.mmio_start     = 0,
	.mmio_len       = 0,
	.accel          = FB_ACCEL_FUJITSU_MB86H61,
};
static struct fb_fix_screeninfo mb86h61fb_fix1 = {
	.id             = "MB86H61 OSD2",
	.smem_start     = FAPEX_SYS_MMAP_OSD2_PLANE_START,
	.smem_len       = FAPEX_SYS_MMAP_OSD2_PLANE_BYTES,
	.type           = FB_TYPE_PACKED_PIXELS,
	.xpanstep       = 0,
	.ypanstep       = 1,
	.ywrapstep      = 0,
	.mmio_start     = 0,
	.mmio_len       = 0,
	.accel          = FB_ACCEL_NONE,
};
static struct fb_var_screeninfo mb86h61fb_var_default = {
	.xres           = 720,
	.yres           = 480,
	.xres_virtual   = 720,
	.yres_virtual   = 480*2,
	.bits_per_pixel = 32,
	.red            = {16, 8, 0},
	.green          = { 8, 8, 0},
	.blue           = { 0, 8, 0},
	.transp         = {24, 8, 0},
	.activate       = FB_ACTIVATE_NOW,
	.vmode          = FB_VMODE_NONINTERLACED,
};

static struct platform_device mb86h61fb_device;

static char *mode = NULL;	/* user-specified resolution */
static int   bpp  = 0;		/* user-specified bits per pixel */
static char *mode1 = NULL;	/* user-specified resolution */
static int   bpp1  = 0;		/* user-specified bits per pixel */
module_param(mode, charp,  S_IRUGO);
module_param(bpp,  uint,   S_IRUGO);
module_param(mode1, charp,  S_IRUGO);
module_param(bpp1,  uint,   S_IRUGO);

static int outputWidth;		/* HD output X */
static int outputHeight;	/* HD output Y */
static int progressive;		/* 1:progressive, 0:interlaced */
static int open_cnt = 0;

/*
 * Frame buffer operations:
 *	mb86h61fb_open()
 *	mb86h61fb_release()
 *	mb86h61fb_check_var()
 *	mb86h61fb_set_par()
 *	mb86h61fb_setcolreg()
 *	mb86h61fb_blank()
 */

#define H61FB_NAME       "mb86h61fb"
#define H61FB_MODSTR_LEN  13	/* XXXXxYYYY-BB */
#define H61FB_CMAP_LEN    256

#define H61FB0             0
#define H61FB1             1

#define REG_OSDMODE	  		0x4044  /* VO_OSDMODE   */
#define REG_OSD1_A1       	0x4048  /* VO_OSDALPHA1 */
#define REG_OSD1_A2       	0x404C  /* VO_OSDALPHA2 */
/*H60:128 aligned,H61:8 aligned*/
#define ROUND128(xres)         (((xres) + (8-1)) & ~(8-1))
#define LINE_LENGTH(xres, bpp) ROUND128(((xres) * (bpp) + 7) / 8)
#define H61FB_XRES(xres, bpp)  (LINE_LENGTH(xres, bpp) * 8 / bpp)
#define H61FB_YRES(yres)       (((yres) + (2-1)) & ~(2-1))

#define H61FB_CLEAR_FB    0     /* clear buffer memory in open() */
#define H61FB_ENABLE      1     /* enable fb after resize in open() */
#define H61FB_X11_HACK    1     /* work arounds for X11 */
#define H61FB_ALPHA       1	/* enable pixel-based alpha blending */

#define DEBUG            	0
#if DEBUG
#define ASSERT(cond, string) BUG_ON(!(cond))
#define DBG(format, arg...)  my_dbg(&(mb86h61fb_device.dev), format, ## arg)
#define my_dbg(dev, format, arg...)            \
        dev_printk(KERN_DEBUG , dev , format , ## arg)
#else
#define ASSERT(cond, string)  ;
#define DBG(format, arg...)   ;
#endif /* DEBUG */

//#define DBG  printk

static void mb86h61fb_rgb(
    int bpp, struct fb_var_screeninfo *var,FAPI_SYS_HandleT handle)
{
	var->red.msb_right    = 0;
	var->green.msb_right  = 0;
	var->blue.msb_right   = 0;
	var->transp.msb_right = 0;
	switch (bpp) {
	case 4:	/* CLUT 4 */
		var->red.offset    = 16;
		var->red.length    = 8;
		var->green.offset  = 8;
		var->green.length  = 8;
		var->blue.offset   = 0;
		var->blue.length   = 8;
		var->transp.offset = 24;
		var->transp.length = 0;
		break;
	case 8:	/* CLUT 8 */
		var->red.offset    = 16;
		var->red.length    = 8;
		var->green.offset  = 8;
		var->green.length  = 8;
		var->blue.offset   = 0;
		var->blue.length   = 8;
		var->transp.offset = 24;
		var->transp.length = 0;
		break;
	case 16: /* 1555 (ARGB=1555) */
		var->red.offset    = 10;
		var->red.length    = 5;
		var->green.offset  = 5;
		var->green.length  = 5;
		var->blue.offset   = 0;
		var->blue.length   = 5;
		/* DirectFB/df_dok do not seem to accept pixel based alpha.*/
		if(0 != var->transp.length)
		{
			var->transp.offset = 15;
			var->transp.length = 1;
		}
		break;
	case 32: /* 32_24BIT (ARGB=8888) */
		var->red.offset    = 16;
		var->red.length    = 8;
		var->green.offset  = 8;
		var->green.length  = 8;
		var->blue.offset   = 0;
		var->blue.length   = 8;
		/* DirectFB/df_dok do not seem to accept pixel based alpha.*/
		if(0 != var->transp.length)
		{
			var->transp.offset = 24;
			var->transp.length = 8;
		}
		break;
	}

	/* DirectFB/df_dok do not seem to accept pixel based alpha.*/
        if(0 != var->transp.length)
        {
	    	viscale_funcStr.Alpha(handle,FAPI_VISCALE_PARAMETER_OFF,0xFF);
        }
        else
        {
	    	viscale_funcStr.Alpha(handle,FAPI_VISCALE_PARAMETER_ON,0xFF);
	    	viscale_funcStr.Alpha(handle,FAPI_VISCALE_PARAMETER_SET,0xFF);
        }

        DBG("mb86h61fb_rgb() 2: bpp=%d\n"
            "\tred   =(%u,%u,%u)\n"
            "\tgreen =(%u,%u,%u)\n"
            "\tblue  =(%u,%u,%u)\n"
            "\talpha =(%u,%u,%u)\n",
            bpp,
            var->red.offset,var->red.length,var->red.msb_right,
            var->green.offset,var->green.length,var->green.msb_right,
            var->blue.offset,var->blue.length,var->blue.msb_right,
            var->transp.offset,var->transp.length,var->transp.msb_right);
	return;
}

/**
 *      xxxfb_open - Optional function. Called when the framebuffer is
 *                   first accessed.
 *      @info: frame buffer structure that represents a single frame buffer
 *      @user: tell us if the userland (value=1) or the console is accessing
 *             the framebuffer. 
 *
 *     This function is the first function called in the framebuffer api.
 *      Usually you don't need to provide this function. The case where it 
 *      is used is to change from a text mode hardware state to a graphics
 *      mode state. 
 *
 *      Returns negative errno on error, or zero on success.
 *
 *      MB86H61:
 *      Input window size should be equal to Input picture size.
 *      Input picture size is set to HD output resolution
 *      unless user specifies them explicitly.
 *
 *      Frame buffer			FAPI VISCALE
 *      bits_per_pixel  visual		osdLayout osdMode
 *      --------------  ------		--------- -------
 *      4		PSEUDOCOLOR	ARGB	  CLUT_4BIT
 *      8		PSEUDOCOLOR	ARGB	  CLUT_8BIT
 *      16		TRUECOLOR	ARGB	  1555
 *      32		TRUECOLOR	ARGB	  32_24BIT
 */

static int mb86h61fb_open(struct fb_info *info, int user)
{
	FAPI_SYS_HandleT handle;
	FAPI_VISCALE_OpenParamsT params;
	int inputWidth = 0, inputHeight = 0, inputBpp = 0;
	FAPI_VISCALE_SystemModeEnumT sysMode;
	char modeString[H61FB_MODSTR_LEN], *token;
	FAPI_VISCALE_ScalerParamsT scalerParams;
	int retval = 0;
	struct page * p,*p0;
	
	DBG("mb86h61fb_open(),count(%d)\n", open_cnt);
	if (open_cnt > 0)
	{
		open_cnt++;
		return 0;
	}
	/* Open the device */
	params.version = FAPI_VISCALE_VERSION;
	
	switch(info->node)
	{
		
		case H61FB0:
			params.scalerType = FAPI_VISCALE_SCALER_OSD1;	
			
			if(mb86h61fb_fix.smem_start==NULL)
			{
	
			}
			else
			{
				info->screen_base=IO_ADDRESS(mb86h61fb_fix.smem_start);
			}
			info->fix = mb86h61fb_fix;		
			memset((char*)info->screen_base,0,info->fix.smem_len);
			break;
		case H61FB1:
			params.scalerType = FAPI_VISCALE_SCALER_OSD2;  	
			
			if(mb86h61fb_fix1.smem_start==NULL)
			{
		
			}
			else
			{
				info->screen_base=IO_ADDRESS(mb86h61fb_fix1.smem_start);
			}
			info->fix = mb86h61fb_fix1;		
			memset((char*)info->screen_base,0,info->fix.smem_len);
			break;
			default:
				return 0;
				
	}
	params.encoderType = FAPI_VISCALE_ENCODER_NTSC;

	retval = 0;
	handle = viscale_funcStr.Open(&params, &retval);
	DBG("VISCALE Open returned 0x%x.\n", retval);
	if (handle != NULL)
		info->par = (void *)handle;
	else if ((retval == FAPI_VISCALE_ERR_NOT_INITIALIZED) ||
		 (retval == FAPI_VISCALE_ERR_EXCEEDED_MAX_HANDLES)) {
		return -ENODEV;
	} else /* retval != FAPI_OK */
		return -EINVAL;

	/* Bind VIENC handle and get HD output size */
	retval = FAPEXFDLVIOUT_SetLinuxFB((void *)handle);
	DBG("FAPEXFDLVIOUT returned %d.\n", retval);
	if (retval == FAPEX_VIOUT_ERR_INVALID_STATUS) {
		retval = -ENODEV;
		goto open_err;
	} else if (retval != FAPI_OK) {
		retval = -EINVAL;
		goto open_err;
	}
	sysMode = viscale_funcStr.GetSystemType(handle);
	DBG("VISCALE GetSystemType returned 0x%x.\n", sysMode);
	switch (sysMode) {
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x2397P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x24P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x25P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x2997P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x30P:
		outputWidth  = 1920;
		outputHeight = 540;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x50I:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x5994I:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x60I:
		outputWidth  = 1920;
		outputHeight = 1080;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x2397P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x24P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x25P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x2997P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x30P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x50P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x5994P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x60P:
		outputWidth  = 1280;
		outputHeight = 720;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x576x50P:
		outputWidth  = 720;
		outputHeight = 576;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x576x50I:
		outputWidth  = 720;
		outputHeight = 576;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x480x5994P:
		outputWidth  = 720;
		outputHeight = 480;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x480x5994I:
		outputWidth  = 720;
		outputHeight = 480;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_LCD_640x480x50P_LQ080V3DG01:
		outputWidth  = 640;
		outputHeight = 480;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_LCD_1024x768x50P_SVA150XG10TB:
		outputWidth  = 1024;
		outputHeight = 768;
		progressive  = 1;
		break;
	default:
		retval = -EINVAL;
		goto open_err;
	}

	/* Parse user-specified mode */
	if (mode) {
		strncpy(modeString, mode, H61FB_MODSTR_LEN-1);
		modeString[H61FB_MODSTR_LEN-1] = '\0';
		inputWidth = simple_strtoul(modeString, &token, 10);
		if (*token != 'x') {
			retval = -EINVAL;
			goto open_err;
		}
		++token;
		inputHeight = simple_strtoul(token, &token, 10);
		if (*token == '-') {
			++token;
			inputBpp = simple_strtoul(token, &token, 10);
			if (*token != '\0') {
				retval = -EINVAL;
				goto open_err;
			}
			if (inputBpp == 0) {
				retval = -EINVAL;
				goto open_err;
			}
		} else if (*token != '\0') {
			retval = -EINVAL;
			goto open_err;
		}
	} else {
		inputWidth = outputWidth;
		inputHeight = outputHeight;
	}
	if (bpp)
		inputBpp = bpp;
	else if (inputBpp == 0)
		inputBpp = 16;
	if ((inputBpp != 4)  && (inputBpp != 8) &&
	    (inputBpp != 16) && (inputBpp != 32)) {
		retval = -EINVAL;
		goto open_err;
	}
	DBG("mb86h61fb_open(),input(%d,%d, %d), output(%d,%d,%d)\n",
		inputWidth, inputHeight, inputBpp,
		outputWidth, outputHeight, progressive);
	ASSERT(inputWidth != 0,  "inputWidth != 0");
	ASSERT(inputHeight != 0, "inputHeight != 0");
	ASSERT(inputBpp != 0,    "inputBpp != 0");
	/*what does it mean?*/
	inputWidth=outputWidth  ;    
	inputHeight=outputHeight  ;  
	inputBpp=32;
	if(H61FB1 == info->node)
	{
		inputBpp=16;
	}
	DBG("mb86h61fb_open(),info->screen_base is 0x%08x\n", (unsigned int)info->screen_base);
	/* Set up info->var */
	info->var = mb86h61fb_var_default;
	info->var.xres_virtual = H61FB_XRES(inputWidth, inputBpp);
	info->var.yres_virtual = 2*H61FB_YRES(inputHeight);
	info->var.bits_per_pixel = inputBpp;
	info->var.xres = inputWidth;
	info->var.yres = inputHeight;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.vmode = (progressive == 1) ? FB_VMODE_NONINTERLACED:
					       FB_VMODE_INTERLACED;
	mb86h61fb_rgb(info->var.bits_per_pixel, &(info->var),info->par);
	DBG("mb86h61fb_open(),(xres,yres)=(%d,%d),(xres_virt,yres_virt)=(%d,%d)\n",info->var.xres,info->var.yres,info->var.xres_virtual,info->var.yres_virtual);
	/* Set up info->fix */
	info->fix.line_length =
		LINE_LENGTH(info->var.xres_virtual, info->var.bits_per_pixel);
	if (info->fix.line_length * info->var.yres_virtual > 
		info->fix.smem_len) {
		retval = -ENOMEM;
		goto open_err;
	}
#if H61FB_CLEAR_FB
	memset(info->screen_base, 0x0,
	       info->fix.line_length * info->var.yres_virtual);
#endif /* H61FB_CLEAR_FB */
	switch (info->var.bits_per_pixel) {
	case 4:
	case 8:
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
		break;
	case 16:
	case 32:
		info->fix.visual = FB_VISUAL_TRUECOLOR;
		break;
	}

	/* Start the device */
#if H61FB_ENABLE
	scalerParams.layerEnable = 0;
#else
	scalerParams.layerEnable = 1;
#endif /* H61FB_ENABLE */
	scalerParams.osdLayout = FAPI_VISCALE_OSD_LAYOUT_ARGB;
	switch (info->var.bits_per_pixel) {
	case 4:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_CLUT_4BIT;
		break;
	case 8:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_CLUT_8BIT;
		break;
	case 16:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_5551;
		break;
	case 32:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_32_24BIT;
		break;
	}
	scalerParams.picBase = (unsigned int)info->screen_base;
	DBG("mb86h61fb_open(),scalerParams.osdLayout :0x%x\n",scalerParams.osdLayout);
	DBG("mb86h61fb_open(),scalerParams.osdMode :0x%x\n",scalerParams.osdMode);
	DBG("mb86h61fb_open(),scalerParams.picBase :0x%x\n",scalerParams.picBase);
	scalerParams.frameField = progressive;
#if H61FB_ALPHA
	scalerParams.useGlobalAlpha = 0;
#else
	scalerParams.useGlobalAlpha = 1;
	scalerParams.globalAlpha = 255;
#endif /* H61FB_ALPHA */
	memset(scalerParams.colourLut, 0, sizeof(scalerParams.colourLut));
	scalerParams.inputPicWidth = info->var.xres_virtual;
	scalerParams.inputPicHeight = info->var.yres_virtual;
	scalerParams.inputWindowX0 = 0;
	scalerParams.inputWindowY0 = 0;
	scalerParams.inputWindowWidth = info->var.xres;
	scalerParams.inputWindowHeight = info->var.yres;
	scalerParams.outputWindowX0 = 0;
	scalerParams.outputWindowY0 = 0;
	scalerParams.outputWindowWidth = outputWidth;
	scalerParams.outputWindowHeight = outputHeight;
	/*used in the version newer then v1.1 such as v4576, the following line should be open in the future. 
		return -30104 (0xffff8a68) 
		viscale module
	*/
	scalerParams.endianness = FAPEX_VISCALE_ENDIAN_DEFAULT;
	scalerParams.pixelOrder = FAPEX_VISCALE_PIXEL_ORDER_DEFAULT;
	retval = viscale_funcStr.Start(handle, &scalerParams);
	
	DBG("VISCALE Start returned 0x%x.\n", retval);
	ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
	      "retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
	if (retval != FAPI_OK) {
		retval = -EINVAL;
		goto open_err;
	}

#if H61FB_ALPHA
	iowrite32(0xff,  MB86HXX_VIDOUT_VA_BASE + REG_OSD1_A1);/*0x0-->0xff, by Thomas*/
	iowrite32(0xff, MB86HXX_VIDOUT_VA_BASE + REG_OSD1_A2);
#endif /* H61FB_ALPHA */
	if(H61FB1 != info->node)
	{
		retval = viscale_funcStr.Resize(handle, &scalerParams);
		ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
		      "retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
		DBG("VISCALE Resize returned %d.\n", retval);
		if (retval != FAPI_OK) {
			if (retval == FAPI_VISCALE_ERR_OPERATION_TIMEOUT)
				retval = -ENODEV;
			else if (retval != FAPI_OK)
				retval = -EINVAL;
			goto open_err;
		}
	}
#if H61FB_ENABLE
	retval = viscale_funcStr.EnableDisable(handle,
					FAPI_VISCALE_PARAMETER_ON);
	ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
	      "retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
	if (retval != FAPI_OK) {
		retval = -ENODEV;
		goto open_err;
	}
#endif /* H61FB_ENABLE */

	if (retval == 0)
		open_cnt++;

	return retval;

open_err:
	if (handle)
		viscale_funcStr.Close(handle);
	return retval;
}

/**
 *      xxxfb_release - Optional function. Called when the framebuffer 
 *                      device is closed. 
 *      @info: frame buffer structure that represents a single frame buffer
 *      @user: tell us if the userland (value=1) or the console is accessing
 *             the framebuffer. 
 *      
 *      Thus function is called when we close /dev/fb or the framebuffer 
 *      console system is released. Usually you don't need this function.
 *      The case where it is usually used is to go from a graphics state
 *      to a text mode state.
 *
 *      Returns negative errno on error, or zero on success.
 */

static int mb86h61fb_release(struct fb_info *info, int user)
{
	FAPI_SYS_HandleT handle = (FAPI_SYS_HandleT)info->par;
	int retval = 0,offs;
	struct page *page;
	DBG("mb86h61fb_release()\n");
	ASSERT(handle != 0, "handle != 0");
	DBG("close cnt(%d)\n", open_cnt);

	if (open_cnt > 0)
		open_cnt--;
	if (open_cnt > 0)
		return 0;
	retval = viscale_funcStr.Stop(handle);
	if (retval != FAPI_OK)
	{
		DBG("%s: viscale_funcStr.Stop failed\n",__FUNCTION__);	
		return  -ENODEV;
	}
	retval = viscale_funcStr.Close(handle);
	if (retval != FAPI_OK)
		retval = -ENODEV;
	else {
		info->par = NULL;
		retval = 0;
	}

	return retval;
}

/**
 *      xxxfb_check_var - Optional function. Validates a var passed in. 
 *      @var: frame buffer variable screen structure
 *      @info: frame buffer structure that represents a single frame buffer 
 *
 *      Checks to see if the hardware supports the state requested by
 *      var passed in. This function does not alter the hardware state!!! 
 *      This means the data stored in struct fb_info and struct xxx_par do 
 *      not change. This includes the var inside of struct fb_info. 
 *      Do NOT change these. This function can be called on its own if we
 *      intent to only test a mode and not actually set it. The stuff in 
 *      modedb.c is a example of this. If the var passed in is slightly 
 *      off by what the hardware can support then we alter the var PASSED in
 *      to what we can do.
 *
 *      For values that are off, this function must round them _up_ to the
 *      next value that is supported by the hardware.  If the value is
 *      greater than the highest value supported by the hardware, then this
 *      function must return -EINVAL.
 *
 *      Exception to the above rule:  Some drivers have a fixed mode, ie,
 *      the hardware is already set at boot up, and cannot be changed.  In
 *      this case, it is more acceptable that this function just return
 *      a copy of the currently working var (info->var). Better is to not
 *      implement this function, as the upper layer will do the copying
 *      of the current var for you.
 *
 *      Note:  This is the only function where the contents of var can be
 *      freely adjusted after the driver has been registered. If you find
 *      that you have code outside of this function that alters the content
 *      of var, then you are doing something wrong.  Note also that the
 *      contents of info->var must be left untouched at all times after
 *      driver registration.
 *
 *      Returns negative errno on error, or zero on success.
 *
 *      MB86H61:
 *      Resolution should be equal to virtual resolution,
 *      and offsets be always (0, 0).
 *      X resolution is rounded up to multiples of 128.
 *      X resolution is rounded up to multiples of 2.
 *      Bits per pixel and vmode may not be changed after open.
 */

static int mb86h61fb_check_var(struct fb_var_screeninfo *var,
				struct fb_info *info)
{
	/* Check resolution and position */
	if (var->xres <= 0 || var->yres <= 0)
		return -EINVAL;
	DBG("mb86h61fb_check_var, 1.(xres,yres) (%d %d) on (xres_virt,yres_virt)(%d %d)\n", var->xres, var->yres,
					       var->xres_virtual,
					       var->yres_virtual);
	var->xoffset = var->yoffset = 0;

	if(info->fix.ypanstep)
	{
		var->yoffset=var->yres;
	}
	if (var->xres_virtual < var->xres + var->xoffset)
		var->xres_virtual = var->xres + var->xoffset;
	if (var->yres_virtual < var->yres + var->yoffset)
		var->yres_virtual = var->yres + var->yoffset*info->fix.ypanstep;
	var->xres_virtual = H61FB_XRES(var->xres_virtual,
					    info->var.bits_per_pixel);
	var->yres_virtual = H61FB_YRES(var->yres_virtual);
	DBG("mb86h61fb_check_var, 2.(xres,yres) (%d %d) on (xres_virt,yres_virt)(%d %d)\n", var->xres, var->yres,
					       var->xres_virtual,
					       var->yres_virtual);
	DBG("bpp=%d, vmode=%d\n", var->bits_per_pixel, var->vmode);
	DBG("bpp=%d, vmode=%d\n", info->var.bits_per_pixel, info->var.vmode);
	var->yoffset=0;
#if H61FB_X11_HACK
	var->bits_per_pixel = info->var.bits_per_pixel;
	mb86h61fb_rgb(var->bits_per_pixel, var,info->par);
#endif /* H61FB_X11_HACK */
	/* Check visual mode */
	if (var->bits_per_pixel != info->var.bits_per_pixel)
		return -EINVAL;
	ASSERT((var->bits_per_pixel == 4  || var->bits_per_pixel == 8 ||
		var->bits_per_pixel == 16 || var->bits_per_pixel == 32),
		"var->bits_per_pixel == 4, 8, 16 or 32");
#if H61FB_X11_HACK
	var->vmode = info->var.vmode;
#else
	if (var->vmode != info->var.vmode)
		return -EINVAL;
#endif /* H61FB_X11_HACK */
	ASSERT((var->vmode == FB_VMODE_NONINTERLACED ||
		var->vmode == FB_VMODE_INTERLACED),
		"var->vmode is NONINTERLACED or INTERLACED");

	/* Check buffer size */
	if (LINE_LENGTH(var->xres_virtual, var->bits_per_pixel)
	    * var->yres_virtual > info->fix.smem_len)
		return -ENOMEM;

	return 0;
}

/**
 *      xxxfb_set_par - Optional function. Alters the hardware state.
 *      @info: frame buffer structure that represents a single frame buffer
 *
 *      Using the fb_var_screeninfo in fb_info we set the resolution of the
 *      this particular framebuffer. This function alters the par AND the
 *      fb_fix_screeninfo stored in fb_info. It doesn't not alter var in 
 *      fb_info since we are using that data. This means we depend on the
 *      data in var inside fb_info to be supported by the hardware. 
 *
 *      This function is also used to recover/restore the hardware to a
 *      known working state.
 *
 *      xxxfb_check_var is always called before xxxfb_set_par to ensure that
 *      the contents of var is always valid.
 *
 *      Again if you can't change the resolution you don't need this function.
 *
 *      However, even if your hardware does not support mode changing,
 *      a set_par might be needed to at least initialize the hardware to
 *      a known working state, especially if it came back from another
 *      process that also modifies the same hardware, such as X.
 *
 *      If this is the case, a combination such as the following should work:
 *
 *      static int xxxfb_check_var(struct fb_var_screeninfo *var,
 *                                struct fb_info *info)
 *      {
 *              *var = info->var;
 *              return 0;
 *      }
 *
 *      static int xxxfb_set_par(struct fb_info *info)
 *      {
 *              init your hardware here
 *      }
 *
 *      Returns negative errno on error, or zero on success.
 */

static int mb86h61fb_set_par(struct fb_info *info)
{
	FAPI_SYS_HandleT handle = (FAPI_SYS_HandleT)info->par;
	FAPI_VISCALE_ScalerParamsT scalerParams;
	int retval;
	FAPI_VISCALE_SystemModeEnumT sysMode;
	FAPI_VISCALE_OpenParamsT params;

	ASSERT(handle != NULL, "handle != NULL");
	retval = viscale_funcStr.Stop(handle);
	if (retval != FAPI_OK)
	{
		DBG("%s: viscale_funcStr.Stop failed\n",__FUNCTION__);	
		return  -ENODEV;
	}

	retval = viscale_funcStr.Close(handle);
	if (retval != FAPI_OK)
	{
		DBG("%s: viscale_funcStr.Close failed\n",__FUNCTION__);	
		return  -ENODEV;
	}
	else 
	{
		info->par = NULL;
		handle = NULL;
	}
	/* Open the device */
	params.version = FAPI_VISCALE_VERSION;
	if(H61FB0 == info->node)
	{
  		params.scalerType = FAPI_VISCALE_SCALER_OSD1;
	}
	else if(H61FB1 == info->node) 
	{ 
 		params.scalerType = FAPI_VISCALE_SCALER_OSD2;     
	}
	params.encoderType = FAPI_VISCALE_ENCODER_NTSC;
	
	handle = viscale_funcStr.Open(&params, &retval);
	if (handle != NULL)
	{
		info->par = (void *)handle;
	}
	else if ((retval == FAPI_VISCALE_ERR_NOT_INITIALIZED) ||
		(retval == FAPI_VISCALE_ERR_EXCEEDED_MAX_HANDLES)) 
	{
		DBG("%s: viscale_funcStr.Open\n",__FUNCTION__);		
		return -ENODEV;
	}
	else 
	{
		DBG("%s: viscale_funcStr.Open\n",__FUNCTION__);			
		return -EINVAL;
	}
	
	/* Bind VIENC handle and get HD output size */
	retval = FAPEXFDLVIOUT_SetLinuxFB((void *)handle);
	if (retval == FAPEX_VIOUT_ERR_INVALID_STATUS) 
	{
		DBG("%s: FAPEXFDLVIOUT_SetLinuxFB failed\n",__FUNCTION__);
		return  -ENODEV;
	} 
	else if (retval != FAPI_OK) 
	{
		DBG("%s: FAPEXFDLVIOUT_SetLinuxFB failed\n",__FUNCTION__);	
		return-EINVAL;
	}
	
	sysMode = viscale_funcStr.GetSystemType(handle);
	
	switch (sysMode) {
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x2397P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x24P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x25P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x2997P:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x30P:
		outputWidth  = 1920;
		outputHeight = 1080;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x50I:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x5994I:
	case FAPI_VISCALE_SYSTEM_MODE_1920x1080x60I:
		outputWidth  = 1920;
		outputHeight = 540;
		progressive  = 0;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x2397P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x24P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x25P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x2997P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x30P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x50P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x5994P:
	case FAPI_VISCALE_SYSTEM_MODE_1280x720x60P:
		outputWidth  = 1280;
		outputHeight = 720;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x576x50P:
		outputWidth  = 720;
		outputHeight = 576;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x576x50I:
		outputWidth  = 720;
		outputHeight = 576;
		progressive  = 0;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x480x5994P:
		outputWidth  = 720;
		outputHeight = 480;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_720x480x5994I:
		outputWidth  = 720;
		outputHeight = 480;
		progressive  = 0;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_LCD_640x480x50P_LQ080V3DG01:
		outputWidth  = 640;
		outputHeight = 480;
		progressive  = 1;
		break;
	case FAPI_VISCALE_SYSTEM_MODE_LCD_1024x768x50P_SVA150XG10TB:
		outputWidth  = 1024;
		outputHeight = 768;
		progressive  = 1;
		break;
	default:
		DBG("%s: invalid sysMode %d\n",__FUNCTION__,sysMode);
		return  -EINVAL;
	}
	DBG("mb86h61fb_set_par(),yres:%d,yres_virt:%d\n",info->var.yres,info->var.yres_virtual);
	/* Start the device */
#if H61FB_ENABLE
	scalerParams.layerEnable = 0;
#else
	scalerParams.layerEnable = 1;
#endif /* H61FB_ENABLE */
	scalerParams.osdLayout = FAPI_VISCALE_OSD_LAYOUT_ARGB;
	switch (info->var.bits_per_pixel) {
	case 4:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_CLUT_4BIT;
		break;
	case 8:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_CLUT_8BIT;
		break;
	case 16:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_5551;
		break;
	case 32:
		scalerParams.osdMode = FAPI_VISCALE_OSD_MODE_32_24BIT;
		break;
	}

	DBG("mb86h61fb_set_par(),info->var.yoffset:%d,ypanstep=%d\n",info->var.yoffset,info->fix.ypanstep);
	scalerParams.picBase = (unsigned int)(info->screen_base + (info->fix.line_length*info->var.yoffset));
	DBG("mb86h61fb_set_par(),picBase :0x%x\n",scalerParams.picBase);
	scalerParams.frameField = progressive;
#if H61FB_ALPHA
	scalerParams.useGlobalAlpha = 0;
#else
	scalerParams.useGlobalAlpha = 1;
	scalerParams.globalAlpha = 255;
#endif /* H61FB_ALPHA */
	
	memset(scalerParams.colourLut, 0, sizeof(scalerParams.colourLut));
	scalerParams.inputPicWidth = info->var.xres_virtual;
	scalerParams.inputPicHeight = info->var.yres_virtual;
	scalerParams.inputWindowX0 = 0;
	scalerParams.inputWindowY0 = 0;
	scalerParams.inputWindowWidth = info->var.xres;
	scalerParams.inputWindowHeight = info->var.yres;
	scalerParams.outputWindowX0 = 0;
	scalerParams.outputWindowY0 = 0;
	scalerParams.outputWindowWidth = outputWidth;
	scalerParams.outputWindowHeight = outputHeight;
	
	scalerParams.endianness = FAPEX_VISCALE_ENDIAN_DEFAULT;
	scalerParams.pixelOrder = FAPEX_VISCALE_PIXEL_ORDER_DEFAULT;

	retval = viscale_funcStr.Start(handle, &scalerParams);
	if (retval != FAPI_OK) 
	{
		DBG("%s: viscale_funcStr.Start \n",__FUNCTION__);		
		return  -EINVAL;
	}
	
#if H61FB_ALPHA
	iowrite32(0xff,  MB86HXX_VIDOUT_VA_BASE + REG_OSD1_A1);
	iowrite32(0xff, MB86HXX_VIDOUT_VA_BASE + REG_OSD1_A2);/*0x0 --> 0xff by Thomas*/
#endif /* H61FB_ALPHA */
	
	if(H61FB1 != info->node)
	{
		retval = viscale_funcStr.Resize(handle, &scalerParams);
		if (retval != FAPI_OK) 
		{
			if (retval == FAPI_VISCALE_ERR_OPERATION_TIMEOUT)
			{
				DBG("%s: viscale_funcStr.Resize,timeout \n",__FUNCTION__);		
				return  -ENODEV;
			}
			else if (retval != FAPI_OK)
			{
				DBG("%s: viscale_funcStr.Resize,fail \n",__FUNCTION__);			
				return  -EINVAL;
			}
		}
	}
#if H61FB_ENABLE
	retval = viscale_funcStr.EnableDisable(handle,
		FAPI_VISCALE_PARAMETER_ON);
	ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
		"retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
	if (retval != FAPI_OK) 
	{
		return -ENODEV;
	}
#endif /* H61FB_ENABLE */
	
	/* Update line length */
	info->fix.line_length = LINE_LENGTH(info->var.xres_virtual,
					    info->var.bits_per_pixel);
	DBG("mb86h61fb_set_par(),line_length=%d from xres_v=%d bpp=%d\n", info->fix.line_length,
		info->var.xres_virtual, info->var.bits_per_pixel);
	return 0;
}

/**
 *      xxxfb_setcolreg - Optional function. Sets a color register.
 *      @regno: Which register in the CLUT we are programming 
 *      @red: The red value which can be up to 16 bits wide 
 *      @green: The green value which can be up to 16 bits wide 
 *      @blue:  The blue value which can be up to 16 bits wide.
 *      @transp: If supported, the alpha value which can be up to 16 bits wide.
 *      @info: frame buffer info structure
 * 
 *      Set a single color register. The values supplied have a 16 bit
 *      magnitude which needs to be scaled in this function for the hardware. 
 *      Things to take into consideration are how many color registers, if
 *      any, are supported with the current color visual. With truecolor mode
 *      no color palettes are supported. Here a pseudo palette is created
 *      which we store the value in pseudo_palette in struct fb_info. For
 *      pseudocolor mode we have a limited color palette. To deal with this
 *      we can program what color is displayed for a particular pixel value.
 *      DirectColor is similar in that we can program each color field. If
 *      we have a static colormap we don't need to implement this function. 
 * 
 *      Returns negative errno on error, or zero on success.
 *
 *      MB86H61:
 *      Have concerns on PSEUDOCOLOR and TRUECOLOR only.
 */

static int mb86h61fb_setcolreg(unsigned int regno, unsigned int red,
			       unsigned int green, unsigned int blue,
			       unsigned int transp, struct fb_info *info)
{
	FAPI_SYS_HandleT handle = (FAPI_SYS_HandleT)info->par;
	unsigned int color;
	int retval = 0;

	ASSERT(handle != NULL, "handle != NULL");
    /* Directcolor:
     *   var->{color}.offset contains start of bitfield
     *   var->{color}.length contains length of bitfield
     *   {hardwarespecific} contains width of DAC
     *   pseudo_palette[X] is programmed to (X << red.offset) |
     *                                      (X << green.offset) |
     *                                      (X << blue.offset)
     *   RAMDAC[X] is programmed to (red, green, blue)
     *   color depth = SUM(var->{color}.length)
     *
     * Pseudocolor:
     *    var->{color}.offset is 0
     *    var->{color}.length contains width of DAC or the number of unique
     *                        colors available (color depth)
     *    pseudo_palette is not used
     *    RAMDAC[X] is programmed to (red, green, blue)
     *    color depth = var->{color}.length
     *
     * Static pseudocolor:
     *    same as Pseudocolor, but the RAMDAC is not programmed (read-only)
     *
     * Mono01/Mono10:
     *    Has only 2 values, black on white or white on black (fg on bg),
     *    var->{color}.offset is 0
     *    white = (1 << var->{color}.length) - 1, black = 0
     *    pseudo_palette is not used
     *    RAMDAC does not exist
     *    color depth is always 2
     *
     * Truecolor:
     *    does not use RAMDAC (usually has 3 of them).
     *    var->{color}.offset contains start of bitfield
     *    var->{color}.length contains length of bitfield
     *    pseudo_palette is programmed to (red << red.offset) |
     *                                    (green << green.offset) |
     *                                    (blue << blue.offset) |
     *                                    (transp << transp.offset)
     *    RAMDAC does not exist
     *    color depth = SUM(var->{color}.length})
     *
     *  The color depth is used by fbcon for choosing the logo and also
     *  for color palette transformation if color depth < 4
     *
     *  As can be seen from the above, the field bits_per_pixel is _NOT_
     *  a criteria for describing the color visual.
     *
     *  A common mistake is assuming that bits_per_pixel <= 8 is pseudocolor,
     *  and higher than that, true/directcolor.  This is incorrect, one needs
     *  to look at the fix->visual.
     *
     *  Another common mistake is using bits_per_pixel to calculate the color
     *  depth.  The bits_per_pixel field does not directly translate to color
     *  depth. You have to compute for the color depth (using the color
     *  bitfields) and fix->visual as seen above.
     */

	color =
         ((red   << info->var.red.offset)   |
         (green  << info->var.green.offset) |
         (blue   << info->var.blue.offset)  |
         (transp << info->var.transp.offset));
#if DEBUG
	if (regno < 16)
		DBG("setcolreg[%d]: 0x%08x = (%d, %d, %d, %d)\n",
		     regno, color, red, green, blue, transp);
#endif /* DEBUG */

	if ((info->fix.visual == FB_VISUAL_PSEUDOCOLOR) &&
	    (regno < H61FB_CMAP_LEN)) {
	/*
	 * This is the point where the function feeds the color to the hardware
	 * palette after converting the colors to something acceptable by
	 * the hardware. Note, only FB_VISUAL_DIRECTCOLOR and
	 * FB_VISUAL_PSEUDOCOLOR visuals need to write to the hardware palette.
	 * If you have code that writes to the hardware CLUT, and it's not
	 * any of the above visuals, then you are doing something wrong.
	 */

		/* Change HW CLUT */
		retval = viscale_funcStr.UpdateSingleClut(handle, regno, color);
		ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
		      "retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
#if DEBUG
		if (regno < 16)
			DBG("VISCALE UpClut returned 0x%x.\n", retval);
#endif /* DEBUG */
		if (retval != FAPI_OK)
			retval = -EINVAL;
		else
			retval = 0;
	} else if ((info->fix.visual == FB_VISUAL_TRUECOLOR) &&
		   (regno < 16)) {

	/* This is the point were you need to fill up the contents of
	 * info->pseudo_palette. This structure is used _only_ by fbcon, thus
	 * it only contains 16 entries to match the number of colors supported
	 * by the console. The pseudo_palette is used only if the visual is
	 * in directcolor or truecolor mode.  With other visuals, the
	 * pseudo_palette is not used. (This might change in the future.)
	 *
	 * The contents of the pseudo_palette is in raw pixel format.  Ie, each
	 * entry can be written directly to the framebuffer without any conversion.
	 * The pseudo_palette is (void *).  However, if using the generic
	 * drawing functions (cfb_imageblit, cfb_fillrect), the pseudo_palette
	 * must be casted to (u32 *) _regardless_ of the bits per pixel. If the
	 * driver is using its own drawing functions, then it can use whatever
	 * size it wants.
	 */
 
		/* Change pseudo palette for fb console */
		((unsigned int *)(info->pseudo_palette))[regno] = color;
	} else
		retval = -EINVAL;

	return retval;
}

/**
 *      xxxfb_blank - NOT a required function. Blanks the display.
 *      @blank_mode: the blank mode we want. 
 *      @info: frame buffer structure that represents a single frame buffer
 *
 *      Blank the screen if blank_mode != FB_BLANK_UNBLANK, else unblank.
 *      Return 0 if blanking succeeded, != 0 if un-/blanking failed due to
 *      e.g. a video mode which doesn't support it.
 *
 *      Implements VESA suspend and powerdown modes on hardware that supports
 *      disabling hsync/vsync:
 *
 *      FB_BLANK_NORMAL = display is blanked, syncs are on.
 *      FB_BLANK_HSYNC_SUSPEND = hsync off
 *      FB_BLANK_VSYNC_SUSPEND = vsync off
 *      FB_BLANK_POWERDOWN =  hsync and vsync off
 *
 *      If implementing this function, at least support FB_BLANK_UNBLANK.
 *      Return !0 for any modes that are unimplemented.
 *
 * 	MB86H61:
 * 	Support BLANK_NORMAL and BLANK_UNBLANK only.
 */

static int mb86h61fb_blank(int blank_mode, struct fb_info *info)
{
	FAPI_SYS_HandleT handle = (FAPI_SYS_HandleT)info->par;
	FAPI_VISCALE_ParameterModeEnumT mode;
	int retval = 0;

	ASSERT(handle != NULL, "handle != NULL");
	switch (blank_mode) {
	case FB_BLANK_NORMAL:
		mode = FAPI_VISCALE_PARAMETER_OFF;
		retval = viscale_funcStr.EnableDisable(handle, mode);
		ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
		      "retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
		if (retval != FAPI_OK)
			retval = -ENODEV;
		else
			retval = 0;
		break;
	case FB_BLANK_UNBLANK:
		mode = FAPI_VISCALE_PARAMETER_ON;
		retval = viscale_funcStr.EnableDisable(handle, mode);
		ASSERT(retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY,
		      "retval != FAPI_VISCALE_ERR_UNSUPPORTED_FUNCTIONALITY");
		if (retval != FAPI_OK)
			retval = -ENODEV;
		else
			retval = 0;
		break;
	default:
		retval = -EINVAL;
	}

	return retval;
}
static int
mb86h61fb_pan_display(struct fb_var_screeninfo *var,struct fb_info *info )
{
	int size;
	DBG("mb86h61fb_pan_display(),yoffset:%d\n",var->yoffset);
	size=info->fix.line_length*(var->yoffset);
	*(volatile uint32_t* )0xff614000=io_virt_to_phys(info->screen_base)+ size;
	return 0;	
}
static struct fb_ops mb86h61fb_ops = {
	.owner          = THIS_MODULE,
	.fb_open        = mb86h61fb_open,
	.fb_release     = mb86h61fb_release,
	.fb_check_var   = mb86h61fb_check_var,
	.fb_set_par     = mb86h61fb_set_par,
	.fb_setcolreg   = mb86h61fb_setcolreg,
	.fb_blank       = mb86h61fb_blank,
	.fb_fillrect    = cfb_fillrect,
	.fb_copyarea    = cfb_copyarea,
	.fb_imageblit   = cfb_imageblit,
	.fb_pan_display = mb86h61fb_pan_display,

};


/*
 * Initialization & Termination:
 * 	mb86h61fb_setup()
 * 	mb86h61fb_probe()
 * 	mb86h61fb_remove()
 * 	mb86h61fb_init()
 * 	mb86h61fb_exit()
 */

#ifndef MODULE
/*
 * mb86h61fb_setup - Parse command line option.
 * @options: option string
 *
 * Always return 0
 */

static int __init mb86h61fb_setup(char *options)
{
	char *opt;

	if (!options || !*options)
		return 0;
	/*
	 *  exmaple:
	 * 	options="mb86h61fb:800x600,bpp=16"
	 * 	  --->  mode="800x600",
	 * 	        bpp=16
	 */
	while ((opt = strsep(&options, ",")) != NULL) {
		if (!*opt)
			continue;
		else if (!strncmp(opt, "bpp=",4))
			bpp = simple_strtoul(opt + 4, NULL, 0);
		else
			mode = opt;
	}
	return 0;
}
#endif /* MODULE */

/*
 * mb86h61fb_probe - Initialize frame buffer device.
 * @dev: platform device structure
 *
 */

static int __init mb86h61fb_probe(struct platform_device *dev)
{
	struct fb_info *info;
	int retval = 0,i;

	/* Initialize info */
	info = framebuffer_alloc((sizeof(FAPI_SYS_HandleT) + 
				  sizeof(void *)*H61FB_CMAP_LEN), &dev->dev);
	if (!info)
		return -ENOMEM;
	info->flags = FBINFO_DEFAULT | FBINFO_HWACCEL_NONE;
	info->var = mb86h61fb_var_default;
	info->fbops = &mb86h61fb_ops;
	info->pseudo_palette = &info->par + sizeof(FAPI_SYS_HandleT);
	info->par = NULL;
	
	/* Allocate color map */
	retval = fb_alloc_cmap(&info->cmap, H61FB_CMAP_LEN, 1);
	if (retval < 0) {
		framebuffer_release(info);
		return retval;
	}

	/* Register frame buffer device */
	retval = register_framebuffer(info);
	if (retval < 0) {
		fb_dealloc_cmap(&info->cmap);
		framebuffer_release(info);
		return retval;
	}
	platform_set_drvdata(dev, info);
	return 0;
}

/*
 * mb86h61fb_remove - Clean up frame buffer device.
 * @dev: platform device structure
 *
 */

static int mb86h61fb_remove(struct platform_device *dev)
{
	struct fb_info *info;

	DBG("mb86h61fb_remove()\n");
	info = platform_get_drvdata(dev);
	if (info) {
		unregister_framebuffer(info);
		fb_dealloc_cmap(&info->cmap);
		framebuffer_release(info);
	}
	return 0;
}

static void mb86h61fb_platform_release(struct device *device)
{
        /* This is called when the reference count goes to zero. */
}

static struct platform_driver mb86h61fb_driver = {
	.probe = mb86h61fb_probe,
	.remove = mb86h61fb_remove,
	.driver = {
		.name = H61FB_NAME,
	},
};

static struct platform_device mb86h61fb_device = {
	.name = H61FB_NAME,
	.id = H61FB0,
	.num_resources = 0,
        .dev = {
		.release = mb86h61fb_platform_release,
	},
};
static struct platform_device mb86h61fb_device1 = {
	.name = H61FB_NAME,
	.id = H61FB1,
	.num_resources = 0,
        .dev = {
		.release = mb86h61fb_platform_release,
	},
};
/*
 * mb86h61fb_init - Initialize platform device.
 */

static int __init mb86h61fb_init(void)
{
	int retval = 0;

#ifndef MODULE
	char *option = NULL;

	/* Identify this device and get command line option */
	if (fb_get_options(H61FB_NAME, &option))
		return  -ENODEV;
	mb86h61fb_setup(option);
#endif /* MODULE */

	/* Register platform device */
	retval = platform_driver_register(&mb86h61fb_driver);
	if (!retval) {
		retval = platform_device_register(&mb86h61fb_device);
		retval = platform_device_register(&mb86h61fb_device1);
		if (retval) {
			platform_driver_unregister(&mb86h61fb_driver);
		}
	}

	return retval;
}

module_init(mb86h61fb_init);

#ifdef MODULE
/*
 * mb86h61fb_exit - Clean up platform device.
 */

static void __exit mb86h61fb_exit(void)
{
	DBG("mb86h61fb_exit()\n");
	platform_device_unregister(&mb86h61fb_device);
	platform_device_unregister(&mb86h61fb_device1);
	platform_driver_unregister(&mb86h61fb_driver);
}

module_exit(mb86h61fb_exit);

MODULE_AUTHOR("Fujitsu Microelectronics Limited");
MODULE_DESCRIPTION("Frame buffer driver for Fujitsu MB86H61");
MODULE_LICENSE("GPL");
#endif /* MODULE */
